@@ -1,9 +1,11 @@
 
 #include "dos_inc.h"
 #include "../src/builtin/zip.h"
+#include "../src/builtin/eval.h"
 #include "../src/builtin/cdplay.h"
 #include "../src/builtin/dosmid.h"
 #include "../src/builtin/mpxplay.h"
+#include "../src/builtin/ne2000.h"
 #include "../src/builtin/glide2x.h"
 #include "../src/builtin/textutil.h"
 #include "../src/builtin/4DOS_img.h"
@@ -17,10 +19,10 @@ extern struct BuiltinFileBlob bfb_DOS4GW_EXE;		// DOS4GW.EXE
 extern struct BuiltinFileBlob bfb_HEXMEM16_EXE;		// HEXMEM16.EXE
 extern struct BuiltinFileBlob bfb_HEXMEM32_EXE;		// HEXMEM32.EXE
 extern struct BuiltinFileBlob bfb_DEBUG_EXE;		// DEBUG.EXE
+extern struct BuiltinFileBlob bfb_REPLACE_EXE;		// REPLACE.EXE
 extern struct BuiltinFileBlob bfb_SORT_EXE;		// SORT.EXE
-extern struct BuiltinFileBlob bfb_TREE_EXE;		// TREE.EXE
 extern struct BuiltinFileBlob bfb_MOVE_EXE;		// MOVE.EXE
-extern struct BuiltinFileBlob bfb_MEM_COM;		// MEM.COM
+extern struct BuiltinFileBlob bfb_MEM_EXE;		// MEM.EXE
 extern struct BuiltinFileBlob bfb_FIND_EXE;		// FIND.EXE
 extern struct BuiltinFileBlob bfb_DOSIDLE_EXE;		// DOSIDLE.EXE
 extern struct BuiltinFileBlob bfb_LASTDRIV_COM;		// LASTDRIV.COM
@@ -31,18 +33,41 @@ extern struct BuiltinFileBlob bfb_EDIT_COM;		// EDIT.COM
 extern struct BuiltinFileBlob bfb_DEVICE_COM;		// DEVICE.COM
 extern struct BuiltinFileBlob bfb_BUFFERS_COM;		// BUFFERS.COM
 extern struct BuiltinFileBlob bfb_COPY_EXE;		// COPY.EXE
+extern struct BuiltinFileBlob bfb_EVAL_EXE;		// EVAL.EXE
+extern struct BuiltinFileBlob bfb_EVAL_HLP;		// EVAL.HLP
 extern struct BuiltinFileBlob bfb_28_COM;		// 28.COM
 extern struct BuiltinFileBlob bfb_28_COM_ega;	// 28.COM
-extern struct BuiltinFileBlob bfb_50_COM;		// 50.COM
 extern struct BuiltinFileBlob bfb_25_COM;		// 25.COM
 extern struct BuiltinFileBlob bfb_25_COM_ega;	// 25.COM
 extern struct BuiltinFileBlob bfb_25_COM_other;	// 25.COM
+extern struct BuiltinFileBlob bfb_50_COM;		// 50.COM
 extern struct BuiltinFileBlob bfb_4DOS_COM;		// 4DOS.COM
 extern struct BuiltinFileBlob bfb_4DOS_HLP;		// 4DOS.HLP
 extern struct BuiltinFileBlob bfb_4HELP_EXE;	// 4HELP.EXE
+extern struct BuiltinFileBlob bfb_BATCOMP_EXE;	// BATCOM.EXE
+extern struct BuiltinFileBlob bfb_EXAMPLES_BTM;	// EXAMPLES.BTM
+extern struct BuiltinFileBlob bfb_LICENSE_TXT;	// LICENSE.TXT
+extern struct BuiltinFileBlob bfb_OPTION_EXE;	// OPTION.EXE
 extern struct BuiltinFileBlob bfb_CDPLAY_EXE;	// CDPLAY.EXE
 extern struct BuiltinFileBlob bfb_DOSMID_EXE;	// DOSMID.EXE
 extern struct BuiltinFileBlob bfb_MPXPLAY_EXE;	// MPXPLAY.EXE
+extern struct BuiltinFileBlob bfb_NE2000_COM;	// NE2000.COM
 extern struct BuiltinFileBlob bfb_GLIDE2X_OVL;	// GLIDE2X.OVL
-extern struct BuiltinFileBlob bfb_TEXTUTIL_ZIP;	// TEXTUTIL.ZIP
-
+extern struct BuiltinFileBlob bfb_VGA_COM;	// VGA.COM
+extern struct BuiltinFileBlob bfb_SCANRES_COM;	// SCANRES.COM
+extern struct BuiltinFileBlob bfb_EGA_COM;	// EGA.COM
+extern struct BuiltinFileBlob bfb_CLR_COM;	// CLR.COM
+extern struct BuiltinFileBlob bfb_CGA_COM;	// CGA.COM
+extern struct BuiltinFileBlob bfb_80X60_COM;	// 80X60.COM
+extern struct BuiltinFileBlob bfb_80X50_COM;	// 80X50.COM
+extern struct BuiltinFileBlob bfb_80X43_COM;	// 80X43.COM
+extern struct BuiltinFileBlob bfb_80X25_COM;	// 80X25.COM
+extern struct BuiltinFileBlob bfb_132X60_COM;	// 132X60.COM
+extern struct BuiltinFileBlob bfb_132X50_COM;	// 132X50.COM
+extern struct BuiltinFileBlob bfb_132X43_COM;	// 132X43.COM
+extern struct BuiltinFileBlob bfb_132X25_COM;	// 132X25.COM
+extern struct BuiltinFileBlob bfb_DEFRAG_EXE;	// DEFRAG.EXE (FreeDOS)
+extern struct BuiltinFileBlob bfb_FDISK_EXE;	// FDISK.EXE (FreeDOS)
+extern struct BuiltinFileBlob bfb_FORMAT_EXE;	// FORMAT.EXE (FreeDOS)
+extern struct BuiltinFileBlob bfb_CHKDSK_EXE;	// CHKDSK.EXE (FreeDOS)
+extern struct BuiltinFileBlob bfb_SYS_COM;	// SYS.COM (FreeDOS)
