@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,10 +19,7 @@
 #ifndef DOSBOX_FPU_H
 #define DOSBOX_FPU_H
 
-#ifndef DOSBOX_MEM_H
-#include "mem.h"
-#endif
-
+#include "logging.h"
 #include "mmx.h"
 
 void FPU_ESC0_Normal(Bitu rm);
@@ -148,11 +145,81 @@ enum FPU_Tag {
 	TAG_Empty = 3
 };
 
-enum FPU_Round {
-	ROUND_Nearest = 0,		
-	ROUND_Down    = 1,
-	ROUND_Up      = 2,	
-	ROUND_Chop    = 3
+template<class T, unsigned bitno, unsigned nbits=1>
+struct RegBit
+{
+	enum { basemask = (1 << nbits) - 1 };
+	enum { mask = basemask << bitno };
+	T& data;
+	RegBit(T& reg) : data(reg) {}
+	template <class T2> RegBit& operator=(T2 val)
+	{
+		data = (data & ~mask) | ((nbits > 1 ? val & basemask : !!val) << bitno);
+		return *this;
+	}
+	operator unsigned() const { return (data & mask) >> bitno; }
+};
+
+struct FPUControlWord
+{
+	uint16_t reg;
+	RegBit<FPUControlWord, 0>     IM;  // Invalid operation mask
+	RegBit<FPUControlWord, 1>     DM;  // Denormalized operand mask
+	RegBit<FPUControlWord, 2>     ZM;  // Zero divide mask
+	RegBit<FPUControlWord, 3>     OM;  // Overflow mask
+	RegBit<FPUControlWord, 4>     UM;  // Underflow mask
+	RegBit<FPUControlWord, 5>     PM;  // Precision mask
+	RegBit<FPUControlWord, 7>     M;   // Interrupt mask   (8087-only)
+	RegBit<FPUControlWord, 8, 2>  PC;  // Precision control
+	RegBit<FPUControlWord, 10, 2> RC;  // Rounding control
+	RegBit<FPUControlWord, 12>    IC;  // Infinity control (8087/80287-only)
+
+	enum
+	{
+		mask8087     = 0x1fff,
+		maskNon8087  = 0x1f7f,
+		reservedMask = 0x40,
+		initValue    = 0x37f
+	};
+	enum RoundMode
+	{
+		Nearest = 0,
+		Down    = 1,
+		Up      = 2,
+		Chop    = 3
+	};
+
+	FPUControlWord() : reg(initValue), IM(*this), DM(*this), ZM(*this), OM(*this),
+					   UM(*this), PM(*this), M(*this), PC(*this), RC(*this), IC(*this) {}
+	FPUControlWord(const FPUControlWord& other) = default;
+	FPUControlWord& operator=(const FPUControlWord& other)
+	{
+		reg = other.reg;
+		return *this;
+	}
+	template<class T>
+	FPUControlWord& operator=(T val)
+	{
+		reg = (val & (CPU_ArchitectureType==CPU_ARCHTYPE_8086 ? mask8087 : maskNon8087)) | reservedMask;
+		return *this;
+	}
+	operator unsigned() const
+	{
+		return reg;
+	}
+	template <class T>
+	FPUControlWord& operator |=(T val)
+	{
+		*this = reg | val;
+		return *this;
+	}
+	void init() { reg = initValue; }
+	FPUControlWord allMasked() const
+	{
+		auto masked = *this;
+		masked |= IM.mask | DM.mask | ZM.mask | OM.mask | UM.mask | PM.mask;
+		return masked;
+	}
 };
 
 typedef struct {
@@ -169,10 +236,9 @@ typedef struct {
 	bool		use80[9];		// if set, use the 80-bit precision version
 #endif
 	FPU_Tag		tags[9];
-	uint16_t		cw,cw_mask_all;
+	FPUControlWord  cw;
 	uint16_t		sw;
 	uint32_t		top;
-	FPU_Round	round;
 } FPU_rec;
 
 
@@ -198,18 +264,6 @@ static INLINE void FPU_SetTag(uint16_t tag){
 		fpu.tags[i] = static_cast<FPU_Tag>((tag >>(2*i))&3);
 }
 
-static INLINE void FPU_SetCW(Bitu word){
-	// HACK: Bits 13-15 are not defined. Apparently, one program likes to test for
-	//       Cyrix EMC87 by trying to set bit 15. We want the test program to see
-	//       us as an Intel 287 when cputype == 286.
-	word &= 0x7FFF;
-
-	fpu.cw = (uint16_t)word;
-	fpu.cw_mask_all = (uint16_t)(word | 0x3f);
-	fpu.round = (FPU_Round)((word >> 10) & 3);
-}
-
-
 static INLINE uint8_t FPU_GET_TOP(void) {
 	return (fpu.sw & 0x3800U) >> 11U;
 }
