@@ -1,9 +1,4 @@
 
-#include "dosbox.h"
-
-#include <vector>
-#include <string>
-
 #ifndef DOSBOX_REGIONALLOCTRACKING_H
 #define DOSBOX_REGIONALLOCTRACKING_H
 
@@ -40,4 +35,3 @@ public:
 };
 
 #endif /* DOSBOX_REGIONALLOCTRACKING_H */
-
