@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -20,17 +20,7 @@
 #ifndef DOSBOX_SHELL_H
 #define DOSBOX_SHELL_H
 
-#include <ctype.h>
-#ifndef DOSBOX_DOSBOX_H
-#include "dosbox.h"
-#endif
-#ifndef DOSBOX_PROGRAMS_H
 #include "programs.h"
-#endif
-
-#include <string>
-#include <list>
-#include <map>
 
 #include <SDL.h>
 #if SDL_VERSION_ATLEAST(2, 0, 0)
@@ -93,12 +83,14 @@ private:
 	
 private:
 	void ProcessCmdLineEnvVarStitution(char * line);
+	static bool hasExecutableExtension(const char* name);
 
 public:
 
 	DOS_Shell();
 	virtual ~DOS_Shell();
 
+	void Prepare(void);
     /*! \brief      Program entry point, when the command is run
      */
 	void Run(void);
@@ -117,6 +109,10 @@ public:
      */
 	Bitu GetRedirection(char *s, char **ifn, char **ofn, char **toc,bool * append);
 
+    /*! \brief      Build Tab completion
+     */
+	bool BuildCompletions(char * line, uint16_t str_len);
+
     /*! \brief      Command line input and keyboard handling
      */
 	void InputCommand(char * line);
@@ -256,6 +252,10 @@ public:
      */
 	void CMD_SHIFT(char * args);
 
+    /*! \brief      List directory tree (TREE)
+     */
+	void CMD_TREE(char * args);
+
     /*! \brief      File verification switch
      */
 	void CMD_VERIFY(char * args);
@@ -286,7 +286,10 @@ public:
 
     /*! \brief      Change country code
      */
+	void CMD_CHCP(char * args);
 	void CMD_COUNTRY(char * args);
+	void CMD_PUSHD(char * args);
+	void CMD_POPD(char * args);
     void CMD_TRUENAME(char * args);
     void CMD_DXCAPTURE(char * args);
 
