@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,15 +16,12 @@
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
-
-#include <stdlib.h>
+#include <assert.h>
 #include <string.h>
 
-#include "dosbox.h"
 #include "callback.h"
 #include "logging.h"
 #include "bios.h"
-#include "mem.h"
 #include "cpu.h"
 
 #if C_EMSCRIPTEN
