@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,11 +19,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 
-#include "dosbox.h"
-#include "mem.h"
 #include "cpu.h"
 #include "lazyflags.h"
-#include "inout.h"
 #include "callback.h"
 #include "pic.h"
 #include "fpu.h"
@@ -222,7 +219,7 @@ Bits CPU_Core_Simple_Trap_Run(void) {
     cpu.trap_skip = false;
 
     Bits ret=CPU_Core_Simple_Run();
-    if (!cpu.trap_skip) CPU_HW_Interrupt(1);
+    if (!cpu.trap_skip) CPU_DebugException(DBINT_STEP,reg_eip);
     CPU_Cycles = oldCycles-1;
     cpudecoder = &CPU_Core_Simple_Run;
 
