@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -29,7 +29,9 @@
 #include "cpu.h"
 #include "debug.h"
 #include "setup.h"
+#include "logging.h"
 
+extern bool dos_kernel_disabled;
 PagingBlock paging;
 
 // Pagehandler implementation
@@ -267,10 +269,9 @@ void PrintPageInfo(const char* string, PhysPt lin_addr, bool writing, bool prepa
 */
 
 bool use_dynamic_core_with_paging = false; /* allow dynamic core even with paging (AT YOUR OWN RISK!!!!) */
+bool auto_determine_dynamic_core_paging = false; /* enable use_dynamic_core_with_paging when paging is enabled */
 bool dosbox_allow_nonrecursive_page_fault = false;	/* when set, do nonrecursive mode (when executing instruction) */
 
-bool CPU_IsDynamicCore(void);
-
 void PAGING_PageFault(PhysPt lin_addr,Bitu page_addr,Bitu faultcode) {
 	/* Save the state of the cpu cores */
 	LazyFlags old_lflags;
@@ -280,7 +281,7 @@ void PAGING_PageFault(PhysPt lin_addr,Bitu page_addr,Bitu faultcode) {
 	cpudecoder=&PageFaultCore;
 	paging.cr2=lin_addr;
 	PF_Entry * entry=&pf_queue.entries[pf_queue.used++];
-	LOG(LOG_PAGING,LOG_NORMAL)("PageFault at %X type [%x] queue %d",lin_addr,faultcode,pf_queue.used);
+	LOG(LOG_PAGING,LOG_NORMAL)("PageFault at %lX type [%lx] queue %d",(unsigned long)lin_addr,(unsigned long)faultcode,(int)pf_queue.used);
 //	LOG_MSG("EAX:%04X ECX:%04X EDX:%04X EBX:%04X",reg_eax,reg_ecx,reg_edx,reg_ebx);
 //	LOG_MSG("CS:%04X EIP:%08X SS:%04x SP:%08X",SegValue(cs),reg_eip,SegValue(ss),reg_esp);
 	entry->cs=SegValue(cs);
@@ -295,7 +296,7 @@ void PAGING_PageFault(PhysPt lin_addr,Bitu page_addr,Bitu faultcode) {
 #endif
 	DOSBOX_RunMachine();
 	pf_queue.used--;
-	LOG(LOG_PAGING,LOG_NORMAL)("Left PageFault for %x queue %d",lin_addr,pf_queue.used);
+	LOG(LOG_PAGING,LOG_NORMAL)("Left PageFault for %lx queue %d",(unsigned long)lin_addr,(int)pf_queue.used);
 	memcpy(&lflags,&old_lflags,sizeof(LazyFlags));
 	cpudecoder=old_cpudecoder;
 //	LOG_MSG("SS:%04x SP:%08X",SegValue(ss),reg_esp);
@@ -312,7 +313,7 @@ static void PAGING_NewPageFault(PhysPt lin_addr, Bitu page_addr, bool prepare_on
 	if (prepare_only) {
 		cpu.exception.which = EXCEPTION_PF;
 		cpu.exception.error = faultcode;
-	} else if (dosbox_allow_nonrecursive_page_fault && !CPU_IsDynamicCore()) {
+	} else if (dosbox_allow_nonrecursive_page_fault) {
 		throw GuestPageFaultException(lin_addr,page_addr,faultcode);
 	} else {
 		// Save the state of the cpu cores
@@ -641,22 +642,22 @@ static INLINE void InitPageCheckPresence(PhysPt lin_addr,bool writing,X86PageEnt
 	Bitu d_index=lin_page >> 10;
 	Bitu t_index=lin_page & 0x3ff;
 	Bitu table_addr=(paging.base.page<<12)+d_index*4;
-	table.load=phys_readd(table_addr);
+	table.load=phys_readd((PhysPt)table_addr);
 	if (!table.block.p) {
 		LOG(LOG_PAGING,LOG_NORMAL)("NP Table");
 		PAGING_PageFault(lin_addr,table_addr,
 			(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04));
-		table.load=phys_readd(table_addr);
+		table.load=phys_readd((PhysPt)table_addr);
 		if (GCC_UNLIKELY(!table.block.p))
 			E_Exit("Pagefault didn't correct table");
 	}
 	Bitu entry_addr=(table.block.base<<12)+t_index*4;
-	entry.load=phys_readd(entry_addr);
+	entry.load=phys_readd((PhysPt)entry_addr);
 	if (!entry.block.p) {
 //		LOG(LOG_PAGING,LOG_NORMAL)("NP Page");
 		PAGING_PageFault(lin_addr,entry_addr,
 			(writing?0x02:0x00) | (((cpu.cpl&cpu.mpl)==0)?0x00:0x04));
-		entry.load=phys_readd(entry_addr);
+		entry.load=phys_readd((PhysPt)entry_addr);
 		if (GCC_UNLIKELY(!entry.block.p))
 			E_Exit("Pagefault didn't correct page");
 	}
@@ -667,7 +668,7 @@ static INLINE bool InitPageCheckPresence_CheckOnly(PhysPt lin_addr,bool writing,
 	Bitu d_index=lin_page >> 10;
 	Bitu t_index=lin_page & 0x3ff;
 	Bitu table_addr=(paging.base.page<<12)+d_index*4;
-	table.load=phys_readd(table_addr);
+	table.load=phys_readd((PhysPt)table_addr);
 	if (!table.block.p) {
 		paging.cr2=lin_addr;
 		cpu.exception.which=EXCEPTION_PF;
@@ -675,7 +676,7 @@ static INLINE bool InitPageCheckPresence_CheckOnly(PhysPt lin_addr,bool writing,
 		return false;
 	}
 	Bitu entry_addr=(table.block.base<<12)+t_index*4;
-	entry.load=phys_readd(entry_addr);
+	entry.load=phys_readd((PhysPt)entry_addr);
 	if (!entry.block.p) {
 		paging.cr2=lin_addr;
 		cpu.exception.which=EXCEPTION_PF;
@@ -856,11 +857,11 @@ initpage_retry:
 		if (paging.enabled) {
 			X86PageEntry table;
 			X86PageEntry entry;
-			InitPageCheckPresence(lin_addr,false,table,entry);
+			InitPageCheckPresence((PhysPt)lin_addr,false,table,entry);
 
 			if (!table.block.a) {
 				table.block.a=1;		//Set access
-				phys_writed((paging.base.page<<12)+(lin_page >> 10)*4,table.load);
+				phys_writed((PhysPt)((paging.base.page<<12)+(lin_page >> 10)*4),table.load);
 			}
 			if (!entry.block.a) {
 				entry.block.a=1;					//Set access
@@ -1319,6 +1320,8 @@ public:
 		return true;
 	}
 	bool InitPage(PhysPt lin_addr, bool writing, bool prepare_only) {
+		(void)prepare_only;
+		(void)writing;
 		Bitu lin_page=lin_addr >> 12;
 		Bitu phys_page;
 		if (paging.enabled) {
@@ -1329,12 +1332,12 @@ public:
 			InitPageCheckPresence(lin_addr,true,table,entry);
 
 			LOG(LOG_PAGING,LOG_NORMAL)("Page access denied: cpl=%i, %x:%x:%x:%x",
-				cpu.cpl,entry.block.us,table.block.us,entry.block.wr,table.block.wr);
+				(int)cpu.cpl,entry.block.us,table.block.us,entry.block.wr,table.block.wr);
 			PAGING_PageFault(lin_addr,(table.block.base<<12)+(lin_page & 0x3ff)*4,0x07);
 
 			if (!table.block.a) {
 				table.block.a=1;		//Set access
-				phys_writed((paging.base.page<<12)+(lin_page >> 10)*4,table.load);
+				phys_writed((PhysPt)((paging.base.page<<12)+(lin_page >> 10)*4),table.load);
 			}
 			if ((!entry.block.a) || (!entry.block.d)) {
 				entry.block.a=1;	//Set access
@@ -1351,6 +1354,7 @@ public:
 		return false;
 	}
 	Bitu InitPageCheckOnly(PhysPt lin_addr,uint32_t val) {
+		(void)val;
 		Bitu lin_page=lin_addr >> 12;
 		if (paging.enabled) {
 			if (!USERWRITE_PROHIBITED) return 2;
@@ -1361,7 +1365,7 @@ public:
 
 			if (InitPage_CheckUseraccess(entry.block.us,table.block.us) || (((entry.block.wr==0) || (table.block.wr==0)))) {
 				LOG(LOG_PAGING,LOG_NORMAL)("Page access denied: cpl=%i, %x:%x:%x:%x",
-					cpu.cpl,entry.block.us,table.block.us,entry.block.wr,table.block.wr);
+					(int)cpu.cpl,entry.block.us,table.block.us,entry.block.wr,table.block.wr);
 				paging.cr2=lin_addr;
 				cpu.exception.which=EXCEPTION_PF;
 				cpu.exception.error=0x07;
@@ -1382,11 +1386,11 @@ public:
 		if (paging.enabled) {
 			X86PageEntry table;
 			X86PageEntry entry;
-			InitPageCheckPresence(lin_addr,true,table,entry);
+			InitPageCheckPresence((PhysPt)lin_addr,true,table,entry);
 
 			if (!table.block.a) {
 				table.block.a=1;		//Set access
-				phys_writed((paging.base.page<<12)+(lin_page >> 10)*4,table.load);
+				phys_writed((PhysPt)((paging.base.page<<12)+(lin_page >> 10)*4),table.load);
 			}
 			if (!entry.block.a) {
 				entry.block.a=1;	//Set access
@@ -1404,7 +1408,7 @@ public:
 static InitPageUserROHandler init_page_handler_userro;
 
 bool PAGING_ForcePageInit(Bitu lin_addr) {
-	PageHandler * handler=get_tlb_readhandler(lin_addr);
+	PageHandler * handler=get_tlb_readhandler((PhysPt)lin_addr);
 	if (handler==&init_page_handler) {
 		init_page_handler.InitPageForced(lin_addr);
 		return true;
@@ -1433,10 +1437,16 @@ void PAGING_SetWP(bool wp) {
 		PAGING_ClearTLB();
 }
 
+int CPU_IsDynamicCore(void);
+
 void PAGING_Enable(bool enabled) {
 	/* If paging is disabled, we work from a default paging table */
 	if (paging.enabled==enabled) return;
 	paging.enabled=enabled;
+	if (auto_determine_dynamic_core_paging) {
+		int coretype=CPU_IsDynamicCore();
+		if (coretype) use_dynamic_core_with_paging = coretype==1?enabled&&dos_kernel_disabled:enabled&&!dos_kernel_disabled;
+	}
 	if (enabled) {
 //		LOG(LOG_PAGING,LOG_NORMAL)("Enabled");
 		PAGING_SetDirBase(paging.cr3);
