@@ -379,6 +379,8 @@ static char const *fop_15[] = { "fprem", "fyl2xp1", "fsqrt", "fsincos",
                    "frndint", "fscale", "fsin", "fcos" };
 static char const *fop_21[] = { 0, "fucompp", 0, 0, 0, 0, 0, 0 };
 static char const *fop_28[] = { "[fneni]", "[fndis]", "fclex", "finit", "[fnsetpm]", "[frstpm]", 0, 0 };
+static char const* fop_29[] = { "*fucomi %GF" };
+static char const* fop_30[] = { "*fcomi %GF" };
 static char const *fop_32[] = { "*fadd %GF,st" };
 static char const *fop_33[] = { "*fmul %GF,st" };
 static char const *fop_34[] = { "*fcom %GF,st" };
@@ -403,16 +405,18 @@ static char const *fop_54[] = { "*fdivrp %GF,st" };
 static char const *fop_55[] = { "*fdivp %GF,st" };
 static char const *fop_56[] = { "*ffreep %GF" };
 static char const *fop_60[] = { "fstsw ax", 0, 0, 0, 0, 0, 0, 0 };
+static char const* fop_61[] = { "*fucomip %GF" };
+static char const* fop_62[] = { "*fcomip %GF" };
 
 static char const **fspecial[] = { /* 0=use st(i), 1=undefined 0 in fop_* means undefined */
   0, 0, 0, 0, 0, 0, 0, 0,
   fop_8, fop_9, fop_10, fop_11, fop_12, fop_13, fop_14, fop_15,
   f0, f0, f0, f0, f0, fop_21, f0, f0,
-  f0, f0, f0, f0, fop_28, f0, f0, f0,
+  f0, f0, f0, f0, fop_28, fop_29, fop_30, f0,
   fop_32, fop_33, fop_34, fop_35, fop_36, fop_37, fop_38, fop_39,
   fop_40, fop_41, fop_42, fop_43, fop_44, fop_45, f0, f0,
   fop_48, fop_49, fop_50, fop_51, fop_52, fop_53, fop_54, fop_55,
-  fop_56, f0, f0, f0, fop_60, f0, f0, f0,
+  fop_56, f0, f0, f0, fop_60, fop_61, fop_62, f0,
 };
 
 static const char *floatops[] = { /* assumed " %EF" at end of each.  mod != 3 only */
@@ -422,15 +426,15 @@ static const char *floatops[] = { /* assumed " %EF" at end of each.  mod != 3 on
        "fldenv", "fldcw", "fstenv", "fstcw",
 /*16*/ "fiadd", "fimul", "ficomw", "ficompw",
        "fisub", "fisubr", "fidiv", "fidivr",
-/*24*/ "fild", 0, "fist", "fistp",
+/*24*/ "fild", "fisttp", "fist", "fistp",
        "frstor", "fldt", 0, "fstpt",
 /*32*/ "faddq", "fmulq", "fcomq", "fcompq",
        "fsubq", "fsubrq", "fdivq", "fdivrq",
-/*40*/ "fldq", 0, "fstq", "fstpq",
+/*40*/ "fldq", "fisttpq", "fstq", "fstpq",
        0, 0, "fsave", "fstsw",
 /*48*/ "fiaddw", "fimulw", "ficomw", "ficompw",
        "fisubw", "fisubrw", "fidivw", "fidivr",
-/*56*/ "fildw", 0, "fistw", "fistpw",
+/*56*/ "fildw", "fisttpw", "fistw", "fistpw",
        "fbldt", "fildq", "fbstpt", "fistpq"
 };
 
@@ -439,9 +443,9 @@ static char *addr_to_hex(UINT32 addr, int splitup) {
 
   if (splitup) {
     if (fp_segment(addr)==0 || fp_offset(addr)==0xffff) /* 'coz of wraparound */
-      sprintf(buffer, "%04X", (unsigned)fp_offset(addr) );
+      sprintf(buffer, "%04X", fp_offset(addr) );
     else
-      sprintf(buffer, "%04X:%04X", (unsigned)fp_segment(addr), (unsigned)fp_offset(addr) );
+      sprintf(buffer, "%04X:%04X", fp_segment(addr), fp_offset(addr) );
   } else {
 #if 0
 	  /* Pet outcommented, reducing address size to 4
@@ -608,11 +612,11 @@ static void outhex(char subtype, int extend, int optional, int defsize, int sign
       } else
         signchar = '+';
       if (delta || !optional)
-		uprintf("%c%0*lX", (char)signchar, (int)(extend), (long)delta);
+		uprintf("%c%0*lX", signchar, extend, (long)delta);
     } else {
       if (extend==2)
         delta = (UINT16)delta;
-	  uprintf("%0.*lX", (int)(2*extend), (long)delta );
+	  uprintf("%0.*lX", (2*extend), (long)delta );
     }
     return;
   }
@@ -629,7 +633,7 @@ static void outhex(char subtype, int extend, int optional, int defsize, int sign
        } else
          signchar = '+';
        if (sign)
-		 uprintf("%c%02lX", (char)signchar, delta & 0xFFL);
+		 uprintf("%c%02lX", signchar, delta & 0xFFL);
        else
 		 uprintf("%02lX", delta & 0xFFL);
        break;
@@ -641,7 +645,7 @@ static void outhex(char subtype, int extend, int optional, int defsize, int sign
        } else
          signchar = '+';
        if (sign)
-		 uprintf("%c%04lX", (char)signchar, delta & 0xFFFFL);
+		 uprintf("%c%04lX", signchar, delta & 0xFFFFL);
        else
 		 uprintf("%04lX", delta & 0xFFFFL);
        break;
@@ -653,7 +657,7 @@ static void outhex(char subtype, int extend, int optional, int defsize, int sign
        } else
          signchar = '+';
        if (sign)
-		 uprintf("%c%08lX", (char)signchar, (unsigned long)delta & 0xFFFFFFFFL);
+		 uprintf("%c%08lX", signchar, (unsigned long)delta & 0xFFFFFFFFL);
        else
 		 uprintf("%08lX", (unsigned long)delta & 0xFFFFFFFFL);
        break;
@@ -873,13 +877,13 @@ static void percent(char type, char subtype)
        switch (bytes(subtype)) {              /* sizeof offset value */
        case 1:
             vofs = (INT8)getbyte();
-			name = addr_to_hex((UINT32)vofs+(UINT32)instruction_offset+(UINT32)INSTRUCTION_SIZE,0);
+			name = addr_to_hex((UINT32)vofs+instruction_offset+(UINT32)INSTRUCTION_SIZE,0);
             break;
        case 2:
             vofs  = (INT32)((UINT32)getbyte());
             vofs |= (INT32)((UINT32)getbyte() << 8);
             vofs  = (INT16)vofs;
-			name  = addr_to_hex((UINT32)vofs+(UINT32)instruction_offset+(UINT32)INSTRUCTION_SIZE,0);
+			name  = addr_to_hex((UINT32)vofs+instruction_offset+(UINT32)INSTRUCTION_SIZE,0);
             break;
 			/* i386 */
        case 4:
@@ -887,7 +891,7 @@ static void percent(char type, char subtype)
             vofs |= (INT32)((UINT32)getbyte() << 8);
             vofs |= (INT32)((UINT32)getbyte() << 16);
             vofs |= (INT32)((UINT32)getbyte() << 24);
-			name = addr_to_hex((UINT32)vofs+(UINT32)instruction_offset+(UINT32)INSTRUCTION_SIZE,(addrsize == 32)?0:1);
+			name = addr_to_hex((UINT32)vofs+instruction_offset+(UINT32)INSTRUCTION_SIZE,(addrsize == 32)?0:1);
             break;
        }
 	   if (vofs<0)
