@@ -16,12 +16,13 @@
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
-
+#include <assert.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
+#include <fstream>
 
 #if defined(WIN32)
 #include <conio.h>
@@ -36,6 +37,7 @@
 #include "menu.h"
 #include "debug.h"
 #include "debug_inc.h"
+#include "pic.h"
 
 #include <stdexcept>
 #include <exception>
@@ -44,6 +46,7 @@ using namespace std;
 
 bool log_int21 = false;
 bool log_fileio = false;
+extern bool logging_con;
 
 static bool has_LOG_Init = false;
 static bool has_LOG_EarlyInit = false;
@@ -52,6 +55,9 @@ static bool do_LOG_stderr = false;
 bool logBuffSuppressConsole = false;
 bool logBuffSuppressConsoleNeedUpdate = false;
 
+int debuggerrun = 0;
+int log_dev_con = 0;
+
 _LogGroup loggrp[LOG_MAX]={{"",LOG_NORMAL},{0,LOG_NORMAL}};
 FILE* debuglog = NULL;
 
@@ -106,6 +112,23 @@ static list<string>::iterator logBuffPos = logBuff.end();
 
 extern int old_cursor_state;
 
+void getlogtext(std::string &str) {
+    str = "";
+    std::list<string>::iterator it;
+    for (it = logBuff.begin(); it != logBuff.end(); ++it)
+        str+= (std::string)(*it)+"\n";
+}
+
+bool savetologfile(const char *name) {
+    std::ofstream out(name);
+    if (!out.is_open()) return false;
+    std::list<string>::iterator it;
+    for (it = logBuff.begin(); it != logBuff.end(); ++it)
+        out << (std::string)(*it) << endl;
+    out.close();
+    return true;
+}
+
 const char *DBGBlock::get_winname(int idx) {
     if (idx >= 0 && idx < DBGBlock::WINI_MAX_INDEX)
         return dbg_win_names[idx];
@@ -138,6 +161,15 @@ void DBGBlock::next_window(void) {
     if (order >= 0) active_win = win_order[order];
 }
 
+void DBGBlock::prev_window(void) {
+    int order = win_find_order((int)active_win);
+
+    if (order < 0) order = 0;
+
+    order = win_prev_by_order(order);
+    if (order >= 0) active_win = win_order[order];
+}
+
 void DBGBlock::swap_order(int o1,int o2) {
     if (o1 != o2)
         std::swap(win_order[o1],win_order[o2]);
@@ -411,7 +443,7 @@ static void DestroySubWindows(void) {
         WINDOW* &ref = dbg.get_win_ref((int)wnd);
 
         if (ref != NULL) {
-            if (ref) delwin(ref);
+            delwin(ref);
             ref = NULL;
         }
     }
@@ -479,7 +511,7 @@ static void MakeSubWindows(void) {
 
             /* add height to the window */
             yheight[wndi] += (unsigned int)expand_by;
-            outy += (int)expand_by;
+            outy += expand_by;
             wndi++;
 
             /* move the others down */
@@ -536,6 +568,12 @@ void DBGUI_NextWindow(void) {
 	DEBUG_DrawScreen();
 }
 
+void DBGUI_PrevWindow(void) {
+    dbg.prev_window();
+    DrawBars();
+    DEBUG_DrawScreen();
+}
+
 void DBGUI_NextWindowIfActiveHidden(void) {
     if (dbg.get_active_win() == NULL)
         DBGUI_NextWindow();
@@ -631,6 +669,7 @@ void DEBUG_ShowMsg(char const* format,...) {
 	va_list msg;
 	size_t len;
 
+    if (format==NULL || (log_dev_con == 2 && !logging_con)) return;
     in_debug_showmsg = true;
 
     // in case of runaway error from the CPU core, user responsiveness can be helpful
@@ -681,7 +720,7 @@ void DEBUG_ShowMsg(char const* format,...) {
 		logBuffPos=logBuff.end();
 		DEBUG_RefreshPage(0);
 	}
-	logBuff.push_back(buf);
+	logBuff.emplace_back(buf);
 	if (logBuff.size() > MAX_LOG_BUFFER) {
         logBuffHasDiscarded = true;
         if (logBuffPos == logBuff.begin()) ++logBuffPos; /* keep the iterator valid */
@@ -759,7 +798,7 @@ void DEBUG_ShowMsg(char const* format,...) {
 /* callback function when DOSBox-X exits */
 void LOG::Exit() {
 	if (debuglog != NULL) {
-		fprintf(debuglog,"--END OF LOG--\n");
+		if (log_dev_con != 2) fprintf(debuglog,"--END OF LOG--\n");
 		fclose(debuglog);
 		debuglog = NULL;
 	}
@@ -809,6 +848,7 @@ void LOG::ParseEnableSetting(_LogGroup &group,const char *setting) {
 		group.min_severity = LOG_NORMAL;
 }
 
+void ResolvePath(std::string& in);
 void LOG::Init() {
 	char buf[64];
 
@@ -821,19 +861,25 @@ void LOG::Init() {
 	/* announce */
 	LOG_MSG("Logging init: beginning logging proper. This is the end of the early init logging");
 
+	const Section_prop *sect = static_cast<Section_prop *>(control->GetSection("dos"));
+	assert(sect != NULL);
+    std::string logstr = sect->Get_string("log console");
+    log_dev_con = control->opt_log_con || logstr == "true" || logstr == "1" ? 1 : (logstr == "quiet" ? 2 : 0);
+
 	/* get the [log] section */
-	const Section_prop *sect = static_cast<Section_prop *>(control->GetSection("log"));
+	sect = static_cast<Section_prop *>(control->GetSection("log"));
 	assert(sect != NULL);
 
 	/* do we write to a logfile, or not? */
-	const char *blah = sect->Get_string("logfile");
-	if (blah != NULL && blah[0] != 0) {
-		if ((debuglog=fopen(blah,"wt+")) != NULL) {
-			LOG_MSG("Logging: opened logfile '%s' successfully. All further logging will go to this file.",blah);
+	std::string logfile = sect->Get_string("logfile");
+	if (logfile.size()) {
+		ResolvePath(logfile);
+		if ((debuglog=fopen(logfile.c_str(),"wt+")) != NULL) {
+			LOG_MSG("Logging: opened logfile '%s' successfully. All further logging will go to this file.",logfile.c_str());
 			setbuf(debuglog,NULL);
 		}
 		else {
-			LOG_MSG("Logging: failed to open logfile '%s'. All further logging will be discarded. Error: %s",blah,strerror(errno));
+			LOG_MSG("Logging: failed to open logfile '%s'. All further logging will be discarded. Error: %s",logfile.c_str(),strerror(errno));
 		}
 	}
 	else {
@@ -841,6 +887,14 @@ void LOG::Init() {
 		debuglog=0;
 	}
 
+	const char* debugstr = sect->Get_string("debuggerrun");
+    if (debugstr!=NULL && !strcasecmp(debugstr, "normal"))
+        debuggerrun = 1;
+    else if (debugstr!=NULL && !strcasecmp(debugstr, "watch"))
+        debuggerrun = 2;
+    else
+        debuggerrun = 0;
+
     log_int21 = sect->Get_bool("int21") || control->opt_logint21;
     log_fileio = sect->Get_bool("fileio") || control->opt_logfileio;
 
@@ -916,7 +970,7 @@ void LOG::EarlyInit(void) {
 
 void LOG::SetupConfigSection(void) {
 	const char *log_values[] = {
-		/* compatibility with existing dosbox.conf files */
+		/* compatibility with existing dosbox-x.conf files */
 		"true", "false",
 
 		/* log levels */
@@ -934,6 +988,7 @@ void LOG::SetupConfigSection(void) {
 	Prop_string* Pstring = sect->Add_string("logfile",Property::Changeable::Always,"");
 	Pstring->Set_help("file where the log messages will be saved to");
 	Pstring->SetBasic(true);
+
 	char buf[64];
 	for (Bitu i = LOG_ALL + 1;i < LOG_MAX;i++) {
 		safe_strncpy(buf,loggrp[i].front, sizeof(buf));
@@ -951,5 +1006,10 @@ void LOG::SetupConfigSection(void) {
 
     Pbool = sect->Add_bool("fileio",Property::Changeable::Always,false);
     Pbool->Set_help("Log file I/O through INT 21h");
-}
 
+	const char* debuggerrunopt[] = { "debugger", "normal", "watch", 0};
+	Pstring = sect->Add_string("debuggerrun",Property::Changeable::OnlyAtStart,"debugger");
+	Pstring->Set_help("The run mode when the DOSBox-X Debugger starts.");
+	Pstring->Set_values(debuggerrunopt);
+	Pstring->SetBasic(true);
+}
