@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -28,6 +28,9 @@
 #include <vector>
 #include <fstream>
 #include <sstream>
+#if !defined(HX_DOS) && !(defined(__MINGW32__) && !defined(__MINGW64_VERSION_MAJOR))
+#include <thread>
+#endif
 
 #include "dosbox.h"
 #include "mem.h"
@@ -35,6 +38,7 @@
 #include "SDL.h"
 #include "SDL_thread.h"
 #include "../libs/decoders/SDL_sound.h"
+#include "../libs/libchdr/chd.h"
 
 #if defined(C_SDL2) /* SDL 1.x defines this, SDL 2.x does not */
 /** @name Frames / MSF Conversion Functions
@@ -51,6 +55,7 @@
 	*(M) = value;							\
 }
 #define MSF_TO_FRAMES(M, S, F)	((M)*60*CD_FPS+(S)*CD_FPS+(F))
+#include "../../vs2015/sdl/src/cdrom/compat_SDL_cdrom.h"
 #endif /* C_SDL2 */
 
 #define RAW_SECTOR_SIZE		2352
@@ -64,11 +69,11 @@ enum { CDROM_USE_SDL, CDROM_USE_ASPI, CDROM_USE_IOCTL_DIO, CDROM_USE_IOCTL_DX, C
 //! \description CD-ROM time is represented as minutes, seconds, and frames (75 per second)
 typedef struct SMSF {
     //! \brief Time, minutes field
-    unsigned char   min;
+    unsigned char   min = 0;
     //! \brief Time, seconds field
-    unsigned char   sec;
+    unsigned char   sec = 0;
     //! \brief Time, frame field
-    unsigned char   fr;
+    unsigned char   fr = 0;
 } TMSF;
 
 //! \brief Output and channel control state
@@ -176,10 +181,8 @@ private:
     //! \brief Close the device
 	void	Close				(void);
 
-#if !defined(C_SDL2)
     //! \brief SDL 1.x CD-ROM device object
     SDL_CD* cd = NULL;
-#endif
     int driveID = 0;
     Uint32 oldLeadOut = 0;
 };
@@ -220,14 +223,16 @@ private:
 		TrackFile(uint16_t _chunkSize) : chunkSize(_chunkSize) {}
 	public:
 		virtual          ~TrackFile() = default;
-		virtual bool     read(uint8_t *buffer, int seek, int count) = 0;
-		virtual bool     seek(uint32_t offset) = 0;
+		virtual bool     read(uint8_t *buffer,int64_t seek, int count) = 0;
+		virtual bool     seek(int64_t offset) = 0;
 		virtual uint16_t   decode(uint8_t *buffer) = 0;
 		virtual uint16_t   getEndian() = 0;
 		virtual uint32_t   getRate() = 0;
 		virtual uint8_t    getChannels() = 0;
-		virtual int      getLength() = 0;
+		virtual int64_t    getLength() = 0;
+		virtual void setAudioPosition(uint32_t pos) = 0;
 		const uint16_t chunkSize = 0;
+		uint32_t audio_pos = UINT32_MAX; // last position when playing audio
 	};
 
     //! \brief Binary file reader for the image
@@ -240,13 +245,14 @@ private:
 		BinaryFile      (const BinaryFile&) = delete; // prevent copying
 		BinaryFile&     operator= (const BinaryFile&) = delete; // prevent assignment
 
-		bool            read(uint8_t *buffer, int seek, int count);
-		bool            seek(uint32_t offset);
+		bool            read(uint8_t *buffer,int64_t seek, int count);
+		bool            seek(int64_t offset);
 		uint16_t          decode(uint8_t *buffer);
 		uint16_t          getEndian();
 		uint32_t          getRate() { return 44100; }
 		uint8_t           getChannels() { return 2; }
-		int             getLength();
+		int64_t           getLength();
+		void setAudioPosition(uint32_t pos) { audio_pos = pos; }
 	private:
 		std::ifstream   *file;
 	};
@@ -260,19 +266,54 @@ private:
 		AudioFile       (const AudioFile&) = delete; // prevent copying
 		AudioFile&      operator= (const AudioFile&) = delete; // prevent assignment
 
-		bool            read(uint8_t *buffer, int seek, int count) { (void)buffer; (void)seek; (void)count; return false; }
-		bool            seek(uint32_t offset);
+		bool            read(uint8_t *buffer,int64_t seek, int count) { (void)buffer; (void)seek; (void)count; return false; }
+		bool            seek(int64_t offset);
 		uint16_t          decode(uint8_t *buffer);
 		uint16_t          getEndian();
 		uint32_t          getRate();
 		uint8_t           getChannels();
-		int             getLength();
+		int64_t           getLength();
+        void setAudioPosition(uint32_t pos) { (void)pos;/*unused*/ }
 	private:
 		Sound_Sample    *sample = nullptr;
-		// ensure the first seek isn't cached by starting with an impossibly-large position
-		uint32_t        track_pos = (std::numeric_limits<uint32_t>::max)();
 	};
 
+    class CHDFile : public TrackFile {
+    public:
+        CHDFile(const char* filename, bool& error);
+        ~CHDFile();
+
+        CHDFile() = delete;
+        CHDFile(const CHDFile&) = delete;
+        CHDFile& operator= (const CHDFile&) = delete;
+
+        bool            read(uint8_t* buffer, int64_t seek, int count);
+        bool            seek(int64_t offset);
+        uint16_t        decode(uint8_t* buffer);
+        uint16_t        getEndian();
+        uint32_t        getRate() { return 44100; }
+        uint8_t         getChannels() { return 2; }
+        int64_t         getLength();
+        void setAudioPosition(uint32_t pos) { audio_pos = pos; }
+        chd_file*       getChd() { return this->chd; }
+    private:
+              chd_file*   chd               = nullptr;
+        const chd_header* header            = nullptr; // chd header
+                /*
+                    TODO: cache more than one hunk
+                    or wait for https://github.com/rtissera/libchdr/issues/36
+                */
+              uint8_t*     hunk_buffer       = nullptr; // buffer to hold one hunk // size of hunks in CHD up to 1 MiB
+              uint8_t*     hunk_buffer_next  = nullptr; // index + 1 prefetch
+              int          hunk_buffer_index = -1;      // hunk index for buffer
+#if !defined(HX_DOS) && !(defined(__MINGW32__) && !defined(__MINGW64_VERSION_MAJOR))
+              std::thread* hunk_thread       = nullptr; // used for prefetch
+              bool         hunk_thread_error = true;
+#endif
+    public:
+              bool         skip_sync         = false;   // this will fail if a CHD contains 2048 and 2352 sector tracks
+     };
+
 public:
 	// Nested struct definition
 	struct Track {
@@ -354,6 +395,7 @@ private:
 
 	// Private functions for cue sheet processing
 	bool  LoadCueSheet(char *cuefile);
+	bool  LoadChdFile(char* chdfile);
 	bool  GetRealFileName(std::string& filename, std::string& pathname);
 	bool  GetCueKeyword(std::string &keyword, std::istream &in);
 	bool  GetCueFrame(int &frames, std::istream &in);
