@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -28,6 +28,7 @@
 uint8_t sdriv[260], sattr[260], fattr;
 char sname[260][LFN_NAMELENGTH+1],storect[CTBUF];
 extern int lfn_filefind_handle;
+extern bool forcelfn;
 
 struct finddata {
        uint8_t attr;
@@ -128,9 +129,9 @@ void DOS_InfoBlock::SetLocation(uint16_t segment) {
 	uint16_t sftOffset=offsetof(sDIB,firstFileTable)+0xa2;
 	sSave(sDIB,firstFileTable,RealMake(segment,sftOffset));
 	real_writed(segment,sftOffset+0x00,RealMake(segment+0x26,0));	//Next File Table
-	real_writew(segment,sftOffset+0x04,100);		//File Table supports 100 files
+	real_writew(segment,sftOffset+0x04,DOS_FILES/2);	//File Table supports DOS_FILES/2 files
 	real_writed(segment+0x26,0x00,0xffffffff);		//Last File Table
-	real_writew(segment+0x26,0x04,100);				//File Table supports 100 files
+	real_writew(segment+0x26,0x04,DOS_FILES-DOS_FILES/2);	//File Table supports DOS_FILES/2 files
 }
 
 void DOS_InfoBlock::SetFirstMCB(uint16_t _firstmcb) {
@@ -388,11 +389,10 @@ bool DOS_PSP::SetNumFiles(uint16_t fileNum) {
 	return true;
 }
 
-
 void DOS_DTA::SetupSearch(uint8_t _sdrive,uint8_t _sattr,char * pattern) {
 	unsigned int i;
 
-	if (lfn_filefind_handle<LFN_FILEFIND_NONE) {
+	if (lfn_filefind_handle<LFN_FILEFIND_NONE || forcelfn) {
 		sdriv[lfn_filefind_handle]=_sdrive;
 		sattr[lfn_filefind_handle]=_sattr;
 		for (i=0;i<LFN_NAMELENGTH;i++) {
@@ -478,7 +478,7 @@ uint8_t DOS_DTA::GetSearchDrive(void) {
 }
 
 void DOS_DTA::GetSearchParams(uint8_t & attr,char * pattern, bool lfn) {
-    if (lfn&&lfn_filefind_handle<LFN_FILEFIND_NONE) {
+    if (lfn||(uselfn&&lfn_filefind_handle<LFN_FILEFIND_NONE)) {
 		attr=sattr[lfn_filefind_handle];
         memcpy(pattern,sname[lfn_filefind_handle],LFN_NAMELENGTH);
         pattern[LFN_NAMELENGTH]=0;
