@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -18,12 +18,15 @@
 
 
 #include "dosbox.h"
+#include "logging.h"
 #include "mem.h"
+#include "jfont.h"
 #include "dos_inc.h"
 #include "callback.h"
 #include "control.h"
 #include <assert.h>
 
+extern bool gbk;
 extern int maxfcb;
 extern Bitu DOS_PRIVATE_SEGMENT_Size;
 
@@ -207,6 +210,45 @@ PhysPt DOS_Get_DPB(unsigned int dos_drive) {
     return PhysMake(dos.tables.dpb,dos_drive*dos.tables.dpb_size);
 }
 
+void SetupDBCSTable() {
+    if (enable_dbcs_tables) {
+        if (!dos.tables.dbcs) dos.tables.dbcs=RealMake(DOS_GetMemory(12,"dos.tables.dbcs"),0);
+
+        // write a valid table, or else Windows 3.1 is unhappy.
+        // Values are copied from INT 21h AX=6300h as returned by an MS-DOS 6.22 boot disk
+        if (IS_PC98_ARCH || IS_JEGA_ARCH || IS_JDOSV || dos.loaded_codepage == 932) {   // Japanese
+            mem_writew(Real2Phys(dos.tables.dbcs)+0,0x0006);
+            mem_writeb(Real2Phys(dos.tables.dbcs)+2,0x81);   // low/high DBCS pair 1
+            mem_writeb(Real2Phys(dos.tables.dbcs)+3,0x9F);
+            mem_writeb(Real2Phys(dos.tables.dbcs)+4,0xE0);   // low/high DBCS pair 2
+            mem_writeb(Real2Phys(dos.tables.dbcs)+5,0xFC);
+            mem_writed(Real2Phys(dos.tables.dbcs)+6,0);
+        } else if (IS_PDOSV || dos.loaded_codepage == 936) { // Simplified Chinese
+            mem_writew(Real2Phys(dos.tables.dbcs)+0,0x0004);
+            mem_writeb(Real2Phys(dos.tables.dbcs)+2,gbk?0x81:0xA1);   // low/high DBCS pair
+            mem_writeb(Real2Phys(dos.tables.dbcs)+3,0xFE);
+            mem_writed(Real2Phys(dos.tables.dbcs)+4,0);
+        } else if (IS_KDOSV || dos.loaded_codepage == 949) { // Korean
+            mem_writew(Real2Phys(dos.tables.dbcs)+0,0x0004);
+            mem_writeb(Real2Phys(dos.tables.dbcs)+2,0x81);   // low/high DBCS pair
+            mem_writeb(Real2Phys(dos.tables.dbcs)+3,0xFE);
+            mem_writed(Real2Phys(dos.tables.dbcs)+4,0);
+        } else if (IS_CDOSV || dos.loaded_codepage == 950) { // Traditional Chinese
+            mem_writew(Real2Phys(dos.tables.dbcs)+0,0x0004);
+            mem_writeb(Real2Phys(dos.tables.dbcs)+2,0x81);   // low/high DBCS pair
+            mem_writeb(Real2Phys(dos.tables.dbcs)+3,0xFE);
+            mem_writed(Real2Phys(dos.tables.dbcs)+4,0);
+        } else {
+            mem_writed(Real2Phys(dos.tables.dbcs),0);        //empty table
+        }
+    }
+    else {
+        dos.tables.dbcs=0;
+    }
+}
+
+uint16_t seg_win_startup_info;
+
 void DOS_SetupTables(void) {
 	uint16_t seg;uint16_t i;
 	dos.tables.tempdta=RealMake(DOS_GetMemory(4,"dos.tables.tempdta"),0);
@@ -224,41 +266,55 @@ void DOS_SetupTables(void) {
 	real_writed(DOS_CONSTRING_SEG,0x2a,0x204e4f43);
 
 	/* create a CON device driver */
+	if(IS_DOSV) {
+		seg = DOS_GetMemory(2, "device $IBMADSP");
+		real_writew(seg,0x00,0x0000);			// next ptr
+		real_writew(seg,0x02,DOS_CONDRV_SEG);	// next ptr
+		real_writew(seg,0x04,0xc000);			// attributes
+		real_writew(seg,0x06,0x0018);			// strategy routine
+		real_writew(seg,0x08,0x001e);			// interrupt routine
+		real_writed(seg,0x0a,0x4d424924);		// driver name
+		real_writed(seg,0x0e,0x50534441);		// driver name
+		real_writeb(seg,0x18,0x26);				// or word ptr es:[bx+03],0100h .. strategy
+		real_writeb(seg,0x19,0x81);
+		real_writeb(seg,0x1a,0x4f);
+		real_writeb(seg,0x1b,0x03);
+		real_writeb(seg,0x1c,0x00);
+		real_writeb(seg,0x1d,0x01);
+		real_writeb(seg,0x1e,0xcb);				// retf .. interrupt
+		// Windows Startup Information Structure
+		// Dummy data is required for the Microsoft version of Japanese Windows 3.1 in enhanced mode.
+		seg_win_startup_info = DOS_GetMemory(2, "Windows startup information");
+		real_writew(seg_win_startup_info,0x00,0x03);					// version major
+		real_writew(seg_win_startup_info,0x01,0x00);					// version minor
+		real_writed(seg_win_startup_info,0x02,0);						// next startup information
+		real_writed(seg_win_startup_info,0x06,0);						// virtual device name
+		real_writed(seg_win_startup_info,0x0a,0);						// virtual device reference data
+		real_writew(seg_win_startup_info,0x0e,0x0012);					// instance data record offset
+		real_writew(seg_win_startup_info,0x10,seg_win_startup_info);	// instance data record segment
+		real_writew(seg_win_startup_info,0x12,0x001c);					// instance data offset
+		real_writew(seg_win_startup_info,0x14,seg_win_startup_info);	// instance data segment
+		real_writew(seg_win_startup_info,0x16,0x0004);					// instance data size
+		real_writed(seg_win_startup_info,0x18,0);						// instance data table end
+		real_writed(seg_win_startup_info,0x1c,0);						// instance data
+		dos_infoblock.SetDeviceChainStart(RealMake(seg,0));
+	}
 	seg=DOS_CONDRV_SEG;
  	real_writed(seg,0x00,0xffffffff);	// next ptr
  	real_writew(seg,0x04,0x8013);		// attributes
   	real_writed(seg,0x06,0xffffffff);	// strategy routine
   	real_writed(seg,0x0a,0x204e4f43);	// driver name
   	real_writed(seg,0x0e,0x20202020);	// driver name
-	dos_infoblock.SetDeviceChainStart(RealMake(seg,0));
-   
+	if(!IS_DOSV) dos_infoblock.SetDeviceChainStart(RealMake(seg,0));
+
 	/* Create a fake Current Directory Structure */
 	seg=DOS_CDS_SEG;
 	real_writed(seg,0x00,0x005c3a43);
 	dos_infoblock.SetCurDirStruct(RealMake(seg,0));
 
+    /* Allocate DBCS DOUBLE BYTE CHARACTER SET LEAD-BYTE TABLE */
+    SetupDBCSTable();
 
-
-    /* Allocate DCBS DOUBLE BYTE CHARACTER SET LEAD-BYTE TABLE */
-    if (enable_dbcs_tables) {
-        dos.tables.dbcs=RealMake(DOS_GetMemory(12,"dos.tables.dbcs"),0);
-
-        if (IS_PC98_ARCH) {
-            // write a valid table, or else Windows 3.1 is unhappy.
-            // Values are copied from INT 21h AX=6300h as returned by an MS-DOS 6.22 boot disk
-            mem_writeb(Real2Phys(dos.tables.dbcs)+0,0x81);  // low/high DBCS pair 1
-            mem_writeb(Real2Phys(dos.tables.dbcs)+1,0x9F);
-            mem_writeb(Real2Phys(dos.tables.dbcs)+2,0xE0);  // low/high DBCS pair 2
-            mem_writeb(Real2Phys(dos.tables.dbcs)+3,0xFC);
-            mem_writed(Real2Phys(dos.tables.dbcs)+4,0);
-        }
-        else {
-            mem_writed(Real2Phys(dos.tables.dbcs),0); //empty table
-        }
-    }
-    else {
-        dos.tables.dbcs=0;
-    }
 	/* FILENAME CHARACTER TABLE */
 	if (enable_filenamechar) {
 		dos.tables.filenamechar=RealMake(DOS_GetMemory(2,"dos.tables.filenamechar"),0);
