@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -279,7 +279,7 @@ bool isoDrive::FindFirst(const char *dir, DOS_DTA &dta, bool fcb_findfirst) {
 
 	uint8_t attr;
 	char pattern[CROSS_LEN];
-    dta.GetSearchParams(attr, pattern, uselfn);
+    dta.GetSearchParams(attr, pattern, false);
    
 	if (attr == DOS_ATTR_VOLUME) {
 		dta.SetResult(discLabel, discLabel, 0, 0, 0, DOS_ATTR_VOLUME);
@@ -298,7 +298,7 @@ bool isoDrive::FindFirst(const char *dir, DOS_DTA &dta, bool fcb_findfirst) {
 bool isoDrive::FindNext(DOS_DTA &dta) {
 	uint8_t attr;
 	char pattern[CROSS_LEN], findName[DOS_NAMELENGTH_ASCII], lfindName[ISO_MAXPATHNAME];
-    dta.GetSearchParams(attr, pattern, uselfn);
+    dta.GetSearchParams(attr, pattern, false);
 	
 	int dirIterator = lfn_filefind_handle>=LFN_FILEFIND_MAX?dta.GetDirID():sdid[lfn_filefind_handle];
 	bool isRoot = dirIterators[dirIterator].root;
@@ -576,7 +576,7 @@ bool isoDrive :: lookup(isoDirEntry *de, const char *path) {
 	
 	char isoPath[ISO_MAXPATHNAME], longname[ISO_MAXPATHNAME];
 	safe_strncpy(isoPath, path, ISO_MAXPATHNAME);
-	strreplace(isoPath, '\\', '/');
+	strreplace_dbcs(isoPath, '\\', '/');
 	
 	// iterate over all path elements (name), and search each of them in the current de
 	for(char* name = strtok(isoPath, "/"); NULL != name; name = strtok(NULL, "/")) {
