@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -36,6 +36,7 @@ void Set_Label(char const * const input, char * const output, bool cdrom);
 class DriveManager {
 public:
 	static void AppendDisk(int drive, DOS_Drive* disk);
+	static void ChangeDisk(int drive, DOS_Drive* disk);
 	static void InitializeDrive(int drive);
 	static int UnmountDrive(int drive);
 	static char * GetDrivePosition(int drive);
@@ -74,6 +75,7 @@ public:
 	virtual bool SetFileAttr(const char * name,uint16_t attr);
 	virtual bool GetFileAttr(const char * name,uint16_t * attr);
 	virtual bool GetFileAttrEx(char* name, struct stat *status);
+	std::string GetHostName(const char * name);
 	virtual unsigned long GetCompressedSize(char* name);
 #if defined (WIN32)
 	virtual HANDLE CreateOpenFile(char const* const name);
@@ -116,13 +118,13 @@ protected:
     } srchInfo[MAX_OPENDIRS] = {};
 };
 
-#if 0 // nothing uses this
 class physfsDrive : public localDrive {
 private:
 	bool isdir(const char *dir);
+	char driveLetter;
 
 public:
-	physfsDrive(const char * startdir,uint16_t _bytes_sector,uint8_t _sectors_cluster,uint16_t _total_clusters,uint16_t _free_clusters,uint8_t _mediaid);
+	physfsDrive(const char driveLetter, const char * startdir,uint16_t _bytes_sector,uint8_t _sectors_cluster,uint16_t _total_clusters,uint16_t _free_clusters,uint8_t _mediaid, int& error, std::vector<std::string> &options);
 	virtual bool FileOpen(DOS_File * * file,const char * name,uint32_t flags);
 	virtual bool FileCreate(DOS_File * * file,const char * name,uint16_t attributes);
 	virtual bool FileUnlink(const char * name);
@@ -131,22 +133,33 @@ public:
 	virtual bool TestDir(const char * dir);
 	virtual bool FindFirst(const char * _dir,DOS_DTA & dta,bool fcb_findfirst=false);
 	virtual bool FindNext(DOS_DTA & dta);
-	virtual bool GetFileAttr(const char * name,uint16_t * attr);
 	virtual bool Rename(const char * oldname,const char * newname);
+	virtual bool SetFileAttr(const char * name,uint16_t attr);
+	virtual bool GetFileAttr(const char * name,uint16_t * attr);
+	virtual bool GetFileAttrEx(char* name, struct stat *status);
+	virtual unsigned long GetCompressedSize(char* name);
+#if defined (WIN32)
+	virtual HANDLE CreateOpenFile(char const* const name);
+	virtual unsigned long GetSerial();
+#endif
 	virtual bool AllocationInfo(uint16_t * _bytes_sector,uint8_t * _sectors_cluster,uint16_t * _total_clusters,uint16_t * _free_clusters);
 	virtual bool FileExists(const char* name);
 	virtual bool FileStat(const char* name, FileStat_Block * const stat_block);
-	virtual uint8_t GetMediaByte(void);
 	virtual bool isRemote(void);
 	virtual bool isRemovable(void);
 	virtual void *opendir(const char *dir);
 	virtual void closedir(void *handle);
-	virtual bool read_directory_first(void *handle, char* entry_name, bool& is_directory);
-	virtual bool read_directory_next(void *handle, char* entry_name, bool& is_directory);
+	virtual bool read_directory_first(void *handle, char* entry_name, char* entry_sname, bool& is_directory);
+	virtual bool read_directory_next(void *handle, char* entry_name, char* entry_sname, bool& is_directory);
 	virtual const char *GetInfo(void);
+	virtual const char *getOverlaydir(void);
+	virtual bool setOverlaydir(const char * name);
+	Bits UnMount();
 	virtual ~physfsDrive(void);
+
+protected:
+	std::string mountarc;
 };
-#endif
 
 #ifdef _MSC_VER
 #pragma pack (1)
@@ -339,26 +352,6 @@ static_assert(offsetof(direntry_lfn,LDIR_Name3) == 0x1C,"Oops");
 
 #define MAX_DIRENTS_PER_SECTOR (SECTOR_SIZE_MAX / sizeof(direntry))
 
-struct partTable {
-	uint8_t booter[446];
-	struct {
-		uint8_t bootflag;
-		uint8_t beginchs[3];
-		uint8_t parttype;
-		uint8_t endchs[3];
-		uint32_t absSectStart;
-		uint32_t partSize;
-	} pentry[4];
-	uint8_t  magic1; /* 0x55 */
-	uint8_t  magic2; /* 0xaa */
-#ifndef SECTOR_SIZE_MAX
-# pragma warning SECTOR_SIZE_MAX not defined
-#endif
-#if SECTOR_SIZE_MAX > 512
-    uint8_t  extra[SECTOR_SIZE_MAX - 512];
-#endif
-} GCC_ATTRIBUTE(packed);
-
 #ifdef _MSC_VER
 #pragma pack ()
 #endif
@@ -477,6 +470,8 @@ private:
 	uint32_t partSectSize = 0;
 	uint32_t firstDataSector = 0;
 	uint32_t firstRootDirSect = 0;
+	uint32_t physToLogAdj = 0; // Some PC-98 HDI images have larger logical than physical bytes/sector and the partition is not a multiple of it, so this is needed
+	int partition_index = -1;
 
 	uint32_t cwdDirCluster = 0;
 
@@ -537,11 +532,10 @@ private:
 	uint8_t subUnit;	char driveLetter;
 };
 
-#if 0 // nothing uses this
 class physfscdromDrive : public physfsDrive
 {
 public:
-	physfscdromDrive(const char driveLetter, const char * startdir,uint16_t _bytes_sector,uint8_t _sectors_cluster,uint16_t _total_clusters,uint16_t _free_clusters,uint8_t _mediaid, int& error);
+	physfscdromDrive(const char driveLetter, const char * startdir,uint16_t _bytes_sector,uint8_t _sectors_cluster,uint16_t _total_clusters,uint16_t _free_clusters,uint8_t _mediaid, int& error, std::vector<std::string> &options);
 	virtual bool FileOpen(DOS_File * * file,const char * name,uint32_t flags);
 	virtual bool FileCreate(DOS_File * * file,const char * name,uint16_t attributes);
 	virtual bool FileUnlink(const char * name);
@@ -559,7 +553,6 @@ private:
 	uint8_t subUnit;
 	char driveLetter;
 };
-#endif
 
 #ifdef _MSC_VER
 #pragma pack (1)
@@ -754,6 +747,7 @@ public:
 	virtual bool FileUnlink(const char * name);
 	virtual bool SetFileAttr(const char * name,uint16_t attr);
 	virtual bool GetFileAttr(const char * name,uint16_t * attr);
+	std::string GetHostName(const char * name);
 	virtual bool FileExists(const char* name);
 	virtual bool Rename(const char * oldname,const char * newname);
 	virtual bool FileStat(const char* name, FileStat_Block * const stat_block);
