@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,11 +16,14 @@
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+#include <assert.h>
+
 #include "../dos/drives.h"
 #include "control.h"
 #include "cpu.h"
 #include "render.h"
 #include "menu.h"
+#include "menudef.h"
 #include "SDL.h"
 #include "SDL_syswm.h"
 #include "bios_disk.h"
@@ -29,6 +32,9 @@
 #include "keyboard.h"
 #include "timer.h"
 #include "inout.h"
+#include "shell.h"
+#include "jfont.h"
+#include "sdlmain.h"
 
 #if DOSBOXMENU_TYPE == DOSBOXMENU_SDLDRAW
 unsigned int min_sdldraw_menu_width = 500;
@@ -73,10 +79,12 @@ DOSBoxMenu                                          mainMenu;
 
 extern const char*                                  drive_opts[][2];
 extern const char*                                  scaler_menu_opts[][2];
-extern int                                          NonUserResizeCounter;
+extern int                                          NonUserResizeCounter, msgcodepage;
 
+extern bool                                         force_conversion;
 extern bool                                         dos_kernel_disabled;
 extern bool                                         dos_shell_running_program;
+extern SHELL_Cmd                                    cmd_list[];
 
 bool                                                GFX_GetPreventFullscreen(void);
 void                                                DOSBox_ShowConsole();
@@ -102,6 +110,9 @@ static const char *def_menu__toplevel[] =
     "CaptureMenu",
 #endif
     "DriveMenu",
+#if C_DEBUG
+    "DebugMenu",
+#endif
     "HelpMenu",
     NULL
 };
@@ -111,11 +122,11 @@ static const char *def_menu_main[] =
 {
     "mapper_gui",
     "mapper_mapper",
+    "mapper_loadmap",
     "--",
     "MainSendKey",
-#if defined(WIN32) || defined(C_SDL2) || defined(LINUX) && C_X11
+    "MainHostKey",
     "SharedClipboard",
-#endif
     "--",
     "mapper_capmouse",
     "auto_lock_mouse",
@@ -130,11 +141,11 @@ static const char *def_menu_main[] =
     "--",
     "mapper_reset",
     "mapper_reboot",
-#endif
-#if !defined(C_EMSCRIPTEN)//FIXME: Shutdown causes problems with Emscripten
+    "restartinst",
+    "--",
+    "restartconf",
+    "restartlang",
     "--",
-#endif
-#if !defined(C_EMSCRIPTEN)//FIXME: Shutdown causes problems with Emscripten
     "mapper_shutdown",
 #endif
     NULL
@@ -159,12 +170,27 @@ static const char *def_menu_main_sendkey[] =
     NULL
 };
 
+/* main -> host key menu ("MenuHostKey") */
+static const char *def_menu_main_hostkey[] =
+{
+    "hostkey_ctrlalt",
+    "hostkey_ctrlshift",
+    "hostkey_altshift",
+    "--",
+    "hostkey_mapper",
+    NULL
+};
+
 /* main -> mouse wheel menu ("WheelToArrows") */
 static const char *def_menu_main_wheelarrow[] =
 {
     "wheel_updown",
     "wheel_leftright",
     "wheel_pageupdown",
+    "wheel_ctrlupdown",
+    "wheel_ctrlleftright",
+    "wheel_ctrlpageupdown",
+    "wheel_ctrlwz",
     "--",
     "wheel_none",
     "wheel_guest",
@@ -174,22 +200,22 @@ static const char *def_menu_main_wheelarrow[] =
 /* main -> shared clipboard menu ("SharedClipboard") */
 static const char *def_menu_main_clipboard[] =
 {
-#if defined(WIN32) || defined(C_SDL2)
-    "clipboard_quick",
+#if defined(WIN32) || defined(MACOSX) || defined(C_SDL2)
+    "mapper_fastedit",
     "clipboard_right",
     "clipboard_middle",
-#endif
-#if defined(WIN32)
+    "clipboard_arrows",
     "--",
+#endif
     "clipboard_device",
     "clipboard_dosapi",
-#endif
-#if defined(WIN32) || defined(C_SDL2)
+    "clipboard_biospaste",
     "--",
+#if defined(WIN32) || defined(MACOSX) || defined(C_SDL2)
     "mapper_copyall",
 #endif
     "mapper_paste",
-    "clipboard_paste_stop",
+    "mapper_pasteend",
     NULL
 };
 
@@ -227,8 +253,8 @@ static const char *def_menu_cpu_core[] =
     "mapper_dynamic",
 #endif
 #if !defined(C_EMSCRIPTEN)//FIXME: Shutdown causes problems with Emscripten
-    "menu_simple",
-    "menu_full",
+    "mapper_simple",
+    "mapper_full",
 #endif
     NULL
 };
@@ -253,6 +279,8 @@ static const char *def_menu_cpu_type[] =
     "cputype_pentium",
     "cputype_pentium_mmx",
     "cputype_ppro_slow",
+    "--",
+    "cputype_experimental",
     NULL
 };
 
@@ -301,20 +329,30 @@ static const char *def_menu_video_scaler[] =
 static const char *def_menu_video_output[] =
 {
     "output_surface",
-#if !defined(C_SDL2) && !defined(HX_DOS)
-# if (HAVE_D3D9_H) && defined(WIN32)
+#if (HAVE_D3D9_H) && defined(WIN32) && !defined(HX_DOS)
     "output_direct3d",
-# endif
 #endif
 #if defined(C_OPENGL) && !defined(HX_DOS)
     "output_opengl",
     "output_openglnb",
+    "output_openglpp",
+#endif
+#if defined(USE_TTF)
+    "output_ttf",
 #endif
     "--",
     "doublescan",
     NULL
 };
 
+/* video 3dfx menu ("Video3dfxMenu") */
+static const char *def_menu_video_3dfx[] =
+{
+    "3dfx_voodoo",
+    "3dfx_glide",
+    NULL
+};
+
 /* video text-mode menu ("VideoTextmodeMenu") */
 static const char *def_menu_video_textmode[] =
 {
@@ -325,6 +363,7 @@ static const char *def_menu_video_textmode[] =
     "text_blinking",
     "--",
     "line_80x25",
+    "line_80x43",
     "line_80x50",
     "line_80x60",
     "line_132x25",
@@ -334,6 +373,35 @@ static const char *def_menu_video_textmode[] =
     NULL
 };
 
+#if defined(USE_TTF)
+/* video TTF menu ("VideoTTFMenu") */
+static const char *def_menu_video_ttf[] =
+{
+    "mapper_incsize",
+    "mapper_decsize",
+    "--",
+    "ttf_showbold",
+    "ttf_showital",
+    "ttf_showline",
+    "ttf_showsout",
+    "--",
+    "ttf_wpno",
+    "ttf_wpwp",
+    "ttf_wpws",
+    "ttf_wpxy",
+    "ttf_wpfe",
+    "--",
+    "ttf_blinkc",
+#if C_PRINTER
+    "ttf_printfont",
+#endif
+    "ttf_dbcs_sbcs",
+    "ttf_autoboxdraw",
+    "ttf_halfwidthkana",
+    NULL
+};
+#endif
+
 /* video vsync menu ("VideoVsyncMenu") */
 static const char *def_menu_video_vsync[] =
 {
@@ -412,6 +480,7 @@ static const char *def_menu_video[] =
     "mapper_resetsize",
 #endif
     "--",
+    "refresh_rate",
     "scaler_forced",
     "VideoScalerMenu",
     "VideoOutputMenu",
@@ -422,10 +491,22 @@ static const char *def_menu_video[] =
     "VideoOverscanMenu",
     "VideoFrameskipMenu",
     "VideoTextmodeMenu",
+#if defined(USE_TTF)
+    "VideoTTFMenu",
+#endif
     "VideoPC98Menu",
-#ifdef C_D3DSHADERS
+    "Video3dfxMenu",
+#if defined(C_D3DSHADERS) || defined(C_OPENGL)
     "--",
+#endif
+#ifdef C_D3DSHADERS
     "load_d3d_shader",
+#endif
+#ifdef C_OPENGL
+    "load_glsl_shader",
+#endif
+#ifdef USE_TTF
+    "load_ttf_font",
 #endif
     NULL
 };
@@ -433,31 +514,35 @@ static const char *def_menu_video[] =
 /* DOS menu ("DOSMenu") */
 static const char *def_menu_dos[] =
 {
-    "DOSMouseMenu",
-    "--",
-    "DOSVerMenu",
-    "DOSLFNMenu",
-    "DOSEMSMenu",
-    "--",
-#if defined(WIN32) && !defined(HX_DOS)
-    "DOSWinMenu",
-#endif
-    "shell_config_commands",
 #if !defined(HX_DOS)
     "mapper_quickrun",
 #endif
+    "DOSVerMenu",
+    "DOSLFNMenu",
     "--",
+    "DOSMouseMenu",
+    "DOSEMSMenu",
+#if defined(WIN32) && !defined(HX_DOS) || defined(LINUX) || defined(MACOSX)
+    "DOSWinMenu",
+#endif
+    "--",
+    "enable_a20gate",
     "quick_reboot",
+    "sync_host_datetime",
+    "shell_config_commands",
     "--",
     "mapper_swapimg",
     "mapper_swapcd",
+    "change_currentfd",
+    "change_currentcd",
     "--",
-    "mapper_rescanall",
+    "make_diskimage",
     "list_drivenum",
-#if C_PRINTER || C_DEBUG
+    "list_ideinfo",
+    "mapper_rescanall",
     "--",
-#endif
 #if C_PRINTER
+    "print_textscreen",
     "mapper_ejectpage",
 #endif
     NULL
@@ -505,11 +590,12 @@ static const char *def_menu_dos_ems[] =
     NULL
 };
 
-#if defined(WIN32) && !defined(HX_DOS)
+#if defined(WIN32) && !defined(HX_DOS) || defined(LINUX) || defined(MACOSX)
 /* DOS WIN menu ("DOSWinMenu") */
 static const char *def_menu_dos_win[] =
 {
     "dos_win_autorun",
+    "dos_win_transpath",
     "dos_win_wait",
     "dos_win_quiet",
     NULL
@@ -557,8 +643,9 @@ static const char *def_menu_capture[] =
     "mapper_savestate",
     "mapper_loadstate",
     "saveslotmenu",
+    "autosavecfg",
     "browsesavefile",
-    "showstate",
+    "mapper_showstate",
     NULL
 };
 
@@ -577,6 +664,7 @@ static const char *def_menu_capture_format[] =
 /* Save/load options */
 static const char *save_load_options[] =
 {
+    "enable_autosave",
     "noremark_savestate",
     "force_loadstate",
     "usesavefile",
@@ -604,6 +692,7 @@ static const char *def_save_slots[] =
     "slot8",
     "slot9",
     "--",
+    "lastautosaveslot",
     "mapper_prevslot",
     "mapper_nextslot",
     "--",
@@ -647,51 +736,83 @@ static const char *def_menu_drive[] =
     "DriveX",
     "DriveY",
     "DriveZ",
-
     NULL
 };
 
-/* help output debug ("HelpDebugMenu") */
-static const char *def_menu_help_debug[] =
+/* help DOS commands ("HelpCommandMenu") */
+#define MENU_HELP_COMMAND_MAX 512
+static const char *def_menu_help_command[MENU_HELP_COMMAND_MAX];
+char help_command_temp[MENU_HELP_COMMAND_MAX][30];
+
+/* help debug ("DebugMenu" or "HelpDebugMenu") */
+#if C_DEBUG
+static const char* def_menu_debug[] =
 {
-#if C_DEBUG
     "mapper_debugger",
-#endif
-#if !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
-    "show_console",
-    "wait_on_error",
-#endif
-#if C_DEBUG
+    "--",
+    "debugger_rundebug",
+    "debugger_runnormal",
+    "debugger_runwatch",
     "--",
     "debug_blankrefreshtest",
+    "debug_generatenmi",
+    "debug_int2fhook",
     "debug_pageflip",
     "debug_retracepoll",
     "--",
+    "show_codetext",
+    "show_logtext",
+    "save_logas",
+    "--",
+    "show_console",
+    "wait_on_error",
+    "--",
     "debug_logint21",
     "debug_logfileio",
-#endif
     NULL
 };
+#elif !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
+static const char* def_menu_help_debug[] =
+{
+    "show_console",
+    "wait_on_error",
+    NULL
+};
+#endif
 
 /* help menu ("HelpMenu") */
 static const char *def_menu_help[] =
 {
     "help_intro",
+    "HelpCommandMenu",
 #if !defined(HX_DOS)
     "--",
     "help_homepage",
     "help_wiki",
     "help_issue",
 #endif
-#if C_DEBUG || !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
     "--",
+#if C_PCAP
+    "help_nic",
+#endif
+#if C_PRINTER && defined(WIN32)
+    "help_prt",
+#endif
+#if !C_DEBUG && !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
     "HelpDebugMenu",
 #endif
+#if C_PCAP || C_PRINTER && defined(WIN32) || !C_DEBUG && !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
     "--",
+#endif
     "help_about",
     NULL
 };
 
+extern bool is_paused;
+void DOSBox_SetSysMenu(void);
+#if defined(USE_TTF)
+void resetFontSize();
+#endif
 bool DOSBox_isMenuVisible(void) {
     return menu.toggle;
 }
@@ -876,6 +997,10 @@ void DOSBoxMenu::dump_log_debug(void) {
     dump_log_displaylist(display_list, 1);
 }
 
+std::vector<DOSBoxMenu::item> DOSBoxMenu::get_master_list(void) {
+    return master_list;
+}
+
 void DOSBoxMenu::clear_all_menu_items(void) {
     for (auto &id : master_list) {
         if (id.is_allocated())
@@ -1075,6 +1200,42 @@ void* DOSBoxMenu::getNsMenu(void) const {
 }
 #endif
 
+#if defined(WIN32) && !defined(HX_DOS)
+bool isSupportedCP(int newCP);
+LPWSTR getWString(std::string str, wchar_t *def, wchar_t*& buffer) {
+    LPWSTR ret = def;
+    int reqsize = 0, cp = dos.loaded_codepage;
+    Section_prop *section = static_cast<Section_prop *>(control->GetSection("config"));
+    if ((!dos.loaded_codepage || dos_kernel_disabled || force_conversion) && section!=NULL && !control->opt_noconfig) {
+        char *countrystr = (char *)section->Get_string("country"), *r=strchr(countrystr, ',');
+        if (r!=NULL && *(r+1)) {
+            cp = atoi(trim(r+1));
+            if ((cp<1 || !isSupportedCP(cp)) && msgcodepage>0) cp = msgcodepage;
+        } else if (msgcodepage>0)
+            cp = msgcodepage;
+        if (cp<1 || !isSupportedCP(cp)) {
+            if (IS_PC98_ARCH || IS_JEGA_ARCH || IS_JDOSV) cp = 932;
+            else if (IS_PDOSV) cp = 936;
+            else if (IS_KDOSV) cp = 949;
+            else if (IS_CDOSV) cp = 950;
+        }
+    }
+    uint16_t len=(uint16_t)str.size();
+    if (cp>0) {
+        if (cp==808) cp=866;
+        else if (cp==872) cp=855;
+        reqsize = MultiByteToWideChar(cp, 0, str.c_str(), len+1, NULL, 0);
+        buffer = new wchar_t[reqsize];
+        if (reqsize>0 && MultiByteToWideChar(cp, 0, str.c_str(), len+1, buffer, reqsize)==reqsize) ret = (LPWSTR)buffer;
+    } else {
+        buffer = new wchar_t[len+1];
+        mbstowcs(buffer, str.c_str(), len+1);
+        ret = (LPWSTR)buffer;
+    }
+    return ret;
+}
+#endif
+
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU /* Windows menu handle */
 std::string DOSBoxMenu::item::winConstructMenuText(void) {
     std::string r;
@@ -1112,6 +1273,7 @@ std::string DOSBoxMenu::item::winConstructMenuText(void) {
 }
 
 void DOSBoxMenu::item::winAppendMenu(HMENU handle) {
+    wchar_t* buffer = NULL;
     if (type == separator_type_id) {
         AppendMenu(handle, MF_SEPARATOR, 0, NULL);
     }
@@ -1119,8 +1281,13 @@ void DOSBoxMenu::item::winAppendMenu(HMENU handle) {
         AppendMenu(handle, MF_MENUBREAK, 0, NULL);
     }
     else if (type == submenu_type_id) {
-        if (winMenu != NULL)
-            AppendMenu(handle, MF_POPUP | MF_STRING, (uintptr_t)winMenu, winConstructMenuText().c_str());
+        if (winMenu != NULL) {
+            LPWSTR str = getWString(winConstructMenuText(), L"", buffer);
+            if (wcscmp(str, L""))
+                AppendMenuW(handle, MF_POPUP | MF_STRING, (uintptr_t)winMenu, str);
+            else
+                AppendMenu(handle, MF_POPUP | MF_STRING, (uintptr_t)winMenu, winConstructMenuText().c_str());
+        }
     }
     else if (type == item_type_id) {
         unsigned int attr = MF_STRING;
@@ -1128,8 +1295,13 @@ void DOSBoxMenu::item::winAppendMenu(HMENU handle) {
         attr |= (status.checked) ? MF_CHECKED : MF_UNCHECKED;
         attr |= (status.enabled) ? MF_ENABLED : (MF_DISABLED | MF_GRAYED);
 
-        AppendMenu(handle, attr, (uintptr_t)(master_id + winMenuMinimumID), winConstructMenuText().c_str());
+        LPWSTR str = getWString(winConstructMenuText(), L"", buffer);
+        if (wcscmp(str, L""))
+            AppendMenuW(handle, attr, (uintptr_t)(master_id + winMenuMinimumID), str);
+        else
+            AppendMenu(handle, attr, (uintptr_t)(master_id + winMenuMinimumID), winConstructMenuText().c_str());
     }
+    if (buffer != NULL) {delete[] buffer;buffer = NULL;}
 }
 
 bool DOSBoxMenu::winMenuSubInit(DOSBoxMenu::item &p_item) {
@@ -1234,9 +1406,12 @@ void DOSBoxMenu::item::refresh_item(DOSBoxMenu &menu) {
                 attr |= (status.checked) ? MF_CHECKED : MF_UNCHECKED;
                 attr |= (status.enabled) ? MF_ENABLED : (MF_DISABLED | MF_GRAYED);
 
-                ModifyMenu(phandle, (uintptr_t)(master_id + winMenuMinimumID),
-                    attr | MF_BYCOMMAND, (uintptr_t)(master_id + winMenuMinimumID),
-                    winConstructMenuText().c_str());
+                wchar_t* buffer = NULL;
+                LPWSTR str = getWString(winConstructMenuText(), L"", buffer);
+                if (wcscmp(str, L""))
+                    ModifyMenuW(phandle, (uintptr_t)(master_id + winMenuMinimumID), attr | MF_BYCOMMAND, (uintptr_t)(master_id + winMenuMinimumID), str);
+                else
+                    ModifyMenu(phandle, (uintptr_t)(master_id + winMenuMinimumID), attr | MF_BYCOMMAND, (uintptr_t)(master_id + winMenuMinimumID), winConstructMenuText().c_str());
             }
         }
     }
@@ -1331,6 +1506,9 @@ void ConstructMenu(void) {
     /* main sendkey menu */
     ConstructSubMenu(mainMenu.get_item("MainSendKey").get_master_id(), def_menu_main_sendkey);
 
+    /* main hostkey menu */
+    ConstructSubMenu(mainMenu.get_item("MainHostKey").get_master_id(), def_menu_main_hostkey);
+
     /* main mouse wheel movements menu */
     ConstructSubMenu(mainMenu.get_item("WheelToArrow").get_master_id(), def_menu_main_wheelarrow);
 
@@ -1387,6 +1565,11 @@ void ConstructMenu(void) {
     /* video text-mode menu */
     ConstructSubMenu(mainMenu.get_item("VideoTextmodeMenu").get_master_id(), def_menu_video_textmode);
 
+#if defined(USE_TTF)
+    /* video TTF menu */
+    ConstructSubMenu(mainMenu.get_item("VideoTTFMenu").get_master_id(), def_menu_video_ttf);
+#endif
+
     /* video vsync menu */
     ConstructSubMenu(mainMenu.get_item("VideoVsyncMenu").get_master_id(), def_menu_video_vsync);
 
@@ -1396,6 +1579,9 @@ void ConstructMenu(void) {
     /* video PC-98 menu */
     ConstructSubMenu(mainMenu.get_item("VideoPC98Menu").get_master_id(), def_menu_video_pc98);
 
+    /* video 3dfx menu */
+    ConstructSubMenu(mainMenu.get_item("Video3dfxMenu").get_master_id(), def_menu_video_3dfx);
+
     /* sound menu */
     ConstructSubMenu(mainMenu.get_item("SoundMenu").get_master_id(), def_menu_sound);
 
@@ -1414,7 +1600,7 @@ void ConstructMenu(void) {
     /* DOS EMS menu */
     ConstructSubMenu(mainMenu.get_item("DOSEMSMenu").get_master_id(), def_menu_dos_ems);
 
-#if defined(WIN32) && !defined(HX_DOS)
+#if defined(WIN32) && !defined(HX_DOS) || defined(LINUX) || defined(MACOSX)
     /* DOS WIN menu */
     ConstructSubMenu(mainMenu.get_item("DOSWinMenu").get_master_id(), def_menu_dos_win);
 #endif
@@ -1451,7 +1637,53 @@ void ConstructMenu(void) {
     /* help menu */
     ConstructSubMenu(mainMenu.get_item("HelpMenu").get_master_id(), def_menu_help);
 
-#if C_DEBUG || !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
+    uint32_t i=0, cmd_index=0;
+    while (cmd_list[cmd_index].name) {
+        if (!cmd_list[cmd_index].flags) {
+            strcpy(help_command_temp[i], ("command_"+std::string(cmd_list[cmd_index].name)).c_str());
+            def_menu_help_command[i] = help_command_temp[i];
+            i++;
+
+#if DOSBOXMENU_TYPE == DOSBOXMENU_SDLDRAW
+            if ((i % 15) == 14) {
+                strcpy(help_command_temp[i], "||");
+                def_menu_help_command[i]=help_command_temp[i];
+                i++;
+            }
+#endif
+        }
+        cmd_index++;
+    }
+    strcpy(help_command_temp[i], "--");
+    def_menu_help_command[i]=help_command_temp[i];
+    i++;
+    cmd_index=0;
+    while (cmd_list[cmd_index].name) {
+        if (cmd_list[cmd_index].flags && strcmp(cmd_list[cmd_index].name, "CHDIR") && strcmp(cmd_list[cmd_index].name, "ERASE") && strcmp(cmd_list[cmd_index].name, "LOADHIGH") && strcmp(cmd_list[cmd_index].name, "MKDIR") && strcmp(cmd_list[cmd_index].name, "RMDIR") && strcmp(cmd_list[cmd_index].name, "RENAME") && strcmp(cmd_list[cmd_index].name, "DX-CAPTURE") && strcmp(cmd_list[cmd_index].name, "DEBUGBOX")) {
+            strcpy(help_command_temp[i], ("command_"+std::string(cmd_list[cmd_index].name)).c_str());
+            def_menu_help_command[i] = help_command_temp[i];
+            i++;
+
+#if DOSBOXMENU_TYPE == DOSBOXMENU_SDLDRAW
+            if ((i % 15) == 14) {
+                strcpy(help_command_temp[i], "||");
+                def_menu_help_command[i]=help_command_temp[i];
+                i++;
+            }
+#endif
+        }
+        cmd_index++;
+    }
+    def_menu_help_command[i++]=NULL;
+    assert(i <= MENU_HELP_COMMAND_MAX);
+
+    /* help DOS command menu */
+    ConstructSubMenu(mainMenu.get_item("HelpCommandMenu").get_master_id(), def_menu_help_command);
+
+#if C_DEBUG
+    /* debug menu */
+    ConstructSubMenu(mainMenu.get_item("DebugMenu").get_master_id(), def_menu_debug);
+#elif !defined(MACOSX) && !defined(LINUX) && !defined(HX_DOS) && !defined(C_EMSCRIPTEN)
     /* help debug menu */
     ConstructSubMenu(mainMenu.get_item("HelpDebugMenu").get_master_id(), def_menu_help_debug);
 #endif
@@ -1510,7 +1742,7 @@ void SetVal(const std::string& secname, const std::string& preval, const std::st
         static char name[9];
         mcb.GetFileName(name);
         if (strlen(name)) {
-            LOG_MSG("GUI: Exit %s running in DOSBox, and then try again.",name);
+            LOG_MSG("GUI: Exit %s running in DOSBox-X, and then try again.",name);
             return;
         }
     }
@@ -1528,15 +1760,15 @@ void DOSBox_SetMenu(DOSBoxMenu &altMenu) {
 #endif
 #if DOSBOXMENU_TYPE == DOSBOXMENU_NSMENU /* TODO: Move to menu.cpp DOSBox_SetMenu() and add setmenu(NULL) to DOSBox_NoMenu() @emendelson request showmenu=false */
     void sdl_hax_macosx_setmenu(void *nsMenu);
-    void menu_osx_set_menuobj(DOSBoxMenu *altMenu);
+    void menu_macosx_set_menuobj(DOSBoxMenu *altMenu);
     sdl_hax_macosx_setmenu(altMenu.getNsMenu());
-    menu_osx_set_menuobj(&altMenu);
+    menu_macosx_set_menuobj(&altMenu);
 #endif
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
     if(!menu.gui) return;
     if(!menu.toggle) return;
 
-    LOG(LOG_MISC,LOG_DEBUG)("Win32: loading and attaching custom menu resource to DOSBox's window");
+    LOG(LOG_MISC,LOG_DEBUG)("Win32: loading and attaching custom menu resource to DOSBox-X's window");
 
     NonUserResizeCounter=1;
     SDL1_hax_SetMenu(altMenu.getWinMenu());
@@ -1559,7 +1791,7 @@ void DOSBox_SetMenu(void) {
 #if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
     if(!menu.gui) return;
 
-    LOG(LOG_MISC,LOG_DEBUG)("Win32: loading and attaching menu resource to DOSBox's window");
+    LOG(LOG_MISC,LOG_DEBUG)("Win32: loading and attaching menu resource to DOSBox-X's window");
 
     menu.toggle=true;
     NonUserResizeCounter=1;
@@ -1571,10 +1803,11 @@ void DOSBox_SetMenu(void) {
     if(menu.startup) {
         RENDER_CallBack( GFX_CallBackReset );
     }
-
-    void DOSBox_SetSysMenu(void);
-    DOSBox_SetSysMenu();
 #endif
+#if defined(USE_TTF)
+    if (ttf.inUse) resetFontSize();
+#endif
+    DOSBox_SetSysMenu();
 }
 
 void DOSBox_NoMenu(void) {
@@ -1597,10 +1830,11 @@ void DOSBox_NoMenu(void) {
     SDL1_hax_SetMenu(NULL);
     mainMenu.get_item("mapper_togmenu").check(!menu.toggle).refresh_item(mainMenu);
     RENDER_CallBack( GFX_CallBackReset );
-
-    void DOSBox_SetSysMenu(void);
-    DOSBox_SetSysMenu();
 #endif
+#if defined(USE_TTF)
+    if (ttf.inUse) resetFontSize();
+#endif
+    DOSBox_SetSysMenu();
 }
 
 void ToggleMenu(bool pressed) {
@@ -1622,8 +1856,6 @@ void ToggleMenu(bool pressed) {
         menu.toggle=false;
         DOSBox_NoMenu();
     }
-
-    void DOSBox_SetSysMenu(void);
     DOSBox_SetSysMenu();
 }
 
@@ -1640,30 +1872,189 @@ void DOSBox_CheckOS(int &id, int &major, int &minor) {
 }
 #endif
 
-#if defined(WIN32)
-# if defined(HX_DOS) || !defined(C_SDL2)
-HWND GetHWND(void) {
-    SDL_SysWMinfo wmi;
-    SDL_VERSION(&wmi.version);
+void MSG_WM_COMMAND_handle(SDL_SysWMmsg &Message) {
+#if defined(WIN32) && !defined(HX_DOS)
+    bool GFX_GetPreventFullscreen(void);
+    bool MAPPER_IsRunning(void);
+    bool GUI_IsRunning(void);
 
-    if(!SDL_GetWMInfo(&wmi)) {
-        return NULL;
-    }
-    return wmi.window;
-}
-
-HWND GetSurfaceHWND(void) {
-    SDL_SysWMinfo wmi;
-    SDL_VERSION(&wmi.version);
-
-    if (!SDL_GetWMInfo(&wmi)) {
-        return NULL;
-    }
-    return wmi.child_window;
-}
-# endif
+#if defined(C_SDL2)
+    if (Message.msg.win.msg != WM_COMMAND) return;
+#else
+    if (Message.msg != WM_COMMAND) return;
 #endif
 
+    WPARAM wParam;
+#if defined(C_SDL2)
+    wParam=Message.msg.win.wParam;
+#else
+    wParam=Message.wParam;
+#endif
+    if (!MAPPER_IsRunning() && !GUI_IsRunning()) {
+        if (LOWORD(wParam) == ID_WIN_SYSMENU_MAPPER) {
+            extern void MAPPER_Run(bool pressed);
+            MAPPER_Run(false);
+        }
+        if (LOWORD(wParam) == ID_WIN_SYSMENU_CFG_GUI) {
+            extern void GUI_Run(bool pressed);
+            GUI_Run(false);
+        }
+        if (LOWORD(wParam) == ID_WIN_SYSMENU_PAUSE) {
+            extern void PauseDOSBox(bool pressed);
+            PauseDOSBox(true);
+        }
+        if (LOWORD(wParam) == ID_WIN_SYSMENU_RESETSIZE) {
+            void GUI_ResetResize(bool pressed);
+            GUI_ResetResize(true);
+        }
+#if defined(USE_TTF)
+        if (LOWORD(wParam) == ID_WIN_SYSMENU_TTFINCSIZE) {
+            extern void increaseFontSize();
+            increaseFontSize();
+        }
+        if (LOWORD(wParam) == ID_WIN_SYSMENU_TTFDECSIZE) {
+            extern void decreaseFontSize();
+            decreaseFontSize();
+        }
+#endif
+    }
+    std::string fullScreenString = std::string("desktop.fullscreen");
+    if (!menu.gui || GetSetSDLValue(1, fullScreenString, 0)) return;
+    if (!GetMenu(GetHWND())) return;
+#if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
+    if (mainMenu.mainMenuWM_COMMAND((unsigned int)LOWORD(wParam))) return;
+#endif
+#else
+    (void)Message;//UNUSED
+#endif
+}
+
+void DOSBox_SetSysMenu(void) {
+#if defined(WIN32) && !defined(HX_DOS)
+    MENUITEMINFOW mii;
+    HMENU sysmenu;
+
+    sysmenu = GetSystemMenu(GetHWND(), TRUE); // revert, so we can reapply menu items
+    sysmenu = GetSystemMenu(GetHWND(), FALSE);
+    if (sysmenu == NULL) return;
+
+    AppendMenu(sysmenu, MF_SEPARATOR, -1, "");
+
+    std::string get_mapper_shortcut(const char *name), key="", msg="";
+    wchar_t* buffer = NULL;
+
+    {
+        key="togmenu";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = (menu.toggle ? MFS_CHECKED : 0) | (GFX_GetPreventFullscreen() ? MFS_DISABLED : MFS_ENABLED);
+        mii.wID = ID_WIN_SYSMENU_TOGGLEMENU;
+        mii.dwTypeData = getWString(msg, L"Show menu bar", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+
+    {
+        key="pause";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = (is_paused ? MFS_CHECKED : 0) | MFS_ENABLED;
+        mii.wID = ID_WIN_SYSMENU_PAUSE;
+        mii.dwTypeData = getWString(msg, L"Pause emulation", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+
+    AppendMenu(sysmenu, MF_SEPARATOR, -1, "");
+
+    {
+        key="resetsize";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = MFS_ENABLED;
+        mii.wID = ID_WIN_SYSMENU_RESETSIZE;
+        mii.dwTypeData = getWString(msg, L"Reset window size", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+
+#if defined(USE_TTF)
+    {
+        key="incsize";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = TTF_using() ? MFS_ENABLED : MFS_DISABLED;
+        mii.wID = ID_WIN_SYSMENU_TTFINCSIZE;
+        mii.dwTypeData = getWString(msg, L"Increase TTF font size", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+
+    {
+        key="decsize";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = TTF_using() ? MFS_ENABLED : MFS_DISABLED;
+        mii.wID = ID_WIN_SYSMENU_TTFDECSIZE;
+        mii.dwTypeData = getWString(msg, L"Decrease TTF font size", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+#endif
+
+    AppendMenu(sysmenu, MF_SEPARATOR, -1, "");
+
+    {
+        key="gui";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = MFS_ENABLED;
+        mii.wID = ID_WIN_SYSMENU_CFG_GUI;
+        mii.dwTypeData = getWString(msg, L"Configuration tool", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+
+    {
+        key="mapper";
+        msg=mainMenu.get_item("mapper_"+key).get_text()+(get_mapper_shortcut(key.c_str()).size()?"\t"+get_mapper_shortcut(key.c_str()):"");
+        memset(&mii, 0, sizeof(mii));
+        mii.cbSize = sizeof(mii);
+        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
+        mii.fState = MFS_ENABLED;
+        mii.wID = ID_WIN_SYSMENU_MAPPER;
+        mii.dwTypeData = getWString(msg, L"Mapper editor", buffer);
+        mii.cch = (UINT)(wcslen(mii.dwTypeData)+1);
+
+        InsertMenuItemW(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
+        if (buffer != NULL) {delete[] buffer;buffer = NULL;}
+    }
+#endif
+}
 #if defined(WIN32) && !defined(C_SDL2) && !defined(HX_DOS)
 #include <shlobj.h>
 
@@ -1793,79 +2184,6 @@ void Mount_Img(char drive, std::string realpath) {
     (void)drive;
 }
 
-void DOSBox_SetSysMenu(void) {
-#if !defined(HX_DOS)
-    MENUITEMINFO mii;
-    HMENU sysmenu;
-
-    sysmenu = GetSystemMenu(GetHWND(), TRUE); // revert, so we can reapply menu items
-    sysmenu = GetSystemMenu(GetHWND(), FALSE);
-    if (sysmenu == NULL) return;
-
-    AppendMenu(sysmenu, MF_SEPARATOR, -1, "");
-
-    {
-        const char *msg = "Show menu &bar";
-
-        memset(&mii, 0, sizeof(mii));
-        mii.cbSize = sizeof(mii);
-        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
-        mii.fState = (menu.toggle ? MFS_CHECKED : 0) | (GFX_GetPreventFullscreen() ? MFS_DISABLED : MFS_ENABLED);
-        mii.wID = ID_WIN_SYSMENU_TOGGLEMENU;
-        mii.dwTypeData = (LPTSTR)(msg);
-        mii.cch = (UINT)(strlen(msg)+1);
-
-        InsertMenuItem(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
-    }
-
-    AppendMenu(sysmenu, MF_SEPARATOR, -1, "");
-
-    {
-        const char *msg = "&Pause";
-
-        memset(&mii, 0, sizeof(mii));
-        mii.cbSize = sizeof(mii);
-        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
-        mii.fState = MFS_ENABLED;
-        mii.wID = ID_WIN_SYSMENU_PAUSE;
-        mii.dwTypeData = (LPTSTR)(msg);
-        mii.cch = (UINT)(strlen(msg) + 1);
-
-        InsertMenuItem(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
-    }
-
-    AppendMenu(sysmenu, MF_SEPARATOR, -1, "");
-
-    {
-        const char *msg = "Show &mapper interface";
-
-        memset(&mii, 0, sizeof(mii));
-        mii.cbSize = sizeof(mii);
-        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
-        mii.fState = MFS_ENABLED;
-        mii.wID = ID_WIN_SYSMENU_MAPPER;
-        mii.dwTypeData = (LPTSTR)(msg);
-        mii.cch = (UINT)(strlen(msg) + 1);
-
-        InsertMenuItem(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
-    }
-
-    {
-        const char *msg = "Show configuration &GUI";
-
-        memset(&mii, 0, sizeof(mii));
-        mii.cbSize = sizeof(mii);
-        mii.fMask = MIIM_ID | MIIM_STRING | MIIM_STATE;
-        mii.fState = MFS_ENABLED;
-        mii.wID = ID_WIN_SYSMENU_CFG_GUI;
-        mii.dwTypeData = (LPTSTR)(msg);
-        mii.cch = (UINT)(strlen(msg) + 1);
-
-        InsertMenuItem(sysmenu, GetMenuItemCount(sysmenu), TRUE, &mii);
-    }
-#endif
-}
-
 void DOSBox_CheckOS(int &id, int &major, int &minor) {
     OSVERSIONINFO osi;
     ZeroMemory(&osi, sizeof(OSVERSIONINFO));
@@ -1905,12 +2223,12 @@ void DOSBox_RefreshMenu(void) {
         DrawMenuBar(GetHWND());
         return;
     }
-    DOSBox_SetSysMenu();
     if(menu.toggle)
         DOSBox_SetMenu();
     else
         DOSBox_NoMenu();
 #endif
+    DOSBox_SetSysMenu();
 }
 
 void DOSBox_RefreshMenu2(void) {
@@ -1936,10 +2254,8 @@ void DOSBox_RefreshMenu2(void) {
         NonUserResizeCounter=1;
         SDL1_hax_SetMenu(NULL);
     }
-
-    void DOSBox_SetSysMenu(void);
-    DOSBox_SetSysMenu();
 #endif
+    DOSBox_SetSysMenu();
 }
 
 void MENU_Check_Drive(HMENU handle, int cdrom, int floppy, int local, int image, int automount, int umount, char drive) {
@@ -1976,40 +2292,6 @@ void reflectmenu_INITMENU_cb() {
                 the menus before the menu is displayed. */
     Reflect_Menu();
 }
-
-void MSG_WM_COMMAND_handle(SDL_SysWMmsg &Message) {
-    bool GFX_GetPreventFullscreen(void);
-
-    if (Message.msg != WM_COMMAND) return;
-#if defined(WIN32) && !defined(HX_DOS)
-    bool MAPPER_IsRunning(void);
-    bool GUI_IsRunning(void);
-
-    if (!MAPPER_IsRunning() && !GUI_IsRunning()) {
-        if (LOWORD(Message.wParam) == ID_WIN_SYSMENU_MAPPER) {
-            extern void MAPPER_Run(bool pressed);
-            MAPPER_Run(false);
-        }
-        if (LOWORD(Message.wParam) == ID_WIN_SYSMENU_CFG_GUI) {
-            extern void GUI_Run(bool pressed);
-            GUI_Run(false);
-        }
-        if (LOWORD(Message.wParam) == ID_WIN_SYSMENU_PAUSE) {
-            extern void PauseDOSBox(bool pressed);
-            PauseDOSBox(true);
-        }
-    }
-#endif
-    std::string fullScreenString = std::string("desktop.fullscreen");
-    if (!menu.gui || GetSetSDLValue(1, fullScreenString, 0)) return;
-    if (!GetMenu(GetHWND())) return;
-#if DOSBOXMENU_TYPE == DOSBOXMENU_HMENU
-    if (mainMenu.mainMenuWM_COMMAND((unsigned int)LOWORD(Message.wParam))) return;
-#endif
-}
-#else
-void DOSBox_SetSysMenu(void) {
-}
 #endif
 
 #if DOSBOXMENU_TYPE == DOSBOXMENU_SDLDRAW
