@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,12 +16,13 @@
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
-
+#include <cassert>
 #include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include "dosbox.h"
+#include "logging.h"
 #include "mem.h"
 #include "vga.h"
 #include "paging.h"
@@ -29,9 +30,11 @@
 #include "inout.h"
 #include "setup.h"
 #include "cpu.h"
+#include "control.h"
 #include "pc98_cg.h"
 #include "pc98_gdc.h"
 #include "zipfile.h"
+#include "src/ints/int10.h"
 
 unsigned char pc98_pegc_mmio[0x200] = {0}; /* PC-98 memory-mapped PEGC registers at E0000h */
 uint32_t pc98_pegc_banks[2] = {0x0000,0x0000}; /* bank switching offsets */
@@ -42,6 +45,9 @@ extern bool vga_ignore_extended_memory_bit;
 extern bool enable_pc98_256color_planar;
 extern bool enable_pc98_256color;
 
+extern unsigned int vbe_window_granularity;
+extern unsigned int vbe_window_size;
+
 #ifndef C_VGARAM_CHECKED
 #define C_VGARAM_CHECKED 1
 #endif
@@ -59,7 +65,9 @@ extern bool enable_pc98_256color;
 #define CHECKED3(v) ((v)&vga.mem.memmask)
 #define CHECKED4(v) ((v)&(vga.mem.memmask>>2))
 
-#define TANDY_VIDBASE(_X_)  &MemBase[ 0x80000 + (_X_)]
+uint32_t tandy_128kbase = 0x80000;
+
+#define TANDY_VIDBASE(_X_)  &MemBase[ tandy_128kbase + (_X_)]
 
 /* how much delay to add to VGA memory I/O in nanoseconds */
 int vga_memio_delay_ns = 1000;
@@ -1776,15 +1784,18 @@ public:
 
     /* byte-wise */
 	uint8_t readb(PhysPt addr) {
+        VGAMEM_USEC_read_delay();
         return readc<uint8_t>( PAGING_GetPhysicalAddress(addr) );
     }
 	void writeb(PhysPt addr,uint8_t val) {
+        VGAMEM_USEC_write_delay();
         writec<uint8_t>( PAGING_GetPhysicalAddress(addr), val );
     }
 
     /* word-wise.
      * in the style of the 8086, non-word-aligned I/O is split into byte I/O */
 	uint16_t readw(PhysPt addr) {
+        VGAMEM_USEC_read_delay();
         addr = PAGING_GetPhysicalAddress(addr);
         if (!(addr & 1)) /* if WORD aligned */
             return readc<uint16_t>(addr);
@@ -1794,6 +1805,7 @@ public:
         }
     }
 	void writew(PhysPt addr,uint16_t val) {
+        VGAMEM_USEC_write_delay();
         addr = PAGING_GetPhysicalAddress(addr);
         if (!(addr & 1)) /* if WORD aligned */
             writec<uint16_t>(addr,val);
@@ -1935,42 +1947,12 @@ public:
 
 class VGA_AMS_Handler : public PageHandler {
 public:
-	template< bool wrapping>
 	void writeHandler(PhysPt start, uint8_t val) {
+        assert(start < vga.mem.memsize);
 		vga.tandy.mem_base[ start ] = val;
-#ifdef DIJDIJD
-		uint32_t data=ModeOperation(val);
-		/* Update video memory and the pixel buffer */
-		VGA_Latch pixels;
-		pixels.d=((uint32_t*)vga.mem.linear)[start];
-		pixels.d&=vga.config.full_not_map_mask;
-		pixels.d|=(data & vga.config.full_map_mask);
-		((uint32_t*)vga.mem.linear)[start]=pixels.d;
-		uint8_t * write_pixels=&vga.mem.linear[VGA_CACHE_OFFSET+(start<<3)];
-
-		uint32_t colors0_3, colors4_7;
-		VGA_Latch temp;temp.d=(pixels.d>>4) & 0x0f0f0f0f;
-			colors0_3 = 
-			Expand16Table[0][temp.b[0]] |
-			Expand16Table[1][temp.b[1]] |
-			Expand16Table[2][temp.b[2]] |
-			Expand16Table[3][temp.b[3]];
-		*(uint32_t *)write_pixels=colors0_3;
-		temp.d=pixels.d & 0x0f0f0f0f;
-		colors4_7 = 
-			Expand16Table[0][temp.b[0]] |
-			Expand16Table[1][temp.b[1]] |
-			Expand16Table[2][temp.b[2]] |
-			Expand16Table[3][temp.b[3]];
-		*(uint32_t *)(write_pixels+4)=colors4_7;
-		if (wrapping && GCC_UNLIKELY( start < 512)) {
-			*(uint32_t *)(write_pixels+512*1024)=colors0_3;
-			*(uint32_t *)(write_pixels+512*1024+4)=colors4_7;
-		}
-#endif
 	}
-//	template< bool wrapping>
 	uint8_t readHandler(PhysPt start) {
+        assert(start < vga.mem.memsize);
 		return vga.tandy.mem_base[ start ];
 	}
 
@@ -1990,17 +1972,17 @@ public:
 				phys_page&=0x03;
 			return ( phys_page * 4096 ) + ( addr & 0x0FFF );
 		}
-		return ( (PAGING_GetPhysicalAddress(addr) & 0xffff) - 0x8000 ) & ( 32*1024-1 );
+		return ( (PAGING_GetPhysicalAddress(addr) & 0xffff) - 0x8000 ) & ( 16*1024-1 );
 	}
 
 	void writeb(PhysPt addr,uint8_t val) {
 		VGAMEM_USEC_write_delay();
 		addr = wrAddr( addr );
 		Bitu plane = vga.mode==M_AMSTRAD ? vga.amstrad.write_plane : 0x01; // 0x0F?
-		if( plane & 0x08 ) writeHandler<false>(addr+49152,(uint8_t)(val >> 0));
-		if( plane & 0x04 ) writeHandler<false>(addr+32768,(uint8_t)(val >> 0));
-		if( plane & 0x02 ) writeHandler<false>(addr+16384,(uint8_t)(val >> 0));
-		if( plane & 0x01 ) writeHandler<false>(addr+0,(uint8_t)(val >> 0));
+		if( plane & 0x08 ) writeHandler(addr+49152,(uint8_t)(val >> 0));
+		if( plane & 0x04 ) writeHandler(addr+32768,(uint8_t)(val >> 0));
+		if( plane & 0x02 ) writeHandler(addr+16384,(uint8_t)(val >> 0));
+		if( plane & 0x01 ) writeHandler(addr+0,(uint8_t)(val >> 0));
 	}
 	void writew(PhysPt addr,uint16_t val) {
 		VGAMEM_USEC_write_delay();
@@ -2008,26 +1990,26 @@ public:
 		Bitu plane = vga.mode==M_AMSTRAD ? vga.amstrad.write_plane : 0x01; // 0x0F?
 		if( plane & 0x01 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
 		}
 		addr += 16384;
 		if( plane & 0x02 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
 		}
 		addr += 16384;
 		if( plane & 0x04 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
 		}
 		addr += 16384;
 		if( plane & 0x08 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
 		}
 
 	}
@@ -2037,34 +2019,34 @@ public:
 		Bitu plane = vga.mode==M_AMSTRAD ? vga.amstrad.write_plane : 0x01; // 0x0F?
 		if( plane & 0x01 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
-			writeHandler<false>(addr+2,(uint8_t)(val >> 16));
-			writeHandler<false>(addr+3,(uint8_t)(val >> 24));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+2,(uint8_t)(val >> 16));
+			writeHandler(addr+3,(uint8_t)(val >> 24));
 		}
 		addr += 16384;
 		if( plane & 0x02 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
-			writeHandler<false>(addr+2,(uint8_t)(val >> 16));
-			writeHandler<false>(addr+3,(uint8_t)(val >> 24));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+2,(uint8_t)(val >> 16));
+			writeHandler(addr+3,(uint8_t)(val >> 24));
 		}
 		addr += 16384;
 		if( plane & 0x04 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
-			writeHandler<false>(addr+2,(uint8_t)(val >> 16));
-			writeHandler<false>(addr+3,(uint8_t)(val >> 24));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+2,(uint8_t)(val >> 16));
+			writeHandler(addr+3,(uint8_t)(val >> 24));
 		}
 		addr += 16384;
 		if( plane & 0x08 )
 		{
-			writeHandler<false>(addr+0,(uint8_t)(val >> 0));
-			writeHandler<false>(addr+1,(uint8_t)(val >> 8));
-			writeHandler<false>(addr+2,(uint8_t)(val >> 16));
-			writeHandler<false>(addr+3,(uint8_t)(val >> 24));
+			writeHandler(addr+0,(uint8_t)(val >> 0));
+			writeHandler(addr+1,(uint8_t)(val >> 8));
+			writeHandler(addr+2,(uint8_t)(val >> 16));
+			writeHandler(addr+3,(uint8_t)(val >> 24));
 		}
 
 	}
@@ -2180,7 +2162,13 @@ void VGA_ChangedBank(void) {
 void MEM_ResetPageHandler_Unmapped(Bitu phys_page, Bitu pages);
 void MEM_ResetPageHandler_RAM(Bitu phys_page, Bitu pages);
 
+extern void DISP2_SetPageHandler(void);
 void VGA_SetupHandlers(void) {
+	/* This code inherited from DOSBox SVN confuses bank size with bank granularity.
+	 * This code is enforcing bank granularity. Bank size concerns how much memory is
+	 * mapped to A0000-BFFFF. Most cards expose a window of 64KB. Most cards also have
+	 * a bank granularity of 64KB, but some, like Paradise and Cirrus, have 64KB windows
+	 * and 4KB granularity. */
 	vga.svga.bank_read_full = vga.svga.bank_read*vga.svga.bank_size;
 	vga.svga.bank_write_full = vga.svga.bank_write*vga.svga.bank_size;
 
@@ -2204,7 +2192,7 @@ void VGA_SetupHandlers(void) {
 		goto range_done;
 	case MCH_MDA:
 	case MCH_HERC:
-		MEM_SetPageHandler( VGA_PAGE_A0, 16, &vgaph.empty );
+//		MEM_SetPageHandler( VGA_PAGE_A0, 16, &vgaph.empty );        // already done by VGA setup memory function, also interferes with "allow more than 640KB" option
 		vgapages.base=VGA_PAGE_B0;
 		/* NTS: Implemented according to [http://www.seasip.info/VintagePC/hercplus.html#regs] */
 		if (vga.herc.enable_bits & 0x2) { /* bit 1: page in upper 32KB */
@@ -2246,7 +2234,7 @@ void VGA_SetupHandlers(void) {
 		goto range_done;
 //		MEM_SetPageHandler(vga.tandy.mem_bank<<2,vga.tandy.is_32k_mode ? 0x08 : 0x04,range_handler);
 	case MCH_AMSTRAD: // Memory handler.
-		MEM_SetPageHandler( 0xb8, 8, &vgaph.ams );
+		MEM_SetPageHandler( VGA_PAGE_B8, 4, &vgaph.ams );   // 0xb8000 - 0xbbfff
 		goto range_done;
 	case EGAVGA_ARCH_CASE:
         break;
@@ -2364,7 +2352,19 @@ void VGA_SetupHandlers(void) {
                 vgapages.mask = 0xffff & vga.mem.memmask;
                 break;
 		}
-		MEM_SetPageHandler(VGA_PAGE_A0, 32, newHandler );
+		if (CurMode && CurMode->mode >= 0x14/*VESA BIOS or extended mode*/ && vbe_window_size > 0/*user override of window size*/) {
+			unsigned int pages = (vbe_window_size + 0xFFFu) >> 12u; /* bytes to pages, round up */
+			if (pages > 32) pages = 32;
+			assert(pages != 0u);
+
+			/* map only what the window size determines, make the rest empty */
+			MEM_SetPageHandler(VGA_PAGE_A0, pages, newHandler );
+			MEM_SetPageHandler(VGA_PAGE_A0 + pages, 32 - pages, &vgaph.empty );
+		}
+		else {
+			/*full 128KB */
+			MEM_SetPageHandler(VGA_PAGE_A0, 32, newHandler );
+		}
 		break;
 	case 1:
 		vgapages.base = VGA_PAGE_A0;
@@ -2393,6 +2393,9 @@ void VGA_SetupHandlers(void) {
     non_cga_ignore_oddeven_engage = (non_cga_ignore_oddeven && !(vga.mode == M_TEXT || vga.mode == M_CGA2 || vga.mode == M_CGA4));
 
 range_done:
+#if C_DEBUG
+	if (control->opt_display2) DISP2_SetPageHandler();
+#endif
 	PAGING_ClearTLB();
 }
 
@@ -2445,7 +2448,18 @@ void VGA_StartUpdateLFB(void) {
 static bool VGA_Memory_ShutDown_init = false;
 
 static void VGA_Memory_ShutDown(Section * /*sec*/) {
-	MEM_SetPageHandler(VGA_PAGE_A0,32,&vgaph.empty);
+	if (machine == MCH_CGA)
+		MEM_SetPageHandler(VGA_PAGE_B8,8,&vgaph.empty);
+	else if (machine == MCH_HERC || machine == MCH_MDA)
+		MEM_SetPageHandler(VGA_PAGE_B0,8,&vgaph.empty);
+	else
+		MEM_SetPageHandler(VGA_PAGE_A0,32,&vgaph.empty);
+
+#if C_DEBUG
+	if (control->opt_display2)
+		MEM_SetPageHandler(VGA_PAGE_B0,8,&vgaph.empty);
+#endif
+
 	PAGING_ClearTLB();
 
 	if (vga.mem.linear_orgptr != NULL) {
@@ -2490,7 +2504,14 @@ void VGA_SetupMemory() {
 
 	vga.svga.bank_read = vga.svga.bank_write = 0;
 	vga.svga.bank_read_full = vga.svga.bank_write_full = 0;
-	vga.svga.bank_size = 0x10000; /* most common bank size is 64K */
+
+	/* obey user override for "bank size", which this code inherited from DOSBox SVN
+	 * confuses with "bank granularity". If "bank size" were truly a concern it would
+	 * affect how much of the A0000-BFFFF region VGA mapping would expose. */
+	if (vbe_window_granularity > 0)
+		vga.svga.bank_size = vbe_window_granularity; /* allow different sizes for dev testing */
+	else
+		vga.svga.bank_size = 0x10000; /* most common bank size is 64K */
 
 	if (!VGA_Memory_ShutDown_init) {
 		AddExitFunction(AddExitFunctionFuncPair(VGA_Memory_ShutDown));
