@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 #include <math.h>
 #include "dosbox.h"
 #include "inout.h"
+#include "logging.h"
 #include "vga.h"
 #include "mem.h"
 #include "pic.h"
@@ -75,7 +76,10 @@ static void write_crtc_data_other(Bitu /*port*/,Bitu val,Bitu /*iolen*/) {
 		break;
 	case 0x04:		//Vertical total
 		if (vga.other.vtotal ^ val) VGA_StartResize();
-		vga.other.vtotal=(uint8_t)(val&0x7f);
+		if (machine == MCH_TANDY) // FIXME: This is needed for "Math Rabbit", but does Tandy hardware *really* use all 8 bits? Or is it limited to 7 bits like MDA/CGA/Hercules/PCjr? This is default DOSBox SVN behavior.
+			vga.other.vtotal=(uint8_t)(val&0xff);
+		else
+			vga.other.vtotal=(uint8_t)(val&0x7f);
 		break;
 	case 0x05:		//Vertical display adjust
 		if (vga.other.vadjust ^ val) VGA_StartResize();
@@ -83,7 +87,10 @@ static void write_crtc_data_other(Bitu /*port*/,Bitu val,Bitu /*iolen*/) {
 		break;
 	case 0x06:		//Vertical rows
 		if (vga.other.vdend ^ val) VGA_StartResize();
-		vga.other.vdend=(uint8_t)(val&0x7f);
+		if (machine == MCH_TANDY) // FIXME: This is needed for "Math Rabbit", but does Tandy hardware *really* use all 8 bits? Or is it limited to 7 bits like MDA/CGA/Hercules/PCjr? This is default DOSBox SVN behavior.
+			vga.other.vdend=(uint8_t)(val&0xff);
+		else
+			vga.other.vdend=(uint8_t)(val&0x7f);
 		break;
 	case 0x07:		//Vertical sync position
 		vga.other.vsyncp=(uint8_t)val;
@@ -290,8 +297,8 @@ static double hue_offset = 0.0;
 
 static uint8_t cga16_val = 0;
 static void update_cga16_color(void);
-static uint8_t herc_pal = 0;
-static uint8_t mono_cga_pal = 0;
+static MonochromeColor herc_pal = MonochromeColor::Green;
+static MonochromeColor mono_cga_pal = MonochromeColor::Green;
 static uint8_t mono_cga_bright = 0;
 static uint8_t const mono_cga_palettes[8][16][3] =
 {
@@ -343,11 +350,25 @@ static void update_cga16_color(void) {
 	double tv_brightness = 0.0; // hardcoded for simpler implementation
 	double tv_saturation = (new_cga ? 0.7 : 0.6);
 
-	bool bw = (vga.tandy.mode_control&4) != 0;
-	bool color_sel = (cga16_val&0x20) != 0;
-	bool background_i = (cga16_val&0x10) != 0;	// Really foreground intensity, but this is what the CGA schematic calls it.
-	bool bpp1 = (vga.tandy.mode_control&0x10) != 0;
-	uint8_t overscan = cga16_val&0x0f;  // aka foreground colour in 1bpp mode
+	bool bw;
+	bool color_sel;
+	bool background_i;
+	bool bpp1;
+	uint8_t overscan;
+	if (machine==MCH_PCJR) {
+	  tv_saturation = 1.0;
+	  bw = (vga.tandy.mode_control&4) != 0;
+	  color_sel = 1;
+	  background_i = 1;	// Really foreground intensity, but this is what the CGA schematic calls it.
+	  bpp1 = (vga.tandy.gfx_control & 0x08) != 0;
+	  overscan = 15;
+	} else {
+	  bw = (vga.tandy.mode_control&4) != 0;
+	  color_sel = (cga16_val&0x20) != 0;
+	  background_i = (cga16_val&0x10) != 0;	// Really foreground intensity, but this is what the CGA schematic calls it.
+	  bpp1 = (vga.tandy.mode_control&0x10) != 0;
+	  overscan = cga16_val&0x0f;  // aka foreground colour in 1bpp mode
+	}
 
 	double chroma_coefficient = new_cga ? 0.29 : 0.72;
 	double b_coefficient = new_cga ? 0.07 : 0;
@@ -387,7 +408,11 @@ static void update_cga16_color(void) {
 		double d = rgbi_coefficients[o];
 		pixel_clock_delay = (chroma_pixel_delays[o & 7]*chroma_coefficient + rgbi_pixel_delay*d)/(chroma_coefficient + d);
 	}
-	pixel_clock_delay -= 21.5*ns;  // correct for delay of color burst
+	if (machine==MCH_PCJR) {
+	  pixel_clock_delay += 60*ns;  // correct for delay of color burst
+	  if (!bpp1) pixel_clock_delay += 25*ns;  // correct for delay of color burst
+	} else
+	  pixel_clock_delay -= 21.5*ns;  // correct for delay of color burst
 
 	double hue_adjust = (-(90-33)-hue_offset+pixel_clock_delay)*tau/360.0;
 	double chroma_signals[8][4];
@@ -430,6 +455,10 @@ static void update_cga16_color(void) {
 		(Bitu)(4 + (color_sel&&!bw? 1 : 0) + (background_i ? 8 : 0)),
 		(Bitu)(6 + (color_sel||bw ? 1 : 0) + (background_i ? 8 : 0))
 	};
+	Bitu PCJRpal[4] = {
+		vga.attr.palette[0], vga.attr.palette[1],
+		vga.attr.palette[2], vga.attr.palette[3]
+	};
 	for (uint8_t x=0; x<4; x++) {	 // Position of pixel in question
 		bool even = (x & 1) == 0;
 		for (uint8_t bits=0; bits<(even ? 0x10 : 0x40); ++bits) {
@@ -439,11 +468,17 @@ static void update_cga16_color(void) {
 				uint8_t rgbi;
 				if (bpp1)
 					rgbi = ((bits >> (3-p)) & (even ? 1 : 2)) != 0 ? overscan : 0;
-				else
+				else if (machine==MCH_PCJR) {
+					if (even)
+						rgbi = PCJRpal[(bits >> (2-(p&2)))&3];
+					else
+						rgbi = PCJRpal[(bits >> (4-((p+1)&6)))&3];
+				} else {
 					if (even)
 						rgbi = (uint8_t)CGApal[(bits >> (2-(p&2)))&3];
 					else
 						rgbi = (uint8_t)CGApal[(bits >> (4-((p+1)&6)))&3];
+				}
 				uint8_t c = rgbi & 7;
 				if (bw && c != 0)
 					c = 7;
@@ -484,7 +519,7 @@ static void update_cga16_color(void) {
 	}
 }
 
-static void IncreaseHue(bool pressed) {
+void IncreaseHue(bool pressed) {
 	if (!pressed)
 		return;
 	hue_offset += 5.0;
@@ -492,7 +527,7 @@ static void IncreaseHue(bool pressed) {
 	LOG_MSG("Hue at %f",hue_offset); 
 }
 
-static void DecreaseHue(bool pressed) {
+void DecreaseHue(bool pressed) {
 	if (!pressed)
 		return;
 	hue_offset -= 5.0;
@@ -610,20 +645,25 @@ static void write_cga(Bitu port,Bitu val,Bitu /*iolen*/) {
 	}
 }
 
-static void CGAModel(bool pressed) {
+void CGAModel(bool pressed) {
 	if (!pressed) return;
 	new_cga = !new_cga;
 	update_cga16_color();
 	LOG_MSG("%s model CGA selected", new_cga ? "Late" : "Early");
 }
  
-static void Composite(bool pressed) {
+static void PCJr_FindMode(void);
+void Composite(bool pressed) {
 	if (!pressed) return;
 	if (++cga_comp>2) cga_comp=0;
 	LOG_MSG("Composite output: %s",(cga_comp==0)?"auto":((cga_comp==1)?"on":"off"));
 	// switch RGB and Composite if in graphics mode
-	if (vga.tandy.mode_control & 0x2)
-		write_cga(0x3d8,vga.tandy.mode_control,1);
+	if (vga.tandy.mode_control & 0x2) {
+	  if (machine==MCH_PCJR)
+	    PCJr_FindMode();
+	  else
+	    write_cga(0x3d8,vga.tandy.mode_control,1);
+	}
 }
 
 static void tandy_update_palette() {
@@ -673,6 +713,7 @@ static void tandy_update_palette() {
 		default:
 			break;
 		}
+		update_cga16_color();
 	}
 }
 
@@ -702,6 +743,7 @@ static void TANDY_FindMode(void) {
 }
 
 static void PCJr_FindMode(void) {
+	new_cga = 1;
 	if (vga.tandy.mode_control & 0x2) {
 		if (vga.tandy.mode_control & 0x10) {
 			/* bit4 of mode control 1 signals 16 colour graphics mode */
@@ -710,10 +752,20 @@ static void PCJr_FindMode(void) {
 		} else if (vga.tandy.gfx_control & 0x08) {
 			/* bit3 of mode control 2 signals 2 colour graphics mode */
 			VGA_SetMode(M_TANDY2);
+			if (cga_comp==1 || (cga_comp==0 && !(vga.tandy.mode_control&0x4)))
+				VGA_SetMode(M_CGA16);
+			else
+				VGA_SetMode(M_TANDY2);
 		} else {
 			/* otherwise some 4-colour graphics mode */
-			if (vga.mode==M_TANDY16) VGA_SetModeNow(M_TANDY4);
-			else VGA_SetMode(M_TANDY4);
+			if (vga.mode==M_TANDY16) {
+			  if (cga_comp==1)
+			    VGA_SetModeNow(M_CGA16);
+			  else
+			    VGA_SetModeNow(M_TANDY4);
+			} else {
+			  if (cga_comp==1) VGA_SetMode(M_CGA16); else VGA_SetMode(M_TANDY4);
+			}
 		}
 		tandy_update_palette();
 	} else {
@@ -878,39 +930,45 @@ static void write_pcjr(Bitu port,Bitu val,Bitu /*iolen*/) {
 	}
 }
 
-static void CycleHercPal(bool pressed) {
+void CycleHercPal(bool pressed) {
 	if (!pressed) return;
-	if (++herc_pal>3) herc_pal=0;
+	if (++herc_pal>MonochromeColor::Last) herc_pal=MonochromeColor::First;
 	Herc_Palette();
 }
 
-static void CycleMonoCGAPal(bool pressed) {
+void CycleMonoCGAPal(bool pressed) {
 	if (!pressed) return;
-	if (++mono_cga_pal>3) mono_cga_pal=0;
+	if (++mono_cga_pal>MonochromeColor::Last) mono_cga_pal=MonochromeColor::First;
 	Mono_CGA_Palette();
 }
 
-static void CycleMonoCGABright(bool pressed) {
+void CycleMonoCGABright(bool pressed) {
 	if (!pressed) return;
 	if (++mono_cga_bright>1) mono_cga_bright=0;
 	Mono_CGA_Palette();
 }
 	
+void HercBlend(bool pressed) {
+	if (!pressed) return;
+	vga.herc.blend = !vga.herc.blend;
+	VGA_SetupDrawing(0);
+}
+
 void Herc_Palette(void) {	
 	switch (herc_pal) {
-	case 0:	// White
+	case MonochromeColor::White:
 		VGA_DAC_SetEntry(0x7,0x2a,0x2a,0x2a);
 		VGA_DAC_SetEntry(0xf,0x3f,0x3f,0x3f);
 		break;
-	case 1:	// Amber
+	case MonochromeColor::Amber:
 		VGA_DAC_SetEntry(0x7,0x34,0x20,0x00);
 		VGA_DAC_SetEntry(0xf,0x3f,0x34,0x00);
 		break;
-	case 2:	// Paper-white
+	case MonochromeColor::Gray:
 		VGA_DAC_SetEntry(0x7,0x2c,0x2d,0x2c);
 		VGA_DAC_SetEntry(0xf,0x3f,0x3f,0x3b);
 		break;
-	case 3:	// Green
+	case MonochromeColor::Green:
 		VGA_DAC_SetEntry(0x7,0x00,0x26,0x00);
 		VGA_DAC_SetEntry(0xf,0x00,0x3f,0x00);
 		break;
@@ -919,12 +977,6 @@ void Herc_Palette(void) {
 	VGA_DAC_CombineColor(2,0xf);
 }
 
-static void HercBlend(bool pressed) {
-	if (!pressed) return;
-	vga.herc.blend = !vga.herc.blend;
-	VGA_SetupDrawing(0);
-}
-
 void Mono_CGA_Palette(void) {	
 	for (uint8_t ct=0;ct<16;ct++) {
 		VGA_DAC_SetEntry(ct,
@@ -1081,8 +1133,6 @@ void VGA_SetupOther(void) {
 	if (machine==MCH_HERC || machine==MCH_MDA) {
 		for (Bitu i=0;i<256;i++)	memcpy(&vga.draw.font[i*32],&int10_font_14[i*14],14);
 		vga.draw.font_tables[0]=vga.draw.font_tables[1]=vga.draw.font;
-		MAPPER_AddHandler(HercBlend,MK_nothing,0,"hercblend","Herc Blend");
-		MAPPER_AddHandler(CycleHercPal,MK_nothing,0,"hercpal","Herc Pal");
 	}
 	if (machine==MCH_CGA || machine==MCH_MCGA || machine==MCH_AMSTRAD) {
 		vga.amstrad.mask_plane = 0x07070707;
@@ -1103,16 +1153,6 @@ void VGA_SetupOther(void) {
 			IO_RegisterWriteHandler(0x3de,write_cga,IO_MB);
 			IO_RegisterWriteHandler(0x3df,write_cga,IO_MB);
 		}
-
-		if(!mono_cga) {
-            MAPPER_AddHandler(IncreaseHue,MK_nothing,0,"inchue","Inc Hue");
-            MAPPER_AddHandler(DecreaseHue,MK_nothing,0,"dechue","Dec Hue");
-            MAPPER_AddHandler(CGAModel,MK_nothing,0,"cgamodel","CGA Model");
-            MAPPER_AddHandler(Composite,MK_nothing,0,"cgacomp","CGA Comp");
-        } else {
-            MAPPER_AddHandler(CycleMonoCGAPal,MK_nothing,0,"monocgapal","Mono CGA Pal"); 
-            MAPPER_AddHandler(CycleMonoCGABright,MK_nothing,0,"monocgabri","Mono CGA Bright");
-        }
 	}
 	if (machine==MCH_TANDY) {
 		write_tandy( 0x3df, 0x0, 0 );
