@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,6 +16,8 @@
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+#include <assert.h>
+
 #include "dosbox.h"
 #include "callback.h"
 #include "bios.h"
@@ -133,6 +135,7 @@ void incrementFDD(void) {
         equipment|=(numofdisks<<6);
     } else equipment|=1;
     mem_writew(BIOS_CONFIGURATION,equipment);
+    if(IS_EGAVGA_ARCH) equipment &= ~0x30; //EGA/VGA startup display mode differs in CMOS
     CMOS_SetRegister(0x14, (uint8_t)(equipment&0xff));
 }
 
@@ -723,6 +726,8 @@ void IDE_ResetDiskByBIOS(unsigned char disk);
 void IDE_EmuINT13DiskReadByBIOS(unsigned char disk,unsigned int cyl,unsigned int head,unsigned sect);
 void IDE_EmuINT13DiskReadByBIOS_LBA(unsigned char disk,uint64_t lba);
 
+void diskio_delay(Bits value/*bytes*/);
+
 static Bitu INT13_DiskHandler(void) {
     uint16_t segat, bufptr;
     uint8_t sectbuf[512];
@@ -846,6 +851,8 @@ static Bitu INT13_DiskHandler(void) {
         for(i=0;i<reg_al;i++) {
             last_status = imageDiskList[drivenum]->Read_Sector((uint32_t)reg_dh, (uint32_t)(reg_ch | ((reg_cl & 0xc0)<< 2)), (uint32_t)((reg_cl & 63)+i), sectbuf);
 
+            diskio_delay(512);
+
             /* IDE emulation: simulate change of IDE state that would occur on a real machine after INT 13h */
             IDE_EmuINT13DiskReadByBIOS(reg_dl, (uint32_t)(reg_ch | ((reg_cl & 0xc0)<< 2)), (uint32_t)reg_dh, (uint32_t)((reg_cl & 63)+i));
 
@@ -893,6 +900,8 @@ static Bitu INT13_DiskHandler(void) {
                 bufptr++;
             }
 
+            diskio_delay(512);
+
             last_status = imageDiskList[drivenum]->Write_Sector((uint32_t)reg_dh, (uint32_t)(reg_ch | ((reg_cl & 0xc0) << 2)), (uint32_t)((reg_cl & 63) + i), &sectbuf[0]);
             if(last_status != 0x00) {
             CALLBACK_SCF(true);
@@ -1100,6 +1109,8 @@ static Bitu INT13_DiskHandler(void) {
         for(i=0;i<dap.num;i++) {
             last_status = imageDiskList[drivenum]->Read_AbsoluteSector(dap.sector+i, sectbuf);
 
+            diskio_delay(512);
+
             IDE_EmuINT13DiskReadByBIOS_LBA(reg_dl,dap.sector+i);
 
             if((last_status != 0x00) || (killRead)) {
@@ -1133,6 +1144,8 @@ static Bitu INT13_DiskHandler(void) {
                 bufptr++;
             }
 
+            diskio_delay(512);
+
             last_status = imageDiskList[drivenum]->Write_AbsoluteSector(dap.sector+i, &sectbuf[0]);
             if(last_status != 0x00) {
                 CALLBACK_SCF(true);
@@ -1425,7 +1438,7 @@ uint8_t imageDiskVFD::Write_AbsoluteSector(uint32_t sectnum,const void *data) {
     return Write_Sector(h,c,s,data);
 }
 
-imageDiskVFD::imageDiskVFD(FILE *imgFile, uint8_t *imgName, uint32_t imgSizeK, bool isHardDisk) : imageDisk(ID_VFD) {
+imageDiskVFD::imageDiskVFD(FILE *imgFile, const char *imgName, uint32_t imgSizeK, bool isHardDisk) : imageDisk(ID_VFD) {
     (void)isHardDisk;//UNUSED
     unsigned char tmp[16];
 
@@ -1440,7 +1453,7 @@ imageDiskVFD::imageDiskVFD(FILE *imgFile, uint8_t *imgName, uint32_t imgSizeK, b
     diskimg = imgFile;
 
     if (imgName != NULL)
-        diskname = (const char*)imgName;
+        diskname = imgName;
 
     // NOTES:
     // 
@@ -1748,7 +1761,7 @@ uint8_t imageDiskD88::Write_AbsoluteSector(uint32_t sectnum,const void *data) {
     return Write_Sector(h,c,s,data);
 }
 
-imageDiskD88::imageDiskD88(FILE *imgFile, uint8_t *imgName, uint32_t imgSizeK, bool isHardDisk) : imageDisk(ID_D88) {
+imageDiskD88::imageDiskD88(FILE *imgFile, const char *imgName, uint32_t imgSizeK, bool isHardDisk) : imageDisk(ID_D88) {
     (void)isHardDisk;//UNUSED
     D88HEAD head;
 
@@ -1769,7 +1782,7 @@ imageDiskD88::imageDiskD88(FILE *imgFile, uint8_t *imgName, uint32_t imgSizeK, b
     diskimg = imgFile;
 
     if (imgName != NULL)
-        diskname = (const char*)imgName;
+        diskname = imgName;
 
     // NOTES:
     // 
@@ -2044,7 +2057,7 @@ uint8_t imageDiskNFD::Write_AbsoluteSector(uint32_t sectnum,const void *data) {
     return Write_Sector(h,c,s,data);
 }
 
-imageDiskNFD::imageDiskNFD(FILE *imgFile, uint8_t *imgName, uint32_t imgSizeK, bool isHardDisk, unsigned int revision) : imageDisk(ID_NFD) {
+imageDiskNFD::imageDiskNFD(FILE *imgFile, const char *imgName, uint32_t imgSizeK, bool isHardDisk, unsigned int revision) : imageDisk(ID_NFD) {
     (void)isHardDisk;//UNUSED
     union {
         NFDHDR head;
@@ -2066,7 +2079,7 @@ imageDiskNFD::imageDiskNFD(FILE *imgFile, uint8_t *imgName, uint32_t imgSizeK, b
     diskimg = imgFile;
 
     if (imgName != NULL)
-        diskname = (const char*)imgName;
+        diskname = imgName;
 
     // NOTES:
     // 
@@ -2313,3 +2326,150 @@ imageDiskNFD::~imageDiskNFD() {
     }
 }
 
+bool PartitionLoadMBR(std::vector<partTable::partentry_t> &parts,imageDisk *loadedDisk) {
+	partTable smbr;
+
+	parts.clear();
+
+	if (loadedDisk->getSectSize() > sizeof(smbr)) return false;
+	if (loadedDisk->Read_Sector(0,0,1,&smbr) != 0x00) return false;
+	if (smbr.magic1 != 0x55 || smbr.magic2 != 0xaa) return false; /* Must have signature */
+
+	/* first copy the main partition table entries */
+	for (size_t i=0;i < 4;i++)
+		parts.push_back(smbr.pentry[i]);
+
+	/* then, enumerate extended partitions and add the partitions within, doing it in a way that
+	 * allows recursive extended partitions */
+	{
+		size_t i=0;
+
+		do {
+			if (parts[i].parttype == 0x05/*extended*/ || parts[i].parttype == 0x0F/*LBA extended*/) {
+				unsigned long sect = parts[i].absSectStart;
+				unsigned long send = sect + parts[i].partSize;
+
+				/* partitions within an extended partition form a sort of linked list.
+				 * first entry is the partition, sector start relative to parent partition.
+				 * second entry points to next link in the list. */
+
+				/* parts[i] is the parent partition in this loop.
+				 * this loop will add to the parts vector, the parent
+				 * loop will continue enumerating through the vector
+				 * until all have processed. in this way, extended
+				 * partitions will be expanded into the sub partitions
+				 * until none is left to do. */
+
+				/* FIXME: Extended partitions within extended partitions not yet tested,
+				 *        MS-DOS FDISK.EXE won't generate that. */
+
+				while (sect < send) {
+					smbr.magic1 = smbr.magic2 = 0;
+					loadedDisk->Read_AbsoluteSector(sect,&smbr);
+
+					if (smbr.magic1 != 0x55 || smbr.magic2 != 0xAA)
+						break;
+					if (smbr.pentry[0].absSectStart == 0 || smbr.pentry[0].partSize == 0)
+						break; // FIXME: Not sure what MS-DOS considers the end of the linked list
+
+					const size_t idx = parts.size();
+
+					/* Right, get this: absolute start sector in entry #0 is relative to this link in the linked list.
+					 * The pointer to the next link in the linked list is relative to the parent partition. Blegh. */
+					smbr.pentry[0].absSectStart += sect;
+					if (smbr.pentry[1].absSectStart != 0)
+						smbr.pentry[1].absSectStart += parts[i].absSectStart;
+
+					/* if the partition extends past the parent, then stop */
+					if ((smbr.pentry[0].absSectStart+smbr.pentry[0].partSize) > (parts[i].absSectStart+parts[i].partSize))
+						break;
+
+					parts.push_back(smbr.pentry[0]);
+
+					/* Based on MS-DOS 5.0, the 2nd entry is a link to the next entry, but only if
+					 * start sector is nonzero and type is 0x05/0x0F. I'm not certain if MS-DOS allows
+					 * the linked list to go either direction, but for the sake of preventing infinite
+					 * loops stop if the link points to a lower sector number. */
+					if (idx < 256 && (smbr.pentry[1].parttype == 0x05 || smbr.pentry[1].parttype == 0x0F) &&
+						smbr.pentry[1].absSectStart != 0 && smbr.pentry[1].absSectStart > sect) {
+						sect = smbr.pentry[1].absSectStart;
+					}
+					else {
+						break;
+					}
+				}
+			}
+			i++;
+		} while (i < parts.size());
+	}
+
+	return true;
+}
+
+bool PartitionLoadIPL1(std::vector<_PC98RawPartition> &parts,imageDisk *loadedDisk) {
+	unsigned char ipltable[SECTOR_SIZE_MAX];
+
+	parts.clear();
+
+	assert(sizeof(_PC98RawPartition) == 32);
+	if (loadedDisk->getSectSize() > sizeof(ipltable)) return false;
+
+	memset(ipltable,0,sizeof(ipltable));
+	if (loadedDisk->Read_Sector(0,0,2,ipltable) != 0) return false;
+
+	const unsigned int max_entries = (std::min)(16UL,(unsigned long)(loadedDisk->getSectSize() / sizeof(_PC98RawPartition)));
+
+	for (size_t i=0;i < max_entries;i++) {
+		const _PC98RawPartition *pe = (_PC98RawPartition*)(ipltable+(i * 32));
+
+		if (pe->mid == 0 && pe->sid == 0 &&
+			pe->ipl_sect == 0 && pe->ipl_head == 0 && pe->ipl_cyl == 0 &&
+			pe->sector == 0 && pe->head == 0 && pe->cyl == 0 &&
+			pe->end_sector == 0 && pe->end_head == 0 && pe->end_cyl == 0)
+			continue; /* unused */
+
+		parts.push_back(*pe);
+	}
+
+	return true;
+}
+
+std::string PartitionIdentifyType(imageDisk *loadedDisk) {
+	struct partTable mbrData;
+
+	if (loadedDisk->Read_Sector(0,0,1,&mbrData) != 0)
+		return std::string();
+
+	if (!memcmp(mbrData.booter+4,"IPL1",4))
+		return "IPL1"; /* PC-98 IPL1 */
+
+	if (mbrData.magic1 == 0x55 && mbrData.magic2 == 0xaa)
+		return "MBR"; /* IBM PC MBR */
+
+	return std::string();
+}
+
+void LogPrintPartitionTable(const std::vector<_PC98RawPartition> &parts) {
+	for (size_t i=0;i < parts.size();i++) {
+		const _PC98RawPartition &part = parts[i];
+
+		LOG(LOG_DOSMISC,LOG_DEBUG)("IPL #%u: boot=%u active=%u startchs=%u/%u/%u endchs=%u/%u/%u '%s'",
+			(unsigned int)i,(part.mid&0x80)?1:0,(part.sid&0x80)?1:0,
+			part.cyl,part.head,part.sector,part.end_cyl,part.end_head,part.end_sector,
+			std::string((char*)part.name,sizeof(part.name)).c_str());
+	}
+}
+
+void LogPrintPartitionTable(const std::vector<partTable::partentry_t> &parts) {
+	for (size_t i=0;i < parts.size();i++) {
+		const partTable::partentry_t &part = parts[i];
+
+		LOG(LOG_DOSMISC,LOG_DEBUG)("MBR #%u: bootflag=%u parttype=0x%02x beginchs=0x%02x%02x%02x endchs=0x%02x%02x%02x start=%llu size=%llu",
+			(unsigned int)i,(part.bootflag&0x80)?1:0,part.parttype,
+			part.beginchs[0],part.beginchs[1],part.beginchs[2],
+			part.endchs[0],part.endchs[1],part.endchs[2],
+			(unsigned long long)part.absSectStart,
+			(unsigned long long)part.partSize);
+	}
+}
+
