@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,9 +16,11 @@
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+#include <assert.h>
 
 #include "dosbox.h"
 #include "control.h"
+#include "logging.h"
 #include "mem.h"
 #include "menu.h"
 #include "callback.h"
@@ -27,6 +29,8 @@
 #include "int10.h"
 #include "mouse.h"
 #include "setup.h"
+#include "render.h"
+#include "jfont.h"
 
 Int10Data int10;
 bool blinking=true;
@@ -34,19 +38,64 @@ static Bitu call_10 = 0;
 static bool warned_ff=false;
 extern bool enable_vga_8bit_dac;
 extern bool vga_8bit_dac;
+extern bool wpExtChar;
+extern int wpType;
 
+uint16_t GetTextSeg();
+void WriteCharTopView(uint16_t off, int count);
+uint8_t *GetSbcsFont(Bitu code);
+uint8_t *GetSbcs19Font(Bitu code);
+uint8_t *GetSbcs24Font(Bitu code);
+uint8_t *GetDbcsFont(Bitu code);
+uint8_t *GetDbcs24Font(Bitu code);
+void DOSV_FillScreen();
+void ResolvePath(std::string& in);
+void INT10_ReadString(uint8_t row, uint8_t col, uint8_t flag, uint8_t attr, PhysPt string, uint16_t count,uint8_t page);
+bool INT10_SetDOSVModeVtext(uint16_t mode, enum DOSV_VTEXT_MODE vtext_mode);
+#if defined(USE_TTF)
+extern bool colorChanged, justChanged;
+#endif
 Bitu INT10_Handler(void) {
 	// NTS: We do have to check the "current video mode" from the BIOS data area every call.
 	//      Some OSes like Windows 95 rely on overwriting the "current video mode" byte in
 	//      the BIOS data area to play tricks with the BIOS. If we don't call this, tricks
 	//      like the Windows 95 boot logo or INT 10h virtualization in Windows 3.1/9x/ME
 	//      within the DOS "box" will not work properly.
+	if(IS_DOSV && DOSV_CheckCJKVideoMode() && reg_ah != 0x03) DOSV_OffCursor();
 	INT10_SetCurMode();
 
 	switch (reg_ah) {
 	case 0x00:								/* Set VideoMode */
 		Mouse_BeforeNewVideoMode(true);
-		INT10_SetVideoMode(reg_al);
+		SetTrueVideoMode(reg_al);
+		if(IS_DOSV && IS_DOS_CJK && (reg_al == 0x03 || reg_al == 0x70 || reg_al == 0x72 || reg_al == 0x78)) {
+			uint8_t mode = reg_al;
+			if(reg_al == 0x03 || reg_al == 0x72) {
+				INT10_SetVideoMode(0x12);
+				INT10_SetDOSVModeVtext(mode, DOSV_VGA);
+			} else if(reg_al == 0x70 || reg_al == 0x78) {
+				mode = 0x70;
+				enum DOSV_VTEXT_MODE vtext_mode = DOSV_GetVtextMode((reg_al == 0x70) ? 0 : 1);
+				if(vtext_mode == DOSV_VTEXT_XGA || vtext_mode == DOSV_VTEXT_XGA_24) {
+					INT10_SetVideoMode(0x37);
+					INT10_SetDOSVModeVtext(mode, vtext_mode);
+				} else if(vtext_mode == DOSV_VTEXT_SXGA || vtext_mode == DOSV_VTEXT_SXGA_24) {
+					INT10_SetVideoMode(0x3d);
+					INT10_SetDOSVModeVtext(mode, vtext_mode);
+				} else if(vtext_mode == DOSV_VTEXT_SVGA) {
+					INT10_SetVideoMode((svgaCard == SVGA_TsengET3K) ? 0x29 : (svgaCard == SVGA_ParadisePVGA1A) ? 0x58 : 0x6a);
+					INT10_SetDOSVModeVtext(mode, vtext_mode);
+				} else {
+					INT10_SetVideoMode(0x12);
+					INT10_SetDOSVModeVtext(mode, DOSV_VTEXT_VGA);
+				}
+			}
+			DOSV_FillScreen();
+		} else {
+			if(reg_al == 0x74)
+				break;
+			INT10_SetVideoMode(reg_al);
+		}
 		Mouse_AfterNewVideoMode(true);
 		break;
 	case 0x01:								/* Set TextMode Cursor Shape */
@@ -128,11 +177,17 @@ Bitu INT10_Handler(void) {
 		INT10_GetPixel(reg_cx,reg_dx,reg_bh,&reg_al);
 		break;
 	case 0x0E:								/* Teletype OutPut */
-		INT10_TeletypeOutput(reg_al,reg_bl);
+		if(DOSV_CheckCJKVideoMode()) {
+			uint16_t attr;
+			INT10_ReadCharAttr(&attr, 0);
+			INT10_TeletypeOutput(reg_al, attr >> 8);
+		} else
+			INT10_TeletypeOutput(reg_al,reg_bl);
 		break;
 	case 0x0F:								/* Get videomode */
 		reg_bh=real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
-		reg_al=real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE)|(real_readb(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL)&0x80);
+		reg_al=real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
+		if (IS_EGAVGA_ARCH) reg_al|=real_readb(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL)&0x80;
 		reg_ah=(uint8_t)real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS);
 		break;					
 	case 0x10:								/* Palette functions */
@@ -152,12 +207,18 @@ Bitu INT10_Handler(void) {
 		switch (reg_al) {
 		case 0x00:							/* SET SINGLE PALETTE REGISTER */
 			INT10_SetSinglePaletteRegister(reg_bl,reg_bh);
+#if defined(USE_TTF)
+			colorChanged = justChanged = true;
+#endif
 			break;
 		case 0x01:							/* SET BORDER (OVERSCAN) COLOR*/
 			INT10_SetOverscanBorderColor(reg_bh);
 			break;
 		case 0x02:							/* SET ALL PALETTE REGISTERS */
 			INT10_SetAllPaletteRegisters(SegPhys(es)+reg_dx);
+#if defined(USE_TTF)
+			colorChanged = justChanged = true;
+#endif
 			break;
 		case 0x03:							/* TOGGLE INTENSITY/BLINKING BIT */
 			blinking=reg_bl==1;
@@ -238,6 +299,13 @@ Bitu INT10_Handler(void) {
 			INT10_LoadFont(Real2Phys(int10.rom.font_8_first),reg_al==0x12,256,0,reg_bl&0x7f,8);
 			break;
 		case 0x03:			/* Set Block Specifier */
+#if defined(USE_TTF)
+            if (ttf.inUse&&wpType==1) {
+                DOS_Block dos;
+                if (mem_readd(((dos.psp()-1)<<4)+8) == 0x5057)							// Name of MCB PSP should be WP
+                    wpExtChar = reg_bl != 0;
+            }
+#endif
 			IO_Write(0x3c4,0x3);IO_Write(0x3c5,reg_bl);
 			break;
 		case 0x04:			/* Load 8x16 font */
@@ -532,7 +600,46 @@ CX	640x480	800x600	  1024x768/1280x1024
 		}
 		break;
 	case 0x13:								/* Write String */
-		INT10_WriteString(reg_dh,reg_dl,reg_al,reg_bl,SegPhys(es)+reg_bp,reg_cx,reg_bh);
+		if((reg_al & 0x10) != 0 && DOSV_CheckCJKVideoMode())
+			INT10_ReadString(reg_dh,reg_dl,reg_al,reg_bl,SegPhys(es)+reg_bp,reg_cx,reg_bh);
+		else
+			INT10_WriteString(reg_dh,reg_dl,reg_al,reg_bl,SegPhys(es)+reg_bp,reg_cx,reg_bh);
+		break;
+	case 0x18:
+		if(IS_DOSV && DOSV_CheckCJKVideoMode()) {
+			uint8_t *font;
+			Bitu size = 0;
+			if(reg_al == 0) {
+				reg_al = 1;
+				if(reg_bx == 0) {
+					if(reg_dh == 8) {
+						if(reg_dl == 16) {
+							font = GetSbcsFont(reg_cl);
+							size = 16;
+						} else if(reg_dl == 19) {
+							font = GetSbcs19Font(reg_cl);
+							size = 19;
+						}
+					} else if(reg_dh == 16 && reg_dl == 16) {
+						font = GetDbcsFont(reg_cx);
+						size = 2 * 16;
+					} else if(reg_dh == 12 && reg_dl == 24) {
+						font = GetSbcs24Font(reg_cl);
+						size = 2 * 24;
+					} else if(reg_dh == 24 && reg_dl == 24) {
+						font = GetDbcs24Font(reg_cx);
+						size = 3 * 24;
+					}
+					if(size > 0) {
+						uint16_t seg = SegValue(es);
+						uint16_t off = reg_si;
+						for(Bitu ct = 0 ; ct < size ; ct++)
+							real_writeb(seg, off++, *font++);
+						reg_al = 0;
+					}
+				}
+			}
+		}
 		break;
 	case 0x1A:								/* Display Combination */
 		if (!IS_VGA_ARCH && machine != MCH_MCGA) break;
@@ -579,9 +686,22 @@ CX	640x480	800x600	  1024x768/1280x1024
 				break;
 		}
 		break;
+	case 0x1d:
+		if(IS_DOSV && DOSV_CheckCJKVideoMode()) {
+			if(reg_al == 0x00) {
+				real_writeb(BIOSMEM_SEG, BIOSMEM_NB_ROWS, int10.text_row - reg_bl);
+			} else if(reg_al == 0x01) {
+				real_writeb(BIOSMEM_SEG, BIOSMEM_NB_ROWS, int10.text_row);
+			} else if(reg_al == 0x02) {
+				reg_bx = int10.text_row - real_readb(BIOSMEM_SEG, BIOSMEM_NB_ROWS);
+			}
+		}
+		break;
 	case 0x4f:								/* VESA Calls */
 		if ((!IS_VGA_ARCH) || (svgaCard!=SVGA_S3Trio))
 			break;
+		if (int10.vesa_oldvbe10 && reg_al >= 6) /* Functions 6 and up did not exist until VBE 1.2 (or 1.1?) */
+			break;
 		switch (reg_al) {
 		case 0x00:							/* Get SVGA Information */
 			reg_al=0x4f;
@@ -627,7 +747,7 @@ CX	640x480	800x600	  1024x768/1280x1024
 			break;
 		case 0x05:							
 			if (reg_bh==0) {				/* Set CPU Window */
-				reg_ah=VESA_SetCPUWindow(reg_bl,reg_dl);
+				reg_ah=VESA_SetCPUWindow(reg_bl,reg_dx);
 				reg_al=0x4f;
 			} else if (reg_bh == 1) {		/* Get CPU Window */
 				reg_ah=VESA_GetCPUWindow(reg_bl,reg_dx);
@@ -758,6 +878,185 @@ CX	640x480	800x600	  1024x768/1280x1024
 			break;
 		}
 		break;
+	case 0x50:// Set/Read JP/US mode of CRT BIOS
+		switch (reg_al) {
+			case 0x00:
+				LOG(LOG_INT10, LOG_NORMAL)("AX CRT BIOS 5000h is called.");
+				if (INT10_AX_SetCRTBIOSMode(reg_bx)) reg_al = 0x00;
+				else reg_al = 0x01;
+				break;
+			case 0x01:
+				//LOG(LOG_INT10,LOG_NORMAL)("AX CRT BIOS 5001h is called.");
+				reg_bx=INT10_AX_GetCRTBIOSMode();
+				reg_al=0;
+				break;
+			default:
+				LOG(LOG_INT10,LOG_ERROR)("Unhandled AX Function %X",reg_al);
+				reg_al=0x0;
+				break;
+		}
+		break;
+	case 0x51:// Save/Read JFONT pattern
+		if(INT10_AX_GetCRTBIOSMode() == 0x01) break;//exit if CRT BIOS is in US mode
+		switch (reg_al) {
+			//INT 10h/AX=5100h Store user font pattern
+			//IN
+			//ES:BP=Index of store buffer
+			//DX=Char code
+			//BH=width bits of char
+			//BL=height bits of char
+			//OUT
+			//AL=status 00h=Success 01h=Failed
+		case 0x00:
+		{
+			LOG(LOG_INT10, LOG_NORMAL)("AX CRT BIOS 5100h is called.");
+			Bitu buf_es = SegValue(es);
+			Bitu buf_bp = reg_bp;
+			Bitu chr = reg_dx;
+			Bitu w_chr = reg_bh;
+			Bitu h_chr = reg_bl;
+			Bitu font;
+			if (w_chr == 16 && h_chr == 16) {
+				for (Bitu line = 0; line < 16; line++)
+				{
+					//Order of font pattern is different between FONTX2 and AX(JEGA).
+					font = real_readb(buf_es, buf_bp + line);
+					jfont_dbcs_16[chr * 32 + line * 2] = font;
+					font = real_readb(buf_es, buf_bp + line + 16);
+					jfont_dbcs_16[chr * 32 + line * 2 + 1] = font;
+				}
+				reg_al = 0x00;
+			}
+			else
+				reg_al = 0x01;
+			break;
+		}
+		//INT 10h/AX=5101h Read character pattern
+		//IN
+		//ES:BP=Index of read buffer
+		//DX=Char code
+		//BH=width bits of char
+		//BL=height bits of char
+		//OUT
+		//AL=status 00h=Success 01h=Failed
+		case 0x01:
+		{
+			LOG(LOG_INT10, LOG_NORMAL)("AX CRT BIOS 5101h is called.");
+			Bitu buf_es = SegValue(es);
+			Bitu buf_bp = reg_bp;
+			Bitu chr = reg_dx;
+			Bitu w_chr = reg_bh;
+			Bitu h_chr = reg_bl;
+			Bitu font;
+			if (w_chr == 8) {
+				reg_al = 0x00;
+				switch (h_chr)
+				{
+				case 8:
+					for (Bitu line = 0; line < 8; line++)
+						real_writeb(buf_es, buf_bp + line, int10_font_08[chr * 8 + line]);
+					break;
+				case 14:
+					for (Bitu line = 0; line < 14; line++)
+						real_writeb(buf_es, buf_bp + line, int10_font_14[chr * 14 + line]);
+					break;
+				case 19:
+					for (Bitu line = 0; line < 19; line++)
+						real_writeb(buf_es, buf_bp + line, jfont_sbcs_19[chr * 19 + line]);
+					break;
+				default:
+					reg_al = 0x01;
+					break;
+				}
+			}
+			else if (w_chr == 16 && h_chr == 16) {
+				reg_al = 0x00;
+				for (Bitu line = 0; line < 16; line++)
+				{
+					font = jfont_dbcs_16[chr * 32 + line * 2];
+					real_writeb(buf_es, buf_bp + line, font);
+					font = jfont_dbcs_16[chr * 32 + line * 2 + 1];
+					real_writeb(buf_es, buf_bp + line + 16, font);
+				}
+			}
+			else
+				reg_al = 0x01;
+			break;
+		}
+		default:
+			LOG(LOG_INT10,LOG_ERROR)("Unhandled AX Function %X",reg_al);
+			reg_al=0x1;
+			break;
+		}
+		break;
+	case 0x52:// Set/Read virtual text ram buffer when the video mode is JEGA graphic mode
+		if(INT10_AX_GetCRTBIOSMode() == 0x01) break;//exit if CRT BIOS is in US mode
+		LOG(LOG_INT10,LOG_NORMAL)("AX CRT BIOS 52xxh is called.");
+		switch (reg_al) {
+		case 0x00:
+		{
+			if (reg_bx == 0) real_writew(BIOSMEM_AX_SEG, BIOSMEM_AX_VTRAM_SEGADDR, 0);
+			else
+			{
+				LOG(LOG_INT10, LOG_NORMAL)("AX CRT BIOS set VTRAM segment address at %x", reg_bx);
+				real_writew(BIOSMEM_AX_SEG, BIOSMEM_AX_VTRAM_SEGADDR, reg_bx);
+				/* Fill VTRAM with 0x20(Space) */
+				for (int y = 0; y < 25; y++)
+					for (int x = 0; x < 80; x++)
+						SetVTRAMChar(x, y, 0x20, 0x00);
+			}
+			break;
+		}
+		case 0x01:
+		{
+			Bitu vtram_seg = real_readw(BIOSMEM_AX_SEG, BIOSMEM_AX_VTRAM_SEGADDR);
+			if (vtram_seg == 0x0000) reg_bx = 0;
+			else reg_bx = vtram_seg;
+			break;
+		}
+		default:
+			LOG(LOG_INT10,LOG_ERROR)("Unhandled AX Function %X",reg_al);
+			break;
+		}
+		break;
+	case 0x82:// Set/Read the scroll mode when the video mode is JEGA graphic mode
+		if(INT10_AX_GetCRTBIOSMode() == 0x01) break;//exit if CRT BIOS is in US mode
+		LOG(LOG_INT10,LOG_NORMAL)("AX CRT BIOS 82xxh is called.");
+		switch (reg_al) {
+		case 0x00:
+			if (reg_bl == -1) {//Read scroll mode
+				reg_al = real_readb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS) & 0x01;
+			}
+			else {//Set scroll mode
+				uint8_t tmp = real_readb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS);
+				reg_al = tmp & 0x01;//Return previous scroll mode
+				tmp |= (reg_bl & 0x01);
+				real_writeb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS, tmp);
+			}
+			break;
+		default:
+			LOG(LOG_INT10,LOG_ERROR)("Unhandled AX Function %X",reg_al);
+			break;
+		}
+		break;
+	case 0x83:// Read the video RAM address and virtual text video RAM
+		//Out: AX=base address of video RAM, ES:BX=base address of virtual text video RAM
+		if(INT10_AX_GetCRTBIOSMode() == 0x01) break;//exit if CRT BIOS is in US mode
+		LOG(LOG_INT10,LOG_NORMAL)("AX CRT BIOS 83xxh is called.");
+		switch (reg_al) {
+		case 0x00:
+		{
+			reg_ax = CurMode->pstart;
+			Bitu vtram_seg = real_readw(BIOSMEM_AX_SEG, BIOSMEM_AX_VTRAM_SEGADDR);
+			RealMakeSeg(es, vtram_seg >> 4);
+			reg_bx = vtram_seg << 4;
+			break;
+		}
+		default:
+			LOG(LOG_INT10,LOG_ERROR)("Unhandled AX Function %X",reg_al);
+			break;
+		}
+		break;
 	case 0xf0:
 		INT10_EGA_RIL_ReadRegister(reg_bl, reg_dx);
 		break;
@@ -782,9 +1081,19 @@ CX	640x480	800x600	  1024x768/1280x1024
 		reg_bx=RealOff(pt);
 		}
 		break;
+	case 0xfe:
+		if(IS_DOSV && DOSV_CheckCJKVideoMode()) {
+			reg_di = 0x0000;
+			SegSet16(es, GetTextSeg());
+		}
+		break;
 	case 0xff:
-		if (!warned_ff) LOG(LOG_INT10,LOG_NORMAL)("INT10:FF:Weird NC call");
-		warned_ff=true;
+		if(IS_DOSV && DOSV_CheckCJKVideoMode()) {
+			WriteCharTopView(reg_di, reg_cx);
+		} else {
+			if (!warned_ff) LOG(LOG_INT10,LOG_NORMAL)("INT10:FF:Weird NC call");
+			warned_ff=true;
+		}
 		break;
 	default:
 		LOG(LOG_INT10,LOG_ERROR)("Function %4X not supported",reg_ax);
@@ -794,19 +1103,42 @@ CX	640x480	800x600	  1024x768/1280x1024
 	return CBRET_NONE;
 }
 
+bool DISP2_Active(void);
 static void INT10_Seg40Init(void) {
-	// the default char height
-	real_writeb(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT,16);
-	// Clear the screen 
-	real_writeb(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL,0x60);
-	// Set the basic screen we have
-	real_writeb(BIOSMEM_SEG,BIOSMEM_SWITCHES,0xF9);
-	// Set the basic modeset options
-	real_writeb(BIOSMEM_SEG,BIOSMEM_MODESET_CTL,0x51); // why is display switching enabled (bit 6) ?
-	// Set the  default MSR
+	// Clear the screen.
+	//
+	// The byte value at 40:87 (BIOSMEM_VIDEO_CTL) only means something if EGA/VGA.
+	// Should be zero for PCjr (which defines it as function key), and Tandy (which does not define it),
+	// and MDA/Hercules/CGA (which also does not define this byte).
+	// 
+	// Furthermore, some games such as "Road Runner" by Mindscape use that byte as part of it's
+	// detection routines to differentiate between CGA, EGA, and Tandy. For the game to work properly
+	// in Tandy emulation, this byte must be zero.
+	if (IS_EGAVGA_ARCH) {
+		real_writeb(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL,0x60);
+		real_writeb(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT,16);
+		real_writeb(BIOSMEM_SEG,BIOSMEM_SWITCHES,0xF9);
+		// Set the pointer to video save pointer table
+		real_writed(BIOSMEM_SEG, BIOSMEM_VS_POINTER, int10.rom.video_save_pointers);
+	}
+	else {
+		real_writeb(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL,0x00);
+		if (machine == MCH_TANDY || machine == MCH_CGA || machine == MCH_PCJR)
+			real_writeb(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT,8); /* FIXME: INT 10h teletext routines depend on this */
+		else
+			real_writeb(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT,0);
+		real_writeb(BIOSMEM_SEG,BIOSMEM_SWITCHES,0x00);
+	}
+#if C_DEBUG
+	if (control->opt_display2)
+		real_writeb(BIOSMEM_SEG,BIOSMEM_MODESET_CTL,0x10|(DISP2_Active()?0:1));
+    else
+#endif
+	if (IS_EGAVGA_ARCH) {
+		real_writeb(BIOSMEM_SEG,BIOSMEM_MODESET_CTL,0x51); // why is display switching enabled (bit 6) ?
+	}
+	// Set the default MSR
 	real_writeb(BIOSMEM_SEG,BIOSMEM_CURRENT_MSR,0x09);
-	// Set the pointer to video save pointer table
-	real_writed(BIOSMEM_SEG, BIOSMEM_VS_POINTER, int10.rom.video_save_pointers);
 }
 
 static void INT10_InitVGA(void) {
@@ -1000,14 +1332,19 @@ fail:
  * Holds the font as a giant 2048x2048 1-bit monochromatic bitmap. */
 /* We load it separately because I am uncertain whether it is legal or not to
  * incorporate this directly into DOSBox-X. */
-bool Load_Anex86_Font(void) {
+bool Load_Anex86_Font(const char *fontname) {
     unsigned char tmp[(2048/8)*16]; /* enough for one 2048x16 row and bitmap header */
     unsigned int hibyte,lowbyte,r;
     unsigned int bmp_ofs;
     FILE *fp = NULL;
 
+    if (fontname&&*fontname) {
+        std::string name = fontname;
+        ResolvePath(name);
+        fp = fopen(name.c_str(),"rb");
+    }
     /* ANEX86.BMP accurate dump of actual font */
-    fp = fopen("anex86.bmp","rb");
+    if (!fp) fp = fopen("anex86.bmp","rb");
     if (!fp) fp = fopen("ANEX86.bmp","rb");
     if (!fp) fp = fopen("ANEX86.BMP","rb");
 
@@ -1094,15 +1431,26 @@ fail:
     return false;
 }
 
-extern uint8_t int10_font_16[256 * 16];
-
 extern VideoModeBlock PC98_Mode;
+extern uint8_t pc98_freecg_sbcs[256 * 16];
 
-bool Load_VGAFont_As_PC98(void) {
-    unsigned int i;
+bool Load_JFont_As_PC98(void) {
+    unsigned int hibyte,lowbyte,r,o,i,i1,i2;
 
-    for (i=0;i < (256 * 16);i++)
-        vga.draw.font[i] = int10_font_16[i];
+    for (lowbyte=0;lowbyte < 256;lowbyte++) {
+        for (r=0;r < 16;r++)
+            vga.draw.font[(lowbyte*16)+r] = pc98_freecg_sbcs[(lowbyte*16)+r];
+    }
+    for (lowbyte=33;lowbyte < 125;lowbyte++) {
+        for (hibyte=32;hibyte < 128;hibyte++) {
+            i1=(lowbyte+1)/2+(lowbyte<95?112:176);
+            i2=hibyte+(lowbyte%2?31+(hibyte/96):126);
+            i = i1 * 0x100 + i2;
+            o = ((hibyte * 128) + (lowbyte - 32)) * 16 * 2;
+            uint8_t *font = GetDbcsFont(i);
+            for (r=0;r < 32;r++) vga.draw.font[o+r] = font[r];
+        }
+    }
 
     return true;
 }
@@ -1119,6 +1467,61 @@ RealPt GetSystemBiosINT10Vector(void) {
         return 0;
 }
 
+Bitu INT10_AX_GetCRTBIOSMode(void) {
+	if (!IS_JEGA_ARCH) return 0x01;
+	if (real_readb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS) & 0x80) return 0x51;//if in US mode
+	else return 0x01;
+}
+
+bool INT10_AX_SetCRTBIOSMode(Bitu mode) {
+	if (!IS_JEGA_ARCH) return false;
+	uint8_t tmp = real_readb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS);
+	switch (mode) {
+		//Todo: verify written value
+	case 0x01:
+		real_writeb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS, tmp & 0x7F);
+		LOG(LOG_INT10, LOG_NORMAL)("AX CRT BIOS has been set to US mode.");
+		INT10_SetVideoMode(0x03);
+		return true;
+		/* -------------------SET to JP mode in CRT BIOS -------------------- */
+	case 0x51:
+		real_writeb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS, tmp | 0x80);
+		LOG(LOG_INT10, LOG_NORMAL)("AX CRT BIOS has been set to JP mode.");
+		//		Mouse_BeforeNewVideoMode(true);
+		// change to the default video mode (03h) with vram cleared
+		INT10_SetVideoMode(0x03);
+		//		Mouse_AfterNewVideoMode(true);;
+		return true;
+	default:
+		return false;
+	}
+}
+Bitu INT16_AX_GetKBDBIOSMode(void) {
+	if (!IS_JEGA_ARCH) return 0x01;
+	if (real_readb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS) & 0x40) return 0x51;//if in US mode
+	else return 0x01;
+}
+
+bool INT16_AX_SetKBDBIOSMode(Bitu mode) {
+	if (!IS_JEGA_ARCH) return false;
+	uint8_t tmp = real_readb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS);
+	switch (mode) {
+		//Todo: verify written value
+	case 0x01:
+		real_writeb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS, tmp & 0xBF);
+		LOG(LOG_INT10, LOG_NORMAL)("AX KBD BIOS has been set to US mode.");
+		return true;
+	case 0x51:
+		real_writeb(BIOSMEM_AX_SEG, BIOSMEM_AX_JPNSTATUS, tmp | 0x40);
+		LOG(LOG_INT10, LOG_NORMAL)("AX KBD BIOS has been set to JP mode.");
+		return true;
+	default:
+		return false;
+	}
+}
+
+extern bool VGA_BIOS_use_rom;
+
 void INT10_Startup(Section *sec) {
     (void)sec;//UNUSED
 	LOG(LOG_MISC,LOG_DEBUG)("INT 10h reinitializing");
@@ -1133,22 +1536,29 @@ void INT10_Startup(Section *sec) {
 	int16_unmask_irq1_on_read = static_cast<Section_prop *>(control->GetSection("dosbox"))->Get_bool("unmask keyboard on int 16 read");
 	int16_ah_01_cf_undoc = static_cast<Section_prop *>(control->GetSection("dosbox"))->Get_bool("int16 keyboard polling undocumented cf behavior");
 	int10_vga_bios_vector = video_section->Get_bool("int 10h points at vga bios");
+	int size_override = video_section->Get_int("vga bios size override");
+
+	Section_prop * pc98_section = static_cast<Section_prop *>(control->GetSection("pc98"));
+	bool try_font_rom = pc98_section->Get_bool("pc-98 try font rom");
+	const char *anex86_font = pc98_section->Get_string("pc-98 anex86 font");
 
     if (!IS_PC98_ARCH) {
-        INT10_InitVGA();
-        if (IS_TANDY_ARCH) SetupTandyBios();
-        /* Setup the INT 10 vector */
-        call_10=CALLBACK_Allocate();	
-        CALLBACK_Setup(call_10,&INT10_Handler,CB_IRET,"Int 10 video");
-        RealSetVec(0x10,CALLBACK_RealPointer(call_10));
-        //Init the 0x40 segment and init the datastructures in the the video rom area
-        INT10_SetupRomMemory();
-        INT10_Seg40Init();
-        INT10_SetupBasicVideoParameterTable();
+        if (!VGA_BIOS_use_rom) {
+            INT10_InitVGA();
+            if (IS_TANDY_ARCH) SetupTandyBios();
+            /* Setup the INT 10 vector */
+            call_10=CALLBACK_Allocate();	
+            CALLBACK_Setup(call_10,&INT10_Handler,CB_IRET,"Int 10 video");
+            RealSetVec(0x10,CALLBACK_RealPointer(call_10));
+            //Init the 0x40 segment and init the datastructures in the the video rom area
+            INT10_SetupRomMemory();
+            INT10_Seg40Init();
+            INT10_SetupBasicVideoParameterTable();
 
-        LOG(LOG_MISC,LOG_DEBUG)("INT 10: VGA bios used %d / %d memory",(int)int10.rom.used,(int)VGA_BIOS_Size);
-        if (int10.rom.used > VGA_BIOS_Size) /* <- this is fatal, it means the Setup() functions scrozzled over the adjacent ROM or RAM area */
-            E_Exit("VGA BIOS size too small %u > %u",(unsigned int)int10.rom.used,(unsigned int)VGA_BIOS_Size);
+            LOG(LOG_MISC,LOG_DEBUG)("INT 10: VGA bios used %d / %d memory",(int)int10.rom.used,(int)VGA_BIOS_Size);
+            if (int10.rom.used > VGA_BIOS_Size && size_override > -512) /* <- this is fatal, it means the Setup() functions scrozzled over the adjacent ROM or RAM area */
+                E_Exit("VGA BIOS size too small %u > %u",(unsigned int)int10.rom.used,(unsigned int)VGA_BIOS_Size);
+        }
 
         /* NTS: Uh, this does seem bass-ackwards... INT 10h making the VGA BIOS appear. Can we refactor this a bit? */
         if (VGA_BIOS_Size > 0) {
@@ -1161,17 +1571,17 @@ void INT10_Startup(Section *sec) {
         }
 
         INT10_SetVideoMode(0x3);
+        SetTrueVideoMode(0x03);
     }
     else {
         /* load PC-98 character ROM data, if possible */
         {
             /* We can use FONT.ROM as generated by T98Tools */
-            bool ok = Load_FONT_ROM();
+            bool ok = try_font_rom?Load_FONT_ROM():false;
             /* We can use ANEX86.BMP from the Anex86 emulator */
-            if (!ok) ok = Load_Anex86_Font();
-            /* Failing all else we can just re-use the IBM VGA 8x16 font to show SOMETHING on the screen.
-             * Japanese text will not display properly though. */
-            if (!ok) ok = Load_VGAFont_As_PC98();
+            if (!ok) ok = Load_Anex86_Font(anex86_font);
+            /* Failing all else we can use the internal FREECG 8x16 font and default Japanese font to show text on the screen. */
+            if (!ok) ok = Load_JFont_As_PC98();
         }
 
         CurMode = &PC98_Mode;
@@ -1181,21 +1591,25 @@ void INT10_Startup(Section *sec) {
          * Compare that to IBM PC platform, where segment fills only 0x40:0x00 to 0x50:0x00 inclusive and extra state is held in the "Extended BIOS Data Area".
          */
 
+        real_writeb(0x60,0x8A,1); /* kanji/graph mode */
+        real_writeb(0x60,0x8B,' '); /* kanji/graph mode indicator */
+        real_writeb(0x60,0x8C,' '); /* function row mode 2 indicator */
+
         /* number of text rows on the screen.
          * Touhou Project will not clear/format the text layer properly without this variable. */
-        mem_writeb(0x710,0); /* cursor position Y coordinate */
-        mem_writeb(0x711,1); /* function definition display status flag */
-        mem_writeb(0x712,25 - 1 - 1); /* scroll range lower limit (usually 23 when function key row is visible) */
-        mem_writeb(0x713,1); /* normal 25 lines */
-        mem_writeb(0x714,0xE1); /* content erase attribute */
+        real_writeb(0x60,0x110,0); /* cursor position Y coordinate */
+        real_writeb(0x60,0x111,1); /* function definition display status flag */
+        real_writeb(0x60,0x112,25 - 1 - 1); /* scroll range lower limit (usually 23 when function key row is visible) */
+        real_writeb(0x60,0x113,1); /* normal 25 lines */
+        real_writeb(0x60,0x114,0xE1); /* content erase attribute */
 
-        mem_writeb(0x719,0x20); /* content erase character */
+        real_writeb(0x60,0x119,0x20); /* content erase character */
 
-        mem_writeb(0x71B,0x01); /* cursor displayed */
-        mem_writeb(0x71C,0x00); /* cursor position X coordinate */
-        mem_writeb(0x71D,0xE1); /* content display attribute */
-        mem_writeb(0x71E,0x00); /* scroll range upper limit (usually 0) */
-        mem_writeb(0x71F,0x01); /* scrolling speed is normal */
+        real_writeb(0x60,0x11B,0x01); /* cursor displayed */
+        real_writeb(0x60,0x11C,0x00); /* cursor position X coordinate */
+        real_writeb(0x60,0x11D,0xE1); /* content display attribute */
+        real_writeb(0x60,0x11E,0x00); /* scroll range upper limit (usually 0) */
+        real_writeb(0x60,0x11F,0x01); /* scrolling speed is normal */
 
         /* init text RAM */
         for (unsigned int i=0;i < 0x1FE0;i += 2) {
