@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -95,13 +95,17 @@ extern uint32_t S3_LFB_BASE;
 
 /* FIXME: Wait, what?? What the hell kind of preprocessor macro is this??? Kill these macros! --J.C. */
 #define BIOS_NCOLS uint16_t ncols=IS_PC98_ARCH ? 80 : real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS);
-#define BIOS_NROWS uint16_t nrows=IS_PC98_ARCH ? (uint16_t)(real_readb(0x60,0x112)+1u) : (uint16_t)real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1u;
+#define BIOS_NROWS uint16_t nrows=IS_PC98_ARCH ? (uint16_t)(real_readb(0x60,0x112)+1u) : IS_EGAVGA_ARCH?((uint16_t)real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1):25;
+#define BIOS_CHEIGHT uint8_t cheight=IS_EGAVGA_ARCH?real_readb(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT):8;
 
 extern uint8_t int10_font_08[256 * 8];
 extern uint8_t int10_font_14[256 * 14];
 extern uint8_t int10_font_16[256 * 16];
+extern uint8_t int10_font_19[256 * 19];
 extern uint8_t int10_font_14_alternate[20 * 15 + 1];
 extern uint8_t int10_font_16_alternate[19 * 17 + 1];
+extern uint8_t int10_font_14_init[256 * 14];
+extern uint8_t int10_font_16_init[256 * 16];
 
 struct VideoModeBlock {
 	uint16_t	mode;
@@ -127,6 +131,7 @@ typedef struct {
 		RealPt font_16;
 		RealPt font_14_alternate;
 		RealPt font_16_alternate;
+		RealPt font_19;
 		RealPt static_state;
 		RealPt video_save_pointers;
         RealPt video_dynamic_save_area;
@@ -148,6 +153,8 @@ typedef struct {
 	uint16_t vesa_setmode;
 	bool vesa_nolfb;
 	bool vesa_oldvbe;
+	bool vesa_oldvbe10;
+	uint8_t text_row;
 } Int10Data;
 
 #define _EGA_HALF_CLOCK			0x0001
@@ -233,7 +240,7 @@ uint8_t VESA_GetSVGAInformation(uint16_t seg,uint16_t off);
 uint8_t VESA_GetSVGAModeInformation(uint16_t mode,uint16_t seg,uint16_t off);
 uint8_t VESA_SetSVGAMode(uint16_t mode);
 uint8_t VESA_GetSVGAMode(uint16_t & mode);
-uint8_t VESA_SetCPUWindow(uint8_t window,uint8_t address);
+uint8_t VESA_SetCPUWindow(uint8_t window,uint16_t address);
 uint8_t VESA_GetCPUWindow(uint8_t window,uint16_t & address);
 uint8_t VESA_ScanLineLength(uint8_t subcall, uint16_t val, uint16_t & bytes,uint16_t & pixels,uint16_t & lines);
 uint8_t VESA_SetDisplayStart(uint16_t x,uint16_t y,bool wait);
