@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,9 +23,9 @@
 #include <string.h>
 #include <math.h>
 
-
 #include "dosbox.h"
 #include "callback.h"
+#include "logging.h"
 #include "mem.h"
 #include "regs.h"
 #include "cpu.h"
@@ -34,10 +34,12 @@
 #include "inout.h"
 #include "int10.h"
 #include "bios.h"
+#include "jfont.h"
 #include "dos_inc.h"
 #include "support.h"
 #include "setup.h"
 #include "control.h"
+#include "SDL.h"
 
 #if defined(_MSC_VER)
 # pragma warning(disable:4244) /* const fmath::local::uint64_t to double possible loss of data */
@@ -45,11 +47,15 @@
 
 /* ints/bios.cpp */
 void bios_enable_ps2();
+void WriteCharTopView(uint16_t off, int count);
+uint16_t GetTextSeg();
 
 /* hardware/keyboard.cpp */
 void AUX_INT33_Takeover();
 int KEYBOARD_AUX_Active();
 void KEYBOARD_AUX_Event(float x,float y,Bitu buttons,int scrollwheel);
+extern bool MOUSE_IsLocked();
+extern bool usesystemcursor, del_flag;
 
 bool en_int33=false;
 bool en_bios_ps2mouse=false;
@@ -498,6 +504,10 @@ void RestoreCursorBackground() {
 
 void DrawCursor() {
     if (mouse.hidden || mouse.inhibit_draw) return;
+    if (usesystemcursor&&!MOUSE_IsLocked()) {
+        SDL_ShowCursor(SDL_ENABLE);
+        return;
+    }
     INT10_SetCurMode();
     // In Textmode ?
     if (CurMode->type==M_TEXT) {
@@ -507,7 +517,9 @@ void DrawCursor() {
 
     // Check video page. Seems to be ignored for text mode. 
     // hence the text mode handled above this
-    if (real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE)!=mouse.page) return;
+    // >>> removed because BIOS page is not actual page in some cases, e.g. QQP games
+    // if (real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE)!=mouse.page) return;
+
 // Check if cursor in update region
 /*  if ((POS_X >= mouse.updateRegion_x[0]) && (POS_X <= mouse.updateRegion_x[1]) &&
         (POS_Y >= mouse.updateRegion_y[0]) && (POS_Y <= mouse.updateRegion_y[1])) {
@@ -590,6 +602,12 @@ static inline bool GFX_IsFullscreen(void) {
     return false;
 }
 #endif
+#include "render.h"
+
+void KEYBOARD_AUX_LowerIRQ() {
+    if (MOUSE_IRQ != 0)
+        PIC_SetIRQMask(MOUSE_IRQ,false);
+}
 
 extern int  user_cursor_x,  user_cursor_y;
 extern int  user_cursor_sw, user_cursor_sh;
@@ -635,7 +653,7 @@ void Mouse_CursorMoved(float xrel,float yrel,float x,float y,bool emulate) {
     } else if (CurMode != NULL) {
         if (CurMode->type == M_TEXT) {
             mouse.x = x*real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS)*8;
-            mouse.y = y*(real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1)*8;
+            mouse.y = y*(IS_EGAVGA_ARCH?(real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1):25)*8;
         /* NTS: DeluxePaint II enhanced sets a large range (5112x3832) for VGA mode 0x12 640x480 16-color */
         } else {
             if ((mouse.max_x > 0) && (mouse.max_y > 0)) {
@@ -707,9 +725,11 @@ uint8_t Mouse_GetButtonState(void) {
     return mouse.buttons;
 }
 
-#if defined(WIN32) || defined(C_SDL2)
 char text[5000];
+extern bool isDBCSCP();
+extern std::vector<std::pair<int,int>> jtbs;
 const char* Mouse_GetSelected(int x1, int y1, int x2, int y2, int w, int h, uint16_t *textlen) {
+	uint16_t result=0, len=0;
 	uint8_t page = real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
 	uint16_t c=0, r=0;
 	if (IS_PC98_ARCH) {
@@ -717,9 +737,15 @@ const char* Mouse_GetSelected(int x1, int y1, int x2, int y2, int w, int h, uint
 		r=real_readb(0x60,0x113) & 0x01 ? 25 : 20;
 	} else {
 		c=real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS);
-		r=(uint16_t)real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
+		r=(uint16_t)(IS_EGAVGA_ARCH?real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS):24)+1;
 	}
-	int c1=c*x1/w, r1=r*y1/h, c2=c*x2/w, r2=r*y2/h, t;
+    int c1=x1, r1=y1, c2=x2, r2=y2, t;
+    if (w>0&&h>0) {
+        c1=c*x1/w;
+        r1=r*y1/h;
+        c2=c*x2/w;
+        r2=r*y2/h;
+    }
 	if (c1>c2) {
 		t=c1;
 		c1=c2;
@@ -730,10 +756,39 @@ const char* Mouse_GetSelected(int x1, int y1, int x2, int y2, int w, int h, uint
 		r1=r2;
 		r2=t;
 	}
-	uint16_t result=0, len=0;
 	text[0]=0;
+    uint16_t seg = IS_DOSV?GetTextSeg():0;
+#if defined(USE_TTF)
+    if (ttf.inUse&&isDBCSCP()&&!(c1==0&&c2==(int)(ttf.cols-1)&&r1==0&&r2==(int)(ttf.lins-1))) {
+        ttf_cell *curAC = curAttrChar;
+        for (unsigned int y = 0; y < ttf.lins; y++) {
+            if ((int)y>=r1&&(int)y<=r2)
+                for (unsigned int x = 0; x < ttf.cols; x++)
+                    if ((int)x>=c1&&(int)x<=c2&&curAC[x].selected) {
+                        if ((int)x==c1&&c1>0&&curAC[x].skipped&&!curAC[x-1].selected&&curAC[x-1].doublewide) {
+                            ReadCharAttr(x-1,y,page,&result);
+                            text[len++]=result;
+                        }
+                        ReadCharAttr(x,y,page,&result);
+                        text[len++]=result;
+                        if ((int)x==c2&&c2<(int)(ttf.cols-1)&&curAC[x].doublewide&&!curAC[x+1].selected&&curAC[x+1].skipped) {
+                            ReadCharAttr(x+1,y,page,&result);
+                            text[len++]=result;
+                        }
+                    }
+            curAC += ttf.cols;
+        }
+    } else
+#endif
 	for (int i=r1; i<=r2; i++) {
-		for (int j=c1; j<=c2; j++) {
+        bool lead1 = false, lead2 = false;
+        if (IS_DOSV) {
+            for (int k=0; k<c1; k++) {
+                if (lead1) lead1=false;
+                else if (isKanji1(real_readb(seg,(i*c+k)*2))) lead1=true;
+            }
+        }
+		for (int j=c1+(lead1?1:0); j<=c2; j++) {
 			if (IS_PC98_ARCH) {
 				uint16_t address=((i*80)+j)*2;
 				PhysPt where = CurMode->pstart+address;
@@ -744,12 +799,37 @@ const char* Mouse_GetSelected(int x1, int y1, int x2, int y2, int w, int h, uint
 					if (j1>32&&j1<127&&j2>32&&j2<127) {
 						text[len++]=(j1+1)/2+(j1<95?112:176);
 						text[len++]=j2+(j1%2?31+(j2/96):126);
+						if (del_flag && (text[len-1]&0xFF) == 0x7F) text[len-1]++;
 					}
-				} else
-					text[len++]=result;
-			} else {
-				ReadCharAttr(j,i,page,&result);
-				text[len++]=result;
+				} else if (j==c1&&c1>0) {
+                    uint16_t prevres=mem_readw(where-2);
+                    if (!((prevres & 0xFF00u) != 0u && (prevres & 0xFCu) != 0x08u && prevres==result))
+                        text[len++]=result;
+                } else
+                    text[len++]=result;
+            } else if (IS_DOSV) {
+                if (lead2) lead2=false;
+                else if (isKanji1(real_readb(seg,(i*c+j)*2))) lead2=true;
+                result=real_readb(seg,(i*c+j)*2);
+                text[len++]=result;
+                if (!lead2 && del_flag && (text[len-1]&0xFF) == 0x7F) text[len-1]++;
+                if (j==c2&&c2<c-1&&lead2) {
+                    result=real_readb(seg,(i*c+j+1)*2);
+                    text[len++]=result;
+                    if (del_flag && (text[len-1]&0xFF) == 0x7F) text[len-1]++;
+                }
+            } else {
+                bool find = isJEGAEnabled()?std::find(jtbs.begin(), jtbs.end(), std::make_pair(i,j)) != jtbs.end():false;
+                if (!isJEGAEnabled()||j>c1||!find) {
+                    ReadCharAttr(j,i,page,&result);
+                    text[len++]=result;
+                    if (isJEGAEnabled() && find && del_flag && (text[len-1]&0xFF) == 0x7F) text[len-1]++;
+                }
+                if (isJEGAEnabled()&&j==c2&&c2<c-1&&std::find(jtbs.begin(), jtbs.end(), std::make_pair(i,j+1)) != jtbs.end()) {
+                    ReadCharAttr(j+1,i,page,&result);
+                    text[len++]=result;
+                    if (del_flag && (text[len-1]&0xFF) == 0x7F) text[len-1]++;
+                }
 			}
 		}
 		while (len>0&&text[len-1]==32) text[--len]=0;
@@ -758,22 +838,29 @@ const char* Mouse_GetSelected(int x1, int y1, int x2, int y2, int w, int h, uint
 			text[len++]='\n';
 		}
 	}
-	text[len]=0;
+    text[len] = 0;
 	*textlen=len;
 	return text;
 }
 
-void Mouse_Select(int x1, int y1, int x2, int y2, int w, int h) {
-	uint8_t page = real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
-	uint16_t c=0, r=0;
-	if (IS_PC98_ARCH) {
-		c=80;
-		r=real_readb(0x60,0x113) & 0x01 ? 25 : 20;
-	} else {
-		c=real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS);
-		r=(uint16_t)real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
-	}
-	int c1=c*x1/w, r1=r*y1/h, c2=c*x2/w, r2=r*y2/h, t;
+#if defined(WIN32) || defined(MACOSX) || defined(C_SDL2)
+void Mouse_Select(int x1, int y1, int x2, int y2, int w, int h, bool select) {
+    int c1=x1, r1=y1, c2=x2, r2=y2, t;
+    uint8_t page = real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
+    uint16_t c=0, r=0;
+    if (IS_PC98_ARCH) {
+        c=80;
+        r=real_readb(0x60,0x113) & 0x01 ? 25 : 20;
+    } else {
+        c=real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS);
+        r=(uint16_t)(IS_EGAVGA_ARCH?real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS):24)+1;
+    }
+    if (w>0&&h>0) {
+        c1=c*x1/w;
+        r1=r*y1/h;
+        c2=c*x2/w;
+        r2=r*y2/h;
+    }
 	if (c1>c2) {
 		t=c1;
 		c1=c2;
@@ -784,44 +871,30 @@ void Mouse_Select(int x1, int y1, int x2, int y2, int w, int h) {
 		r1=r2;
 		r2=t;
 	}
+    uint16_t seg = IS_DOSV?GetTextSeg():0;
+#if defined(USE_TTF)
+    if (ttf.inUse&&(!IS_EGAVGA_ARCH||CurMode->mode!=3||isDBCSCP())) {
+        ttf_cell *newAC = newAttrChar;
+        for (int y = 0; y < ttf.lins; y++) {
+            if (y>=r1&&y<=r2)
+                for (int x = 0; x < ttf.cols; x++)
+                    if ((x>=c1||((IS_PC98_ARCH||isDBCSCP())&&c1>0&&x==c1-1&&(newAC[x].chr&0xFF00)&&(newAC[x+1].chr&0xFF)==32))&&x<=c2)
+                        newAC[x].selected = select?1:0;
+            newAC += ttf.cols;
+        }
+        void GFX_EndTextLines(bool force=false);
+        GFX_EndTextLines();
+    } else
+#endif
 	for (int i=r1; i<=r2; i++)
 		for (int j=c1; j<=c2; j++) {
 			if (IS_PC98_ARCH) {
-				uint16_t address=((i*80)+j)*2;
-				PhysPt where = CurMode->pstart+address;
-                mem_writeb(where+0x2000,mem_readb(where+0x2000)^16);
-			} else
-				real_writeb(0xb800,(i*c+j)*2+1,real_readb(0xb800,(i*c+j)*2+1)^119);
-		}
-}
-
-void Restore_Text(int x1, int y1, int x2, int y2, int w, int h) {
-	uint8_t page = real_readb(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
-	uint16_t c=0, r=0;
-	if (IS_PC98_ARCH) {
-		c=80;
-		r=real_readb(0x60,0x113) & 0x01 ? 25 : 20;
-	} else {
-		c=real_readw(BIOSMEM_SEG,BIOSMEM_NB_COLS);
-		r=(uint16_t)real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
-	}
-	int c1=c*x1/w, r1=r*y1/h, c2=c*x2/w, r2=r*y2/h, t;
-	if (c1>c2) {
-		t=c1;
-		c1=c2;
-		c2=t;
-	}
-	if (r1>r2) {
-		t=r1;
-		r1=r2;
-		r2=t;
-	}
-	for (int i=r1; i<=r2; i++)
-		for (int j=c1; j<=c2; j++) {
-			if (IS_PC98_ARCH) {
-				uint16_t address=((i*80)+j)*2;
-				PhysPt where = CurMode->pstart+address;
-                mem_writeb(where+0x2000,mem_readb(where+0x2000)^16);
+				PhysPt where = CurMode->pstart+((i*80)+j)*2;
+				mem_writeb(where+0x2000,mem_readb(where+0x2000)^16);
+			} else if (IS_DOSV) {
+				uint8_t attr = real_readb(seg,(i*c+j)*2+1);
+				real_writeb(seg,(i*c+j)*2+1,attr/0x10+(attr&0xF)*0x10);
+				if (j==c2) WriteCharTopView(c*i*2,j+1);
 			} else
 				real_writeb(0xb800,(i*c+j)*2+1,real_readb(0xb800,(i*c+j)*2+1)^119);
 		}
@@ -995,10 +1068,8 @@ void Mouse_BeforeNewVideoMode(bool setmode) {
 
     if (CurMode->type!=M_TEXT) RestoreCursorBackground();
     else RestoreCursorBackgroundText();
-    if (!mouse.hidden) {
-        mouse.hidden = 1;
-        mouse.hidden_at = PIC_FullIndex();
-    }
+    if (!mouse.hidden) mouse.hidden_at = PIC_FullIndex();
+    mouse.hidden = 1;
     mouse.oldhidden = 1;
     mouse.background = false;
 }
@@ -1043,7 +1114,7 @@ void Mouse_AfterNewVideoMode(bool setmode) {
             mouse.max_y = 400 - 1;
         }
         else {
-            Bitu rows = real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS);
+            Bitu rows = IS_EGAVGA_ARCH?real_readb(BIOSMEM_SEG,BIOSMEM_NB_ROWS):24;
             if ((rows == 0) || (rows > 250)) rows = 25 - 1;
             mouse.max_y = 8*(rows+1) - 1;
         }
@@ -1155,35 +1226,55 @@ static void Mouse_Reset(void) {
     mouse.in_UIR = false;
 }
 
+void Mouse_Used(void) {
+    static bool autolock_enabled = false;
+    if(!autolock_enabled) {
+        Mouse_AutoLock(true);
+        autolock_enabled = true;
+    }
+}
+
+void Mouse_Read_Motion_Data() {
+    const auto locked = MOUSE_IsLocked();
+    reg_cx = (uint16_t)static_cast<int16_t>(locked ? mouse.mickey_x : 0);
+    reg_dx = (uint16_t)static_cast<int16_t>(locked ? mouse.mickey_y : 0);
+    mouse.mickey_x = 0;
+    mouse.mickey_y = 0;
+    Mouse_Used();
+}
+
+// Reference: http://www.ctyme.com/intr/int-33.htm
 static Bitu INT33_Handler(void) {
 //  LOG(LOG_MOUSE,LOG_NORMAL)("MOUSE: %04X %X %X %d %d",reg_ax,reg_bx,reg_cx,POS_X,POS_Y);
     switch (reg_ax) {
-    case 0x00:  /* Reset Driver and Read Status */
+    case 0x00:  /* MS MOUSE - RESET DRIVER AND READ STATUS */
         Mouse_ResetHardware();
         goto software_reset;
-    case 0x01:  /* Show Mouse */
+    case 0x01:  /* MS MOUSE v1.0+ - SHOW MOUSE CURSOR */
         if (mouse.hidden) mouse.hidden--;
         mouse.updateRegion_y[1] = -1; //offscreen
-        Mouse_AutoLock(true);
         DrawCursor();
+        if(!mouse.hidden) Mouse_Used();
         break;
-    case 0x02:  /* Hide Mouse */
+    case 0x02:  /* MS MOUSE v1.0+ - HIDE MOUSE CURSOR */
         {
             if (CurMode->type != M_TEXT) RestoreCursorBackground();
             else RestoreCursorBackgroundText();
             if (mouse.hidden == 0) mouse.hidden_at = PIC_FullIndex();
             mouse.hidden++;
+            if (usesystemcursor&&!MOUSE_IsLocked()) SDL_ShowCursor(SDL_DISABLE);
         }
         break;
-    case 0x03:  /* Return position and Button Status */
+    case 0x03:  /* MS MOUSE v1.0+ - RETURN POSITION AND BUTTON STATUS */
         reg_bx = mouse.buttons;
         reg_cx = (uint16_t)POS_X;
         reg_dx = (uint16_t)POS_Y;
         mouse.first_range_setx = false;
         mouse.first_range_sety = false;
         if (en_int33_hide_if_polling) int33_last_poll = PIC_FullIndex();
+        Mouse_Used();
         break;
-    case 0x04:  /* Position Mouse */
+    case 0x04:  /* MS MOUSE v1.0+ - POSITION MOUSE CURSOR */
         /* If position isn't different from current position
          * don't change it then. (as position is rounded so numbers get
          * lost when the rounded number is set) (arena/simulation Wolf) */
@@ -1197,7 +1288,7 @@ static Bitu INT33_Handler(void) {
         DrawCursor();
         if (en_int33_hide_if_polling) int33_last_poll = PIC_FullIndex();
         break;
-    case 0x05:  /* Return Button Press Data */
+    case 0x05:  /* MS MOUSE v1.0+ - RETURN BUTTON PRESS DATA */
         {
             uint16_t but = reg_bx;
             reg_ax = mouse.buttons;
@@ -1207,9 +1298,10 @@ static Bitu INT33_Handler(void) {
             reg_bx = mouse.times_pressed[but];
             mouse.times_pressed[but] = 0;
             if (en_int33_hide_if_polling) int33_last_poll = PIC_FullIndex();
-            break;
         }
-    case 0x06:  /* Return Button Release Data */
+        Mouse_Used();
+        break;
+    case 0x06:  /* MS MOUSE v1.0+ - RETURN BUTTON RELEASE DATA */
         {
             uint16_t but = reg_bx;
             reg_ax = mouse.buttons;
@@ -1219,9 +1311,10 @@ static Bitu INT33_Handler(void) {
             reg_bx = mouse.times_released[but];
             mouse.times_released[but] = 0;
             if (en_int33_hide_if_polling) int33_last_poll = PIC_FullIndex();
-            break;
         }
-    case 0x07:  /* Define horizontal cursor range */
+        Mouse_Used();
+        break;
+    case 0x07:  /* MS MOUSE v1.0+ - DEFINE HORIZONTAL CURSOR RANGE */
         {
             //Lemmings sets 1-640 and wants that. Ironseed sets 0-640 but doesn't like 640
             //Ironseed works if newvideo mode with mode 13 sets 0-639
@@ -1265,7 +1358,7 @@ static Bitu INT33_Handler(void) {
                     else {
                         // Apply sanity rounding.
                         //
-                        // Daggerfall: Sets max to 310 instead of 320, probably to prevent drawing the cursor
+                        // Daggerfall: Sets max x to 310 instead of 320, probably to prevent drawing the cursor
                         //             partially offscreen. */
                         if (nval >= ((int16_t)CurMode->swidth - 32) && nval <= ((int16_t)CurMode->swidth + 32))
                             nval = (int16_t)CurMode->swidth;
@@ -1282,7 +1375,7 @@ static Bitu INT33_Handler(void) {
             }
         }
         break;
-    case 0x08:  /* Define vertical cursor range */
+    case 0x08:  /* MS MOUSE v1.0+ - DEFINE VERTICAL CURSOR RANGE */
         {
             // Not sure what to take instead of the CurMode (see case 0x07 as well)
             // especially the cases where sheight= 400 and we set it with the mouse_reset to 200
@@ -1326,7 +1419,7 @@ static Bitu INT33_Handler(void) {
                     else {
                         // Apply sanity rounding.
                         //
-                        // Daggerfall: Sets max to 310 instead of 320, probably to prevent drawing the cursor
+                        // Daggerfall: Sets max x to 310 instead of 320, probably to prevent drawing the cursor
                         //             partially offscreen. */
                         if (nval >= ((int16_t)CurMode->sheight - 32) && nval <= ((int16_t)CurMode->sheight + 32))
                             nval = (int16_t)CurMode->sheight;
@@ -1343,7 +1436,7 @@ static Bitu INT33_Handler(void) {
             }
         }
         break;
-    case 0x09:  /* Define GFX Cursor */
+    case 0x09:  /* MS MOUSE v3.0+ - DEFINE GRAPHICS CURSOR */
         {
             PhysPt src = SegPhys(es) + reg_dx;
             MEM_BlockRead(src, userdefScreenMask, CURSORY * 2);
@@ -1356,7 +1449,7 @@ static Bitu INT33_Handler(void) {
             DrawCursor();
             break;
         }
-    case 0x0a:  /* Define Text Cursor */
+    case 0x0a:  /* MS MOUSE v3.0+ - DEFINE TEXT CURSOR */
         mouse.cursorType = (reg_bx ? 1 : 0);
         mouse.textAndMask = reg_cx;
         mouse.textXorMask = reg_dx;
@@ -1366,46 +1459,42 @@ static Bitu INT33_Handler(void) {
         }
         DrawCursor();
         break;
-    case 0x0b:  /* Read Motion Data */
-        {
-            extern bool MOUSE_IsLocked();
-            const auto locked = MOUSE_IsLocked();
-            reg_cx = (uint16_t)static_cast<int16_t>(locked ? mouse.mickey_x : 0);
-            reg_dx = (uint16_t)static_cast<int16_t>(locked ? mouse.mickey_y : 0);
-            mouse.mickey_x = 0;
-            mouse.mickey_y = 0;
-            break;
-        }
-    case 0x0c:  /* Define interrupt subroutine parameters */
+    case 0x0b:  /* MS MOUSE v1.0+ - READ MOTION COUNTERS */
+        Mouse_Read_Motion_Data();
+        break;
+    case 0x0c:  /* MS MOUSE v1.0+ - DEFINE INTERRUPT SUBROUTINE PARAMETERS */
         mouse.sub_mask = reg_cx;
         mouse.sub_seg = SegValue(es);
         mouse.sub_ofs = reg_dx;
-        Mouse_AutoLock(true); //Some games don't seem to reset the mouse before using
+        if(mouse.sub_mask) Mouse_Used();
         break;
-    case 0x0d:  /* Mouse light pen emulation on */
+    case 0x0d:  /* MS MOUSE v1.0+ - LIGHT PEN EMULATION ON */
         LOG(LOG_MOUSE, LOG_ERROR)("Mouse light pen emulation on not implemented");
         break;
-    case 0x0e:  /* Mouse light pen emulation off */
+    case 0x0e:  /* MS MOUSE v1.0+ - LIGHT PEN EMULATION OFF */
         LOG(LOG_MOUSE, LOG_ERROR)("Mouse light pen emulation off not implemented");
         break;
-    case 0x0f:  /* Define mickey/pixel rate */
+    case 0x0f:  /* MS MOUSE v1.0+ - DEFINE MICKEY/PIXEL RATIO */
         Mouse_SetMickeyPixelRate((int16_t)reg_cx, (int16_t)reg_dx);
         break;
-    case 0x10:  /* Define screen region for updating */
+    case 0x10:  /* MS MOUSE v1.0+ - DEFINE SCREEN REGION FOR UPDATING */
         mouse.updateRegion_x[0] = (int16_t)reg_cx;
         mouse.updateRegion_y[0] = (int16_t)reg_dx;
         mouse.updateRegion_x[1] = (int16_t)reg_si;
         mouse.updateRegion_y[1] = (int16_t)reg_di;
         DrawCursor();
         break;
-    case 0x11:      /* Get number of buttons */
+    case 0x11:  /* Genius Mouse 9.06 - GET NUMBER OF BUTTONS */
         reg_ax = 0xffff;
         reg_bx = MOUSE_BUTTONS;
         break;
-    case 0x13:      /* Set double-speed threshold */
-        mouse.doubleSpeedThreshold = (reg_bx ? reg_bx : 64);
+    case 0x12:  /* MS MOUSE - SET LARGE GRAPHICS CURSOR BLOCK */
+        LOG(LOG_MOUSE, LOG_ERROR)("Set large graphics cursor block not implemented");
         break;
-    case 0x14: /* Exchange event-handler */
+    case 0x13:  /* MS MOUSE v5.0+ - DEFINE DOUBLE-SPEED THRESHOLD */
+        mouse.doubleSpeedThreshold = (reg_dx ? reg_dx : 64);
+        break;
+    case 0x14:  /* MS MOUSE v3.0+ - EXCHANGE INTERRUPT SUBROUTINES */
         {
             uint16_t oldSeg = mouse.sub_seg;
             uint16_t oldOfs = mouse.sub_ofs;
@@ -1420,78 +1509,76 @@ static Bitu INT33_Handler(void) {
             SegSet16(es, oldSeg);
         }
         break;
-    case 0x15: /* Get Driver storage space requirements */
+    case 0x15:  /* MS MOUSE v6.0+ - RETURN DRIVER STORAGE REQUIREMENTS */
         reg_bx = sizeof(mouse);
         break;
-    case 0x16: /* Save driver state */
+    case 0x16:  /* MS MOUSE v6.0+ - SAVE DRIVER STATE */
         {
             LOG(LOG_MOUSE, LOG_NORMAL)("Saving driver state...");
             PhysPt dest = SegPhys(es) + reg_dx;
             MEM_BlockWrite(dest, &mouse, sizeof(mouse));
         }
         break;
-    case 0x17: /* load driver state */
+    case 0x17:  /* MS MOUSE v6.0+ - RESTORE DRIVER STATE */
         {
             LOG(LOG_MOUSE, LOG_NORMAL)("Loading driver state...");
             PhysPt src = SegPhys(es) + reg_dx;
             MEM_BlockRead(src, &mouse, sizeof(mouse));
             break;
         }
-    case 0x18: /* Set alternate subroutine call mask and address */
-        LOG(LOG_MOUSE, LOG_ERROR)("Set alternate subroutine call mask and address not implemented");
+    case 0x18:  /* MS MOUSE v6.0+ - SET ALTERNATE MOUSE USER HANDLER */
+        LOG(LOG_MOUSE, LOG_ERROR)("Set alternate mouse user handler not implemented");
         break;
-    case 0x19: /* Get user alternate interrupt address*/
-        LOG(LOG_MOUSE, LOG_ERROR)("Get user alternate interrupt address not implemented");
+    case 0x19:  /* MS MOUSE v6.0+ - RETURN USER ALTERNATE INTERRUPT VECTOR */
+        LOG(LOG_MOUSE, LOG_ERROR)("Return user alternate interrupt vector not implemented");
         break;
-    case 0x1a:  /* Set mouse sensitivity */
+    case 0x1a:  /* MS MOUSE v6.0+ - SET MOUSE SENSITIVITY */
         // ToDo : double mouse speed value
         Mouse_SetSensitivity(reg_bx, reg_cx, reg_dx);
         LOG(LOG_MOUSE, LOG_NORMAL)("Set sensitivity used with %d %d (%d)", reg_bx, reg_cx, reg_dx);
         break;
-    case 0x1b:  /* Get mouse sensitivity */
+    case 0x1b:  /* MS MOUSE v6.0+ - RETURN MOUSE SENSITIVITY */
         reg_bx = mouse.senv_x_val;
         reg_cx = mouse.senv_y_val;
         reg_dx = mouse.dspeed_val;
 
         LOG(LOG_MOUSE, LOG_NORMAL)("Get sensitivity %d %d", reg_bx, reg_cx);
         break;
-    case 0x1c:  /* Set interrupt rate */
+    case 0x1c:  /* MS MOUSE v6.0+ - SET INTERRUPT RATE */
         /* Can't really set a rate this is host determined */
         break;
-    case 0x1d:      /* Set display page number */
+    case 0x1d:  /* MS MOUSE v6.0+ - DEFINE DISPLAY PAGE NUMBER */
         mouse.page = reg_bl;
         break;
-    case 0x1e:      /* Get display page number */
+    case 0x1e:  /* MS MOUSE v6.0+ - RETURN DISPLAY PAGE NUMBER */
         reg_bx = mouse.page;
         break;
-    case 0x1f:  /* Disable Mousedriver */
+    case 0x1f:  /* MS MOUSE v6.0+ - DISABLE MOUSE DRIVER */
         /* ES:BX old mouse driver Zero at the moment TODO */
         reg_bx = 0;
         SegSet16(es, 0);
         mouse.enabled = false; /* Just for reporting not doing a thing with it */
         mouse.oldhidden = mouse.hidden;
-        if (!mouse.hidden) {
-            mouse.hidden = 1;
-            mouse.hidden_at = PIC_FullIndex();
-        }
+        if (!mouse.hidden) mouse.hidden_at = PIC_FullIndex();
+        mouse.hidden = 1;
         break;
-    case 0x20:  /* Enable Mousedriver */
+    case 0x20:  /* MS MOUSE v6.0+ - ENABLE MOUSE DRIVER */
         mouse.enabled = true;
         mouse.hidden = mouse.oldhidden;
         break;
-    case 0x21:  /* Software Reset */
+    case 0x21:  /* MS MOUSE v6.0+ - SOFTWARE RESET */
     software_reset:
         extern bool Mouse_Drv;
         if (Mouse_Drv) {
             reg_ax = 0xffff;
             reg_bx = MOUSE_BUTTONS;
             Mouse_Reset();
-            Mouse_AutoLock(true);
+            Mouse_Used();
             AUX_INT33_Takeover();
             LOG(LOG_MOUSE, LOG_NORMAL)("INT 33h reset");
         }
         break;
-    case 0x22:      /* Set language for messages */
+    case 0x22:  /* MS MOUSE v6.0+ - SET LANGUAGE FOR MESSAGES */
             /*
              *                        Values for mouse driver language:
              *
@@ -1508,32 +1595,111 @@ static Bitu INT33_Handler(void) {
              */
         mouse.language = reg_bx;
         break;
-    case 0x23:      /* Get language for messages */
+    case 0x23:  /* MS MOUSE v6.0+ - GET LANGUAGE FOR MESSAGES */
         reg_bx = mouse.language;
         break;
-    case 0x24:  /* Get Software version and mouse type */
+    case 0x24:  /* MS MOUSE v6.26+ - GET SOFTWARE VERSION, MOUSE TYPE, AND IRQ NUMBER */
+	// TODO: If PC-98 mode, return type 0x01 bus mouse, and bus mouse IRQ.
+	//       If emulating a serial mouse, return type 0x02 and serial port IRQ.
         reg_bx = 0x805;   //Version 8.05 woohoo 
         reg_ch = 0x04;    /* PS/2 type */
-        reg_cl = 0;       /* PS/2 (unused) */
+        reg_cl = 0;       /* PS/2 mouse == 0, for any other type, this would be the IRQ */
         break;
-    case 0x26: /* Get Maximum virtual coordinates */
+    case 0x25:  /* MS MOUSE v6.26+ - GET GENERAL DRIVER INFORMATION */
+	// TODO: According to PC sourcebook reference
+	//       Returns:
+	//       AH = status
+	//         bit 7 driver type: 1=sys 0=com
+	//         bit 6: 0=non-integrated 1=integrated mouse driver
+	//         bits 4-5: cursor type  00=software text cursor 01=hardware text cursor 1X=graphics cursor
+	//         bits 0-3: Function 28 mouse interrupt rate
+	//       AL = Number of MDDS (?)
+	//       BX = fCursor lock
+	//       CX = FinMouse code
+	//       DX = fMouse busy
+        LOG(LOG_MOUSE, LOG_ERROR)("Get general driver information not implemented");
+        break;
+    case 0x26:  /* MS MOUSE v6.26+ - GET MAXIMUM VIRTUAL COORDINATES */
         reg_bx = (mouse.enabled ? 0x0000 : 0xffff);
         reg_cx = (uint16_t)mouse.max_x;
         reg_dx = (uint16_t)mouse.max_y;
         break;
-    case 0x2a:  /* Get cursor hot spot */
+    case 0x27:  /* MS MOUSE v7.01+ - GET SCREEN/CURSOR MASKS AND MICKEY COUNTS */
+        reg_ax = mouse.textAndMask;
+        reg_bx = mouse.textXorMask;
+        Mouse_Read_Motion_Data();
+        break;
+    case 0x28:  /* MS MOUSE v7.0+ - SET VIDEO MODE */
+	// TODO: According to PC sourcebook
+	//       Entry:
+	//       CX = Requested video mode
+	//       DX = Font size, 0 for default
+	//       Returns:
+	//       DX = 0 on success, nonzero (requested video mode) if not
+        LOG(LOG_MOUSE, LOG_ERROR)("Set video mode not implemented");
+        break;
+    case 0x29: /* MS MOUSE v7.0+ - ENUMERATE VIDEO MODES */
+	// TODO: According to PC sourcebook
+	//       Entry:
+	//       CX = 0 for first, != 0 for next
+	//       Exit:
+	//       BX:DX = named string far ptr
+	//       CX = video mode number
+        LOG(LOG_MOUSE, LOG_ERROR)("Enumerate video modes not implemented");
+        break;
+    case 0x2a:  /* MS MOUSE v7.02+ - GET CURSOR HOT SPOT */
+        // NTS: Whoever compiled the INT 33 list in the Microsoft PC sourcebook made a big mistake
+        //      here and started counting hexadecimal as decimal. This function is wrongly listed as INT 33h AH=30H.
+        //      You can tell by the summary list which shows both hexadecimal and decimal side by side, this is
+        //      listed as AH=30h (42). Decimal 42 as hexadecimal would be 0x2A (0x20 + 0x0A) == (32 + 10) == 42 == 0x2A.
         reg_al = (uint8_t)-mouse.hidden;    // Microsoft uses a negative byte counter for cursor visibility
         reg_bx = (uint16_t)mouse.hotx;
         reg_cx = (uint16_t)mouse.hoty;
         reg_dx = 0x04;    // PS/2 mouse type
         break;
-    case 0x31: /* Get Current Minimum/Maximum virtual coordinates */
+    case 0x2b:  /* MS MOUSE v7.0+ - LOAD ACCELERATION PROFILES */
+        LOG(LOG_MOUSE, LOG_ERROR)("Load acceleration profiles not implemented");
+        break;
+    case 0x2c:  /* MS MOUSE v7.0+ - GET ACCELERATION PROFILES */
+        LOG(LOG_MOUSE, LOG_ERROR)("Get acceleration profiles not implemented");
+        break;
+    case 0x2d:  /* MS MOUSE v7.0+ - SELECT ACCELERATION PROFILE */
+        LOG(LOG_MOUSE, LOG_ERROR)("Select acceleration profile not implemented");
+        break;
+    case 0x2e:  /* MS MOUSE v8.10+ - SET ACCELERATION PROFILE NAMES */
+        LOG(LOG_MOUSE, LOG_ERROR)("Set acceleration profile names not implemented");
+        break;
+    case 0x2f:  /* MS MOUSE v7.02+ - MOUSE HARDWARE RESET */
+        LOG(LOG_MOUSE, LOG_ERROR)("INT 33 AX=2F mouse hardware reset not implemented");
+        break;
+    case 0x30:  /* MS MOUSE v7.04+ - GET/SET BallPoint INFORMATION */
+        LOG(LOG_MOUSE, LOG_ERROR)("Get/set BallPoint information not implemented");
+        break;
+    case 0x31:  /* MS MOUSE v7.05+ - GET CURRENT MINIMUM/MAXIMUM VIRTUAL COORDINATES */
         reg_ax = (uint16_t)mouse.min_x;
         reg_bx = (uint16_t)mouse.min_y;
         reg_cx = (uint16_t)mouse.max_x;
         reg_dx = (uint16_t)mouse.max_y;
         break;
-    case 0x53C1: /* Logitech CyberMan */
+    case 0x32:  /* MS MOUSE v7.05+ - GET ACTIVE ADVANCED FUNCTIONS */
+        LOG(LOG_MOUSE, LOG_ERROR)("Get active advanced functions not implemented");
+        break;
+    case 0x33:  /* MS MOUSE v7.05+ - GET SWITCH SETTINGS AND ACCELERATION PROFILE DATA */
+        LOG(LOG_MOUSE, LOG_ERROR)("Get switch settings and acceleration profile data not implemented");
+        break;
+    case 0x34:  /* MS MOUSE v8.0+ - GET INITIALIZATION FILE */
+        LOG(LOG_MOUSE, LOG_ERROR)("Get initialization file not implemented");
+        break;
+    case 0x35:  /* MS MOUSE v8.10+ - LCD SCREEN LARGE POINTER SUPPORT */
+        LOG(LOG_MOUSE, LOG_ERROR)("LCD screen large pointer support not implemented");
+        break;
+    case 0x4d:  /* MS MOUSE - RETURN POINTER TO COPYRIGHT STRING */
+        LOG(LOG_MOUSE, LOG_ERROR)("Return pointer to copyright string not implemented");
+        break;
+    case 0x6d:  /* MS MOUSE - GET VERSION STRING */
+        LOG(LOG_MOUSE, LOG_ERROR)("Get version string not implemented");
+        break;
+    case 0x53C1:  /* Logitech CyberMan */
         LOG(LOG_MOUSE, LOG_NORMAL)("Mouse function 53C1 for Logitech CyberMan called. Ignored by regular mouse driver.");
         break;
     default:
@@ -1805,10 +1971,8 @@ void MOUSE_Startup(Section *sec) {
 
     memset(&mouse,0,sizeof(mouse));
 
-    if (!mouse.hidden) {
-        mouse.hidden = 1;
-        mouse.hidden_at = PIC_FullIndex();
-    }
+    if (!mouse.hidden) mouse.hidden_at = PIC_FullIndex();
+    mouse.hidden = 1;
 
     mouse.timer_in_progress = false;
     mouse.mode = 0xFF; //Non existing mode
