@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2020  The DOSBox Team
+ *  Copyright (C) 2002-2021  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,9 +19,12 @@
 
 #include "dosbox.h"
 #include "cross.h"
+#include "logging.h"
 #include "setup.h"
 #include "control.h"
 #include "support.h"
+
+#include <assert.h>
 #include <fstream>
 #include <string>
 #include <sstream>
@@ -93,7 +96,7 @@ Value::operator char const* () const {
     return _string->c_str();
 }
 
-bool Value::operator==(Value const& other) {
+bool Value::operator==(Value const& other) const {
     if (this == &other) return true;
     if (type != other.type) return false;
     switch(type) {
@@ -235,7 +238,7 @@ string Value::ToString() const {
 
 bool Property::CheckValue(Value const& in, bool warn) {
     if (suggested_values.empty()) return true;
-    for(iter it = suggested_values.begin();it != suggested_values.end();++it) {
+    for(const_iter it = suggested_values.begin();it != suggested_values.end();++it) {
         if ( (*it) == in) { //Match!
             return true;
         }
@@ -377,11 +380,12 @@ bool Prop_string::SetValue(std::string const& input) {
     //If there are none then it can be paths and such which are case sensitive
     if (!suggested_values.empty()) lowcase(temp);
     Value val(temp,Value::V_STRING);
-    return SetVal(val,false,true);
+    return SetVal(val,false,true)||(!suggested_values.empty()&&!input.size());
 }
+
 bool Prop_string::CheckValue(Value const& in, bool warn) {
     if (suggested_values.empty()) return true;
-    for(iter it = suggested_values.begin();it != suggested_values.end();it++) {
+    for(const_iter it = suggested_values.begin();it != suggested_values.end();++it) {
         if ( (*it) == in) { //Match!
             return true;
         }
@@ -403,7 +407,7 @@ bool Prop_path::SetValue(std::string const& input) {
     bool retval = SetVal(val,false,true);
 
     if (input.empty()) {
-        realpath = "";
+        realpath.clear();
         return false;
     }
     std::string workcopy(input);
@@ -419,7 +423,7 @@ bool Prop_path::SetValue(std::string const& input) {
 bool Prop_bool::SetValue(std::string const& input) {
     Value val;
     if (!val.SetValue(input.size()?input:default_value.ToString(),Value::V_BOOL)) return false;
-    return SetVal(val,false,/*warn*/true)&&input.size();
+    return SetVal(val,false,/*warn*/true);
 }
 
 bool Prop_hex::SetValue(std::string const& input) {
@@ -436,7 +440,7 @@ void Prop_multival::make_default_value() {
     std::string result = p->Get_Default_Value().ToString();
     while( (p = section->Get_prop(i++)) ) {
         std::string props = p->Get_Default_Value().ToString();
-        if (props == "") continue;
+        if (props.empty()) continue;
         result += separator; result += props;
     }
     Value val(result,Value::V_STRING);
@@ -473,7 +477,7 @@ bool Prop_multival_remain::SetValue(std::string const& input,bool init) {
             local.erase(0,loc+1);
         } else if (local.size()) { //last argument or last property
             in = local;
-            local = "";
+            local.clear();
         }
 
         if (p->Get_type() == Value::V_STRING) {
@@ -527,7 +531,7 @@ bool Prop_multival::SetValue(std::string const& input,bool init) {
             local.erase(0,loc+1);
         } else if (local.size()) { //last argument
             in = local;
-            local = "";
+            local.clear();
         }
         //Test Value. If it fails set default
         Value valtest (in,p->Get_type());
@@ -624,8 +628,8 @@ int Section_prop::Get_int(string const&_propname) const {
 }
 
 bool Section_prop::Get_bool(string const& _propname) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             return ((*tel)->GetValue());
         }
     }
@@ -633,8 +637,8 @@ bool Section_prop::Get_bool(string const& _propname) const {
 }
 
 double Section_prop::Get_double(string const& _propname) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             return ((*tel)->GetValue());
         }
     }
@@ -652,8 +656,8 @@ Prop_path* Section_prop::Get_path(string const& _propname) const {
 }
 
 Prop_multival* Section_prop::Get_multival(string const& _propname) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             Prop_multival* val = dynamic_cast<Prop_multival*>((*tel));
             if (val) return val; else return NULL;
         }
@@ -662,8 +666,8 @@ Prop_multival* Section_prop::Get_multival(string const& _propname) const {
 }
 
 Prop_multival_remain* Section_prop::Get_multivalremain(string const& _propname) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             Prop_multival_remain* val = dynamic_cast<Prop_multival_remain*>((*tel));
             if (val) return val; else return NULL;
         }
@@ -672,15 +676,15 @@ Prop_multival_remain* Section_prop::Get_multivalremain(string const& _propname)
 }
 
 Property* Section_prop::Get_prop(int index) {
-    for(it tel=properties.begin();tel!=properties.end();++tel) {
+    for(it tel = properties.begin();tel != properties.end();++tel){
         if (!index--) return (*tel);
     }
     return NULL;
 }
 
 Property* Section_prop::Get_prop(string const& _propname) {
-    for(it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             return (*tel);
         }
     }
@@ -688,16 +692,16 @@ Property* Section_prop::Get_prop(string const& _propname) {
 }
 
 const char* Section_prop::Get_string(string const& _propname) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             return ((*tel)->GetValue());
         }
     }
     return "";
 }
 Hex Section_prop::Get_hex(string const& _propname) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
-        if ((*tel)->propname==_propname) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
+        if ((*tel)->propname == _propname){
             return ((*tel)->GetValue());
         }
     }
@@ -738,23 +742,26 @@ void Section_prop::PrintData(FILE* outfile,int everything,bool norem) {
     size_t len = 0;
     // Determine maximum length of the props in this section
     for(const_it tel = properties.begin();tel != properties.end();++tel) {
-        if (!(everything>0 || everything==-1 && ((*tel)->basic() || (*tel)->modified()) || !everything && ((!norem && (*tel)->propname == "rem" && (!strcasecmp(GetName(), "4dos") || !strcasecmp(GetName(), "config"))) || (*tel)->modified()))) continue;
+        if (!(everything>0 || (everything==-1 && ((*tel)->basic() || (*tel)->modified())) || (!everything && ((!norem && (*tel)->propname == "rem" && (!strcasecmp(GetName(), "4dos") || !strcasecmp(GetName(), "config"))) || (*tel)->modified())))) continue;
 
         if ((*tel)->propname.length() > len)
             len = (*tel)->propname.length();
     }
 	if (!strcasecmp(GetName(), "config")&&len<11) len=11;
 
+    // fprintf wants int instead of size_t, also refuse to print sections with weirdly long properties.
+    int intlen = (len > 100) ? 100 : static_cast<int>(len);
+    if (intlen == 100) return;
     for(const_it tel = properties.begin();tel != properties.end();++tel) {
-        if (!(everything>0 || everything==-1 && ((*tel)->basic() || (*tel)->modified()) || !everything && ((!norem && (*tel)->propname == "rem" && (!strcasecmp(GetName(), "4dos") || !strcasecmp(GetName(), "config"))) || (*tel)->modified()))) continue;
+        if (!(everything>0 || (everything==-1 && ((*tel)->basic() || (*tel)->modified())) || (!everything && ((!norem && (*tel)->propname == "rem" && (!strcasecmp(GetName(), "4dos") || !strcasecmp(GetName(), "config"))) || (*tel)->modified())))) continue;
 
         std::string pre=everything==2&&!(*tel)->basic()?"#DOSBOX-X-ADV:":"";
-        fprintf(outfile,"%s%-*s = %s\n", pre.c_str(), (unsigned int)len, (*tel)->propname.c_str(), (*tel)->GetValue().ToString().c_str());
+        fprintf(outfile,"%s%-*s = %s\n", pre.c_str(), intlen, (*tel)->propname.c_str(), (*tel)->GetValue().ToString().c_str());
     }
 }
 
 string Section_prop::GetPropValue(string const& _property) const {
-    for(const_it tel=properties.begin();tel!=properties.end();++tel) {
+    for(const_it tel = properties.begin();tel != properties.end();++tel){
         if (!strcasecmp((*tel)->propname.c_str(),_property.c_str())) {
             return (*tel)->GetValue().ToString();
         }
@@ -783,19 +790,21 @@ string Section_line::GetPropValue(string const& /* _property*/) const {
     return NO_SUCH_PROPERTY;
 }
 
+#define HELPLINE_SIZE 256
 bool Config::PrintConfig(char const * const configfilename,int everything,bool norem) const {
-    char temp[50];char helpline[256];
-    FILE* outfile=fopen(configfilename,"w+t");
-    if (outfile==NULL) return false;
+    char temp[50];
+    char helpline[HELPLINE_SIZE] = { 0 };
+    FILE* outfile = fopen(configfilename,"w+t");
+    if (outfile == NULL) return false;
 
     /* Print start of configfile and add a return to improve readibility. */
     if (!norem) {
         fprintf(outfile,MSG_Get("CONFIGFILE_INTRO"),VERSION);
         fprintf(outfile,"\n");
     }
-    for (const_it tel=sectionlist.begin(); tel!=sectionlist.end(); ++tel) {
+    for (const_it tel = sectionlist.begin(); tel != sectionlist.end(); ++tel){
         /* Print out the Section header */
-        strcpy(temp,(*tel)->GetName());
+        safe_strncpy(temp,(*tel)->GetName(),sizeof(temp));
         lowcase(temp);
 
         Section_prop *sec = dynamic_cast<Section_prop *>(*tel);
@@ -804,7 +813,7 @@ bool Config::PrintConfig(char const * const configfilename,int everything,bool n
             Property *p;
             size_t i = 0, maxwidth = 0;
             while ((p = sec->Get_prop(int(i++)))) {
-                if (!(everything>0 || everything==-1 && (p->basic() || p->modified()) || !everything && (p->propname == "rem" && (!strcmp(temp, "4dos") || !strcmp(temp, "config")) || p->modified())))
+                if (!(everything>0 || (everything==-1 && (p->basic() || p->modified())) || (!everything && ((p->propname == "rem" && (!strcmp(temp, "4dos") || !strcmp(temp, "config"))) || p->modified()))))
                     continue;
 
                 size_t w = strlen(p->propname.c_str());
@@ -821,29 +830,35 @@ bool Config::PrintConfig(char const * const configfilename,int everything,bool n
 
             i=0;
             char prefix[80];
+            int intmaxwidth = (maxwidth > 60) ? 60 : static_cast<int>(maxwidth);
+            std::vector<std::string> advopts;
+            advopts.clear();
             if (!norem)
             while ((p = sec->Get_prop(int(i++)))) {
-                if (!(everything>0 || everything==-1 && (p->basic() || p->modified()) || !everything && (p->propname == "rem" && (!strcmp(temp, "4dos") || !strcmp(temp, "config")) || p->modified())))
-                    continue;
+                std::string help = p->Get_help();
+                if (!(everything==1 || (everything==-1 && (p->basic() || p->modified())) || (!everything && ((p->propname == "rem" && (!strcmp(temp, "4dos") || !strcmp(temp, "config"))) || p->modified())))) {
+                    if ((everything==-1 || everything==2) && !p->basic() && !p->modified() && help.size())
+                        advopts.push_back(p->propname);
+                    if (everything!=2) continue;
+                }
 
                 std::string pre=everything==2&&!p->basic()?"#DOSBOX-X-ADV:":"";
-                snprintf(prefix,80, "\n%s#%*s     ", pre.c_str(), (int)maxwidth, "");
-                std::string help = p->Get_help();
+                snprintf(prefix,80, "\n%s#%*s     ", pre.c_str(), intmaxwidth, "");
                 std::string::size_type pos = std::string::npos;
-                while ((pos = help.find("\n", pos+1)) != std::string::npos) {
+                while ((pos = help.find('\n', pos+1)) != std::string::npos) {
                     help.replace(pos, 1, prefix);
                 }
 
                 std::vector<Value> values = p->GetValues();
 
                 if (help != "" || !values.empty()) {
-                    fprintf(outfile, "%s# %*s: %s", pre.c_str(), (int)maxwidth, p->propname.c_str(), help.c_str());
+                    fprintf(outfile, "%s# %*s: %s", pre.c_str(), intmaxwidth, p->propname.c_str(), help.c_str());
 
                     if (!values.empty()) {
                         fprintf(outfile, "%s%s:", prefix, MSG_Get("CONFIG_SUGGESTED_VALUES"));
-                        std::vector<Value>::iterator it = values.begin();
+                        std::vector<Value>::const_iterator it = values.begin();
                         while (it != values.end()) {
-                            if ((*it).ToString() != "%u") { //Hack hack hack. else we need to modify GetValues, but that one is const...
+                            if ((*it).ToString() != "%u" && (strcmp(temp, "config") || p->propname != "numlock" || (*it).ToString() != "")) { //Hack hack hack. else we need to modify GetValues, but that one is const...
                                 if (it != values.begin()) fputs(",", outfile);
                                 fprintf(outfile, " %s", (*it).ToString().c_str());
                             }
@@ -854,29 +869,36 @@ bool Config::PrintConfig(char const * const configfilename,int everything,bool n
                     fprintf(outfile, "\n");
                 }
             }
+            if ((everything==-1 || everything==2) && !advopts.empty()) {
+                fprintf(outfile, everything==2?"#DOSBOX-X-ADV-SEE:#\n#DOSBOX-X-ADV-SEE:# %s:\n#DOSBOX-X-ADV-SEE:# ->":"#\n# %s:\n# ->", MSG_Get("CONFIG_ADVANCED_OPTION"));
+                for (std::vector<std::string>::iterator advopt = advopts.begin(); advopt != advopts.end(); ++advopt)
+                    fprintf(outfile, " %s%c", advopt->c_str(), advopt+1 >= advopts.end()?'\n':';');
+                fprintf(outfile, everything==2?"#DOSBOX-X-ADV-SEE:#\n":"#\n");
+            }
         } else {
             fprintf(outfile,"[%s]\n",temp);
 
             if (!norem) {
                 upcase(temp);
                 strcat(temp,"_CONFIGFILE_HELP");
-                const char * helpstr=MSG_Get(temp);
-                char * helpwrite=helpline;
-                while (*helpstr) {
-                    *helpwrite++=*helpstr;
+                const char * helpstr = MSG_Get(temp);
+                const char * linestart = helpstr;
+                char * helpwrite = helpline;
+                while (*helpstr && helpstr - linestart < HELPLINE_SIZE - 2) {
+                    *helpwrite++ = *helpstr;
                     if (*helpstr == '\n') {
-                        *helpwrite=0;
+                        *helpwrite = 0;
                         fprintf(outfile,"# %s",helpline);
-                        helpwrite=helpline;
-                    }
-                    helpstr++;
+                        helpwrite = helpline;
+                        linestart = ++helpstr;
+                    } else helpstr++;
                 }
             }
         }
 
         (*tel)->PrintData(outfile,everything,norem);
 		if (!strcmp(temp, "config")||!strcmp(temp, "4dos")) {
-			const char * extra = const_cast<char*>(sec->data.c_str());
+			const char * extra = sec->data.c_str();
 			bool used1=false, used2=false;
 			char linestr[CROSS_LEN+1], *lin=linestr;
 			if (extra&&strlen(extra)) {
@@ -906,7 +928,7 @@ bool Config::PrintConfig(char const * const configfilename,int everything,bool n
 			}
 			if (!strcmp(temp, "config")) {
 				if (everything&&!used1) {
-					fprintf(outfile, "%-11s = %s\n", "set path", "Z:\\");
+					fprintf(outfile, "%-11s = %s\n", "set path", "Z:\\;Z:\\SYSTEM;Z:\\BIN;Z:\\DOS;Z:\\4DOS;Z:\\DEBUG;Z:\\TEXTUTIL");
 					fprintf(outfile, "%-11s = %s\n", "set prompt", "$P$G");
 					fprintf(outfile, "%-11s = %s\n", "set temp", "");
 				}
@@ -947,7 +969,7 @@ Section_prop* Config::AddSection_prop(char const * const _name,void (*_initfunct
 
 Section_prop::~Section_prop() {
     /* Delete properties themself (properties stores the pointer of a prop */
-    for(it prop = properties.begin(); prop != properties.end(); prop++) delete (*prop);
+    for(it prop = properties.begin(); prop != properties.end(); ++prop) delete (*prop);
     properties.clear();
 }
 
@@ -963,14 +985,14 @@ void Null_Init(Section *sec);
 
 void AddExitFunction(SectionFunction func,const char *name,bool canchange) {
     /* NTS: Add functions so that iterating front to back executes them in First In Last Out order. */
-    exitfunctions.push_front(Function_wrapper(func,canchange,name));
+    exitfunctions.emplace_front(Function_wrapper(func,canchange,name));
 }
 
 void AddVMEventFunction(enum vm_event event,SectionFunction func,const char *name,bool canchange) {
     assert(event < VM_EVENT_MAX);
 
     /* NTS: First In First Out order */
-    vm_event_functions[event].push_back(Function_wrapper(func,canchange,name));
+    vm_event_functions[event].emplace_back(Function_wrapper(func,canchange,name));
 }
 
 const char *VM_EVENT_string[VM_EVENT_MAX] = {
@@ -1028,14 +1050,14 @@ Config::~Config() {
 }
 
 Section* Config::GetSection(int index) {
-    for (it tel=sectionlist.begin(); tel!=sectionlist.end(); ++tel) {
+    for (it tel = sectionlist.begin(); tel != sectionlist.end(); ++tel){
         if (!index--) return (*tel);
     }
     return NULL;
 }
 
 Section* Config::GetSection(string const& _sectionname) const{
-    for (const_it tel=sectionlist.begin(); tel!=sectionlist.end(); ++tel) {
+    for (const_it tel = sectionlist.begin(); tel != sectionlist.end(); ++tel){
         if (!strcasecmp((*tel)->GetName(),_sectionname.c_str())) return (*tel);
     }
     return NULL;
@@ -1044,7 +1066,7 @@ Section* Config::GetSection(string const& _sectionname) const{
 Section* Config::GetSectionFromProperty(char const * const prop) const{
     bool log=false;
     const_it logsec;
-    for (const_it tel=sectionlist.begin(); tel!=sectionlist.end(); ++tel) {
+    for (const_it tel = sectionlist.begin(); tel != sectionlist.end(); ++tel){
         if ((*tel)->GetPropValue(prop) != NO_SUCH_PROPERTY) {
             if (!strcasecmp((*tel)->GetName(),"log")) {
                 log=true;
@@ -1066,7 +1088,7 @@ bool Config::ParseConfigFile(char const * const configfilename) {
     if (!in) return false;
     const char * settings_type;
     settings_type = (configfiles.size() == 0)? "primary":"additional";
-    configfiles.push_back(configfilename);
+    configfiles.emplace_back(configfilename);
 
     LOG(LOG_MISC,LOG_NORMAL)("Loading %s settings from config file %s", settings_type,configfilename);
 
@@ -1088,7 +1110,6 @@ bool Config::ParseConfigFile(char const * const configfilename) {
         if (!gegevens.size()) continue;
 
         switch(gegevens[0]) {
-        case '%':
         case '\0':
         case '#':
         case ' ':
@@ -1105,10 +1126,12 @@ bool Config::ParseConfigFile(char const * const configfilename) {
             testsec = NULL;
         }
             break;
+        case '%':
+            if (strcasecmp(currentsection->GetName(), "autoexec")) continue;
         default:
             try {
                 if (currentsection) {
-					bool savedata=!strcasecmp(currentsection->GetName(), "pc98")||!strcasecmp(currentsection->GetName(), "4dos")||!strcasecmp(currentsection->GetName(), "config");
+					bool savedata=!strcasecmp(currentsection->GetName(), "pc98")||!strcasecmp(currentsection->GetName(), "ttf")||!strcasecmp(currentsection->GetName(), "4dos")||!strcasecmp(currentsection->GetName(), "config");
 					if (!currentsection->HandleInputline(gegevens)&&strcasecmp(currentsection->GetName(), "autoexec")) savedata=true;
 					if (savedata) {
 						Section_prop *section = static_cast<Section_prop *>(currentsection);
@@ -1156,7 +1179,7 @@ bool CommandLine::FindExist(char const * const name,bool remove) {
     return true;
 }
 
-bool CommandLine::FindHex(char const * const name,int & value,bool remove) {
+bool CommandLine::FindHex(char const * const name,unsigned int & value,bool remove) {
     cmd_it it,it_next;
     if (!(FindEntry(name,it,true))) return false;
     it_next=it;++it_next;
@@ -1193,7 +1216,7 @@ bool CommandLine::FindCommand(unsigned int which,std::string & value) {
 }
 
 bool CommandLine::FindEntry(char const * const name,cmd_it & it,bool neednext) {
-    for (it=cmds.begin();it!=cmds.end();it++) {
+    for (it = cmds.begin(); it != cmds.end(); ++it) {
         const char *d = (*it).c_str();
 
         /* HACK: If the search string starts with -, it's a switch,
@@ -1211,7 +1234,7 @@ bool CommandLine::FindEntry(char const * const name,cmd_it & it,bool neednext) {
 
 bool CommandLine::FindStringBegin(char const* const begin,std::string & value, bool remove) {
     size_t len = strlen(begin);
-    for (cmd_it it=cmds.begin();it!=cmds.end();++it) {
+    for (cmd_it it = cmds.begin(); it != cmds.end();++it) {
         if (strncmp(begin,(*it).c_str(),len)==0) {
             value=((*it).c_str() + len);
             if (remove) cmds.erase(it);
@@ -1221,13 +1244,41 @@ bool CommandLine::FindStringBegin(char const* const begin,std::string & value, b
     return false;
 }
 
+bool CommandLine::FindStringFullBegin(char const* const begin,std::string & value, bool squote, bool remove) {
+    size_t len = strlen(begin);
+    char c = squote?'\'':'\"';
+    for (cmd_it it=cmds.begin();it!=cmds.end();++it) {
+        if (strncmp(begin,(*it).c_str(),len)==0) {
+            bool q=(*it)[len]==c;
+            value=((*it).c_str() + len + (q?1:0));
+            if (remove) cmds.erase(it);
+            if (q) {
+                std::string str=value;
+                if (str.back()==c)
+                    value.pop_back();
+                else while (str.size()&&++it!=cmds.end()) {
+                    str=(*it);
+                    if (remove) cmds.erase(it);
+                    value+=" "+str;
+                    if (str.back()==c) {
+                        value.pop_back();
+                        break;
+                    }
+                }
+            }
+            return true;
+        }
+    }
+    return false;
+}
+
 bool CommandLine::FindStringRemain(char const * const name,std::string & value) {
-    cmd_it it;value="";
+    cmd_it it;value.clear();
     if (!FindEntry(name,it)) return false;
     ++it;
-    for (;it!=cmds.end();++it) {
-        value+=" ";
-        value+=(*it);
+    for (;it != cmds.end();++it) {
+        value += " ";
+        value += (*it);
     }
     return true;
 }
@@ -1237,27 +1288,27 @@ bool CommandLine::FindStringRemain(char const * const name,std::string & value)
  * Restoring quotes back into the commands so command /C mount d "/tmp/a b" works as intended
  */
 bool CommandLine::FindStringRemainBegin(char const * const name,std::string & value) {
-    cmd_it it;value="";
+    cmd_it it;value.clear();
     if (!FindEntry(name,it)) {
         size_t len = strlen(name);
-            for (it=cmds.begin();it!=cmds.end();++it) {
+            for (it = cmds.begin();it != cmds.end();++it) {
                 if (strncasecmp(name,(*it).c_str(),len)==0) {
                     std::string temp = (*it).c_str() + len;
                     //Restore quotes for correct parsing in later stages
-                    if (temp.find(" ") != std::string::npos)
+                    if(temp.find(' ') != std::string::npos)
                         value = std::string("\"") + temp + std::string("\"");
                     else
                         value = temp;
                     break;
                 }
             }
-        if ( it == cmds.end()) return false;
+        if (it == cmds.end()) return false;
     }
     ++it;
-    for (;it!=cmds.end();++it) {
+    for (;it != cmds.end();++it) {
         value += " ";
         std::string temp = *it;
-        if (temp.find(" ") != std::string::npos)
+        if(temp.find(' ') != std::string::npos)
             value += std::string("\"") + temp + std::string("\"");
         else
             value += temp;
@@ -1268,10 +1319,10 @@ bool CommandLine::FindStringRemainBegin(char const * const name,std::string & va
 bool CommandLine::GetStringRemain(std::string & value) {
     if (!cmds.size()) return false;
 
-    cmd_it it=cmds.begin();value=(*it++);
+    cmd_it it = cmds.begin();value = (*it++);
     for(;it != cmds.end();++it) {
-        value+=" ";
-        value+=(*it);
+        value += " ";
+        value += (*it);
     }
     return true;
 }
@@ -1415,7 +1466,7 @@ void CommandLine::EndOpt() {
 }
 
 void CommandLine::FillVector(std::vector<std::string> & vector) {
-    for(cmd_it it=cmds.begin(); it != cmds.end(); ++it) {
+    for(cmd_it it = cmds.begin(); it != cmds.end(); ++it) {
         vector.push_back((*it));
     }
 #ifdef WIN32
@@ -1437,7 +1488,7 @@ int CommandLine::GetParameterFromList(const char* const params[], std::vector<st
         P_START, P_FIRSTNOMATCH, P_FIRSTMATCH
     } parsestate = P_START;
     cmd_it it = cmds.begin();
-    while(it!=cmds.end()) {
+    while(it != cmds.end()) {
         bool found = false;
         for(unsigned int i = 0; params[i] != NULL; i++) {
             if (*params[i] == 0) {
@@ -1480,7 +1531,6 @@ int CommandLine::GetParameterFromList(const char* const params[], std::vector<st
     return retval;
 }
 
-
 CommandLine::CommandLine(int argc,char const * const argv[],enum opt_style opt) {
     if (argc>0) {
         file_name=argv[0];
@@ -1491,7 +1541,7 @@ CommandLine::CommandLine(int argc,char const * const argv[],enum opt_style opt)
         if (!raw_cmdline.empty()) raw_cmdline += " ";
         raw_cmdline += argv[i];
 
-        cmds.push_back(argv[i]);
+        cmds.emplace_back(argv[i]);
         i++;
     }
 }
@@ -1503,13 +1553,12 @@ const std::string& CommandLine::GetRawCmdline(void) {
 uint16_t CommandLine::Get_arglength() {
     if (cmds.empty()) return 0;
     uint16_t i=1;
-    for(cmd_it it=cmds.begin();it != cmds.end();++it)
+    for(cmd_it it = cmds.begin();it != cmds.end(); ++it)
         i+=(*it).size() + 1;
     return --i;
 }
 
-
-CommandLine::CommandLine(char const * const name,char const * const cmdline,enum opt_style opt) {
+CommandLine::CommandLine(char const * const name,char const * const cmdline,enum opt_style opt, bool squote) {
     if (name) file_name=name;
     /* Parse the cmds and put them in the list */
     bool inword,inquote;char c;
@@ -1520,7 +1569,7 @@ CommandLine::CommandLine(char const * const name,char const * const cmdline,enum
     opt_style = opt;
     while ((c=*c_cmdline)!=0) {
         if (inquote) {
-            if (c!='"') str+=c;
+            if (c!='"'&&!squote) str+=c;
             else {
                 inquote=false;
                 cmds.push_back(str);
@@ -1534,7 +1583,7 @@ CommandLine::CommandLine(char const * const name,char const * const cmdline,enum
                 str.erase();
             }
         }
-        else if (c=='\"') { inquote=true;}
+        else if (c=='\"'&&!squote) { inquote=true;}
         else if (c!=' ') { str+=c;inword=true;}
         c_cmdline++;
     }
